/*---------------------------------------------------------------------
 * File:	if_wd.c
 * Programmer:	Michael Davison		(davison@cs.utk.edu)
 * Date:	May 1991
 *
 * Description: 
 *	This file contains a BSD 4.3 device driver for a Western Digital 
 *	80003e ethernet controller in an IBM RT workstation.
 *
 *	Entry points:
 *	-------------
 *	wdprobe()	- determine if device exists.
 *	wdattach()	- insure device is not active
 *	wdinit()	- initialize inactive driver
 *	wdreset()	- reset active driver
 *	wdioctl()	- various functions (set IP address, etc)
 *	wdoutput() 	- transfer NPDU to datalink layer for transmission 
 *	wdint()		- interrupt handler.
 *---------------------------------------------------------------------*/

/*
 * "wd.h" is generated by config, NWD is the  number of devices supported.
 */
#include   "wd.h"
#if NWD > 0

#ifdef IBMRTPC
#ifndef ATR

/* 
 * Include all the standard kernel include files.
 */
#include "../machine/pte.h"

#include "param.h"
#include "systm.h"
#include "mbuf.h"
#include "buf.h"
#include "protosw.h"
#include "socket.h"
#include "vmmac.h"
#include "ioctl.h"
#include "errno.h"

#include "../net/if.h"
#include "../net/netisr.h"
#include "../net/route.h"

#ifdef INET
#include "../netinet/in.h"
#include "../netinet/in_systm.h"
#include "../netinet/in_var.h"
#include "../netinet/ip.h"
#include "../netinet/if_ether.h"
#endif INET

#ifdef NS
#include "../netns/ns.h"
#include "../netns/hs_if.h"
#endif NS

#include "../machine/io.h"
#include "if_wdreg.h"
#include "../machineio/ioccvar.h"
#include "../machine/debug.h"



/*
 * Local function prototypes.
 */
int wdprobe(), wdattach(), wdinit(), wdioctl();

int wdoutput(), wdreset(), wdint(), wdput();
char getCURR();
void wdxint(), wdstart(), wdrint();
struct mbuf *wdget();
u_short getshortatoff();




/*===================================================================
 *                     auto-configuration section.
 *===================================================================*/

/*
 * Define possible csr (control/status register) addresses. This may be 
 * used to validate config file addresses. Addresses for the PC I/O space
 * are translated to 0xf0000XXX and address for the PC memory-mapped 
 * space are translated to 0xf40XXXXX. 
 */
caddr_t wdstd[] = { 
	(caddr_t)0xf0000200, (caddr_t)0xf0000220, (caddr_t)0xf0000240,
	(caddr_t)0xf0000260, (caddr_t)0xf0000280, (caddr_t)0xf00002a0,
	(caddr_t)0xf00002c0, (caddr_t)0xf00002e0, (caddr_t)0xf0000300,
	(caddr_t)0xf0000320, (caddr_t)0xf0000340, (caddr_t)0xf0000360,
	(caddr_t)0xf0000380, (caddr_t)0xf00003a0, (caddr_t)0xf00003c0,
	(caddr_t)0xf00003e0, 0
};


/*
 * Declare "xxdevice" structure. 1 per device. Filled-in by kernel
 * during boot-time auto-config.
 */
struct iocc_device *wdinfo[NWD];


/*
 * Define "xxdriver" structure. 1 per driver.  Used by kernel during
 * run-time. 
 */
struct iocc_driver wddriver = {
	wdprobe,		/* xxprobe(), is device present?	*/
	0,			/* xxslave(), is slave present?		*/
	wdattach,		/* xxattach(), initialize driver	*/
	0,			/* xxdgo(), initialize DMA		*/
	wdstd,			/* device csr addresses			*/
	"wd",			/* name of device			*/
	wdinfo,			/* backpointer to iodinit struct	*/
	0,			/* name of controller			*/
	0,			/* backpointer ot iominit struct	*/ 
	wdint,			/* interrupt service routine		*/
	0,	 		/* read/write csr offset		*/
	0			/* driver suspend routine		*/
};





/*===================================================================
 *                        driver global data.
 *===================================================================*/

/*
 * The network interface structure (struct arpcom us_ac) must be first,
 * and is used by other kernel routines.  The remaining members of this
 * structure are used within this driver only.
 */
struct wd_softc {
	struct    arpcom us_ac;		/* generic network interface	*/
#define us_if     us_ac.ac_if		/* ifnet struct			*/
#define us_addr   us_ac.ac_enaddr	/* ethernet address from ROM	*/

	char      *us_xbuf[2];		/* base of transmit buffers	*/
	unsigned  us_xlen[2];		/* length used of xmit buffers	*/
	char      us_tpsr[2];		/* calculated xmit page start	*/
	unsigned  us_xfull[2];		/* 1 if transmit buffer filled	*/
	unsigned  us_xcurr;		/* which transmit buffer in use	*/
	unsigned  us_oactive;  		/* 1 if output active		*/
	int	  us_missed;		/* missed xmit interrupt count	*/

	char      *us_rbuf;		/* base of receive buffers	*/
	char      us_nextpage;		/* next recieve buffer page	*/
	char      us_pstart;		/* PSTART value for NIC		*/
	char      us_pstop; 		/* PSTOP value for NIC		*/
} wd_softc[NWD];





/*-------------------------------------------------------------------
 * function:	wdprobe, local version of XXprobe.
 * parameters:	dev - address of XXdevice structure.
 *
 * desciption:	Determine if device exists at the  kernel virtual
 *		address This is accomplished by computing a checksum
 *		on a portion of the Western Digital controller. 
 *-------------------------------------------------------------------*/
int wdprobe(wdp)
register struct wddevice *wdp;
{
	char checksum = 0;
	char *ptr = wdp->addr;
	char *lim = &wdp->cksum;

	while( ptr <= lim )
		checksum += IOIN(ptr++);

	return( ( (checksum == 0xff) ? PROBE_NOINT : PROBE_BAD) );
}



/*-----------------------------------------------------------------------
 * function:	wdattach, local version of XXattach.
 * parameters:	cntlr  - address of XXinfo structure.
 *
 * description:	Do boot-time, device-specific setup and initialization. 
 *		Set up the interface structure.  This structure is used
 *		by the network layer (IP and ARP ) to interact with the 
 *		data link layer (Ethernet device driver).  The address of  
 *		the on-controller, memory-mapped buffer is read from the 
 *		"iod_flags" config file argument.  This little hack is the
 *		approved workaround for a limitation in the BSD config file
 *		format - controllers may only have one address.  Lastly 
 *		the hardware address is read, stored, and displayed.
 *-----------------------------------------------------------------------*/
int wdattach( cntlr )
struct iocc_device *cntlr;
{
	struct wddevice *base = (struct wddevice *)cntlr->iod_addr;
	struct wd_softc *us   = &wd_softc[cntlr->iod_unit];
	struct ifnet    *ifp  = &us->us_if;
	char *inp;
	int  i;

	/* 
	 * Reset the controller and leave it deactivated.
 	 */
	IOOUT( &base->cntrl, (char)0x80);
	IOOUT( &base->cntrl, (char)0x00);

	/*
	 * Grab the address of the controller's buffer.
	 */
	us->us_xbuf[0]	= (char *)cntlr->iod_flags;
	us->us_xbuf[1]	= us->us_xbuf[0] + WD_XBUFSIZE;
	us->us_rbuf	= us->us_xbuf[1] + WD_XBUFSIZE;
	us->us_pstart	= (char)( ( (unsigned long)us->us_rbuf >> 8) & 0xff);
	us->us_pstop	= us->us_pstart + (char)(((WD_NUMRBUFS*256) >> 8)&0xff);
	us->us_nextpage	= us->us_pstart + 1; 
	us->us_tpsr[0]	= (char)(((unsigned long)(us->us_xbuf[0]) >> 8) & 0xff);
	us->us_tpsr[1]	= (char)(((unsigned long)(us->us_xbuf[1]) >> 8) & 0xff);

	/*
	 * Read and display the hardware address.
	 */
	for(inp = base->addr, i=0; i<ETH_ADDR_SIZE; i++)
		us->us_addr[i] = IOIN(inp++);

	printf("wd%d: ethernet address ", ifp->if_unit);
	printf("%x", us->us_addr[0] );
	for(i=1; i<ETH_ADDR_SIZE; i++)
		printf(":%x", us->us_addr[i] );
	printf("\n");

	/*
	 * Setup the iterface structure. This structure defines the 
	 * network layer interface. 
	 */
	ifp->if_unit	= cntlr->iod_unit;
	ifp->if_name	= "wd";
	ifp->if_mtu	= ETHERMTU;
	ifp->if_init	= wdinit;
	ifp->if_ioctl	= wdioctl;
	ifp->if_output	= wdoutput;
	ifp->if_reset	= wdreset;
	ifp->if_flags	= IFF_BROADCAST;
	if_attach(ifp); 

	return(1);
}


/*-------------------------------------------------------------------
 * function: 	wdinit, initialize the interface
 * parameters:	unit,   device number
 *
 * description:	Initialize the controller (the card). This includes 
 *		extensive initialization of the NS8390 ethernet chip.
 *-------------------------------------------------------------------*/
int wdinit(unit)
int unit;
{
	struct wddevice *addr = (struct wddevice *)(wdinfo[unit]->iod_addr);
	struct wd_softc *us   = &wd_softc[unit];
	struct ifnet *ifp     = &us->us_if;
	struct mbuf *m;
	int    i, s;
	char   *inp, *outp, *limp;

	if( ifp->if_addrlist == (struct ifaddr *)0 ){
		printf("wd%d: no address\n",unit);
		return(0);
	}

	if( ifp->if_flags & IFF_RUNNING ){
		printf("wd%d: attempt to initialize active interface.\n",unit);
		return(1);
	}

	us->us_oactive	= 0;
	us->us_xcurr	= 0;

	ifp->if_flags |= (IFF_UP|IFF_RUNNING);

	/*
	 * Initialize the WD8003e controller (not the NIC). Toggle the 
	 * reset bit, set the memory-enable bit, set the A18-A13 bits. 
	 */
	IOOUT( &addr->cntrl, (char)0x80);
	IOOUT( &addr->cntrl, (char)0x00);
 	IOOUT( &addr->cntrl, 
	    (char)(0x40 | ( ((unsigned long)us->us_xbuf[0] >> 13) & 0x3f) ) );

	/* 
	 * Initialize page 0 of the NIC
	 */
	IOOUT( &addr->cr,	(char)(RD2|STP) );
	IOOUT( &addr->dcr,	(char)FT1 );
	IOOUT( &addr->rbcr0,	(char)0);
	IOOUT( &addr->rbcr1,	(char)0);
	IOOUT( &addr->rcr,	(char)AB );
	IOOUT( &addr->tcr,	(char)0 );
	IOOUT( &addr->pstart,	us->us_pstart);
	IOOUT( &addr->pstop,	us->us_pstop);
	IOOUT( &addr->bndry,	us->us_pstart);
	IOOUT( &addr->isr,	(char)0xff );
	IOOUT( &addr->imr,	(char)(PRXE|PTXE|RXEE|TXEE|OVWE) ); 

	/*
	 * Initialize page 1 of the NIC, copy the ROM ether address into 
	 * the NIC ether address, set CURR, and set multicast address
	 * registers to 00:00:00:00:00:00 (no multicast reception).
	 */
	IOOUT( &addr->cr, (char)(RD2|PS0|STP) );

	inp  = (char *)us->us_addr;
	outp = (char *)&addr->par0;
	limp = (char *)&addr->par0 + ETH_ADDR_SIZE;

	while( outp < limp )
		IOOUT( outp++, *inp++ );
 
	IOOUT(outp++, us->us_nextpage ); 

	for(i=0; i<8; i++)
		IOOUT(outp++, (char)0 );

	IOOUT( &addr->cr,  (char)(STA|RD2) );

	/*
	 * If there is any ouput queued, start sending it.
	 */
	s = splimp(); 
	if( ifp->if_snd.ifq_head ){
		IF_DEQUEUE(&ifp->if_snd, m);
		if( wdput(us, unit, m, 0) ){
			wdstart(us, addr, 0);
		}
		if( ifp->if_snd.ifq_head ){
			IF_DEQUEUE(&ifp->if_snd, m);
			wdput(us, unit, m, 1);
			if( !us->us_oactive ){
				wdstart(us, addr, 1);
			}
		}
	}	
	splx(s);

	return(1);
}



/*-------------------------------------------------------------------
 * function: 	wdioctl, handle ioctl calls to the interface
 * parameters:	ifp	- address of interface structure
 *		cmd	- command to be executed
 *		data	- address of any parameters 
 *
 * description:	Various fuctions may be performed here. Most important
 *		is the setting of the network address of this end system.
 *-------------------------------------------------------------------*/
int wdioctl(ifp, cmd, data)
struct ifnet *ifp;
int cmd;
caddr_t data;
{
	struct ifaddr *ifa   = (struct ifaddr *)data;
	struct arpcom *ifarp = (struct arpcom *)ifp;
	struct ifreq *ifr   = (struct ifreq *)ifp;
	int flags, error = 0, s = splimp();
#ifdef NS
	struct ns_addr *ina = &(IA_SNS(ifa)->sns_addr);
	struct wd_softc *us = &wd_softc[ifp->if_unit];
#endif NS

	switch(cmd){
		case SIOCSIFADDR:
			switch( ifa->ifa_addr.sa_family ){
#ifdef INET
			case AF_INET:
				wdinit(ifp->if_unit);			
				ifarp->ac_ipaddr = IA_SIN(ifa)->sin_addr;
				arpwhohas(ifarp, &IA_SIN(ifa)->sin_addr);
				break;
#endif INET
#ifdef NS
			case AF_NS:
				if( ns_nullhost(*ina) ){
					ina->x_host = 
					  *(union ns_host *)(us->us_addr);
				}
				else {
					bcopy( (caddr_t)ina->x_host.c_host,
					  (caddr_t)us->us_addr, 
					  sizeof(us->us_addr) );	
					wdinit(ifp->if_unit);			
				}
				break;
#endif NS
			default:
				printf("wd%d: address family not supported\n",
				  ifp->if_unit);
				error = EINVAL;
			}
			break;
		case SIOCSIFFLAGS:	
			flags = ifp->if_flags;
			if( !(flags&IFF_UP) && (flags&IFF_RUNNING) ){
				wdstop(ifp->if_unit);
				ifp->if_flags &= ~IFF_RUNNING;
			}
			else if( (flags&IFF_UP) && !(flags&IFF_RUNNING) ){
				wdinit(ifp->if_unit);
			}
			break;
		case SIOCGIFFLAGS:	
			ifr->ifr_name[0] = 'w';
			ifr->ifr_name[1] = 'd';
			ifr->ifr_name[2] = (char)ifp->if_unit + '0';
			ifr->ifr_name[3] = '\0';
			ifr->ifr_flags = ifp->if_flags;
			break;
		default:
			error = EINVAL;
	}

	splx(s);
	return(error);
}



/*-------------------------------------------------------------------
 * function: 	wdoutput, ouput through interface.
 * parameters:	ifp	- Interface to use. 
 *		m0	- mbuf list of output (NPDU)
 *		dst	- destination sockaddr structure (IP address)
 *
 * description:	Construct the DPDU header, enque it on the ouput queue
 * 		and start the device transmission if device is inactive.	
 *-------------------------------------------------------------------*/
int wdoutput(ifp, m0, dst)
struct ifnet    *ifp;
struct mbuf     *m0;
struct sockaddr *dst;
{
	int  trailer, s, type, off, error=0;
	char edst[ETH_ADDR_SIZE];
	struct wddevice *addr = 
	  (struct wddevice *)wdinfo[ifp->if_unit]->iod_addr;
	struct in_addr idst;
	struct wd_softc *us = &wd_softc[ifp->if_unit];
	struct mbuf *m=m0;
	struct ether_header *eh;

	if( (ifp->if_flags & (IFF_UP|IFF_RUNNING)) != (IFF_UP|IFF_RUNNING) ){
		error = ENETDOWN;
		goto dump;
	}

	switch( dst->sa_family ){
#ifdef INET
	case AF_INET:
		idst = ((struct sockaddr_in *)dst)->sin_addr;
		if( !arpresolve(&us->us_ac, m, &idst, edst, &trailer) ){
			return(0);
		}

		off = ntohs( (u_short)mtod(m, struct ip *)->ip_len) - m->m_len;

		if(trailer && off > 0 && !(off&0x1ff) && m->m_off >= MMINOFF+4){
			type = ETHERTYPE_TRAIL + (off >> 9);
			
			m->m_off -= 4;
			m->m_len += 4;
		
			*mtod(m, u_short *) = htons( (u_short)ETHERTYPE_IP );	
			*(mtod(m, u_short *)+1) = htons( (u_short)m->m_len);

			while( m->m_next )
				m = m->m_next;

			m->m_next = m0;
			m = m0->m_next;
			m0->m_next = 0;
			m0 = m;
		}	
		else {
			type = ETHERTYPE_IP;
		}
		break;
#endif INET
#ifdef NS
	case AF_NS:
		bcopy( (caddr_t)&((struct sockaddr_ns *)dst)->sns_addr.x_host),
		  (caddr_t)edst, sizeof(edst) );
		type 	= ETHERTYPE_NS;
		offset	= 0;
		break;
#endif NS
	case AF_UNSPEC:
		eh = (struct ether_header *)dst->sa_data;
		bcopy((char *)eh->ether_dhost, (char *)edst, sizeof(edst));
		type = eh->ether_type;
		break;
	default:	
		printf("wd%d: address family %d not supported\n",
		  ifp->if_unit,dst->sa_family);
		error = EAFNOSUPPORT;
		goto dump;
	}

	/*
	 * Build LAN packet header. Another mbuf may be needed.
	 * The header includes destination ether address, source
	 * ether address, and the packet type.
	 */

	if( m->m_off > MMAXOFF  ||  MMINOFF+sizeof(*eh) > m->m_off ){
		if( (m = m_get(M_DONTWAIT, MT_HEADER)) == 0){
			error = ENOBUFS;
			goto dump;
		}
		m->m_next = m0;
		m->m_off  = MMINOFF;
		m->m_len  = sizeof(struct ether_header);
	}
	else {
		m->m_off -= sizeof(struct ether_header);
		m->m_len += sizeof(struct ether_header);
	}

	eh = mtod(m, struct ether_header * );
	bcopy((char *)edst, (char *)eh->ether_dhost, ETH_ADDR_SIZE);
	bcopy((char *)us->us_addr, (char *)eh->ether_shost, ETH_ADDR_SIZE); 
	eh->ether_type = htons( (unsigned short)type );

	/*
	 * Catch missed xmit interrupts.  If we miss 2 in row 
	 * display an error message.  
	 */
	if( us->us_oactive  &&  !(IOIN(&addr->cr)&TXP) ){
		if( (us->us_missed += 2) > 3 ){
			printf("wd%d: Missed ",ifp->if_unit);
			printf("transmit interrupt!\n");
			us->us_missed = 0;
		}
		wdxint(ifp->if_unit, us, addr );
	}

	/*
	 * Queue the packet for transmision.  If the transmit buffer
	 * is empty use it.
	 */

	s = splimp();
	if( us->us_xfull[0] == 0 ){
		if( wdput(us, ifp->if_unit, m, 0) && !us->us_oactive ){
			wdstart(us, addr, 0);
		}
	}
	else if( us->us_xfull[1] == 0 ){
		if( wdput(us, ifp->if_unit, m, 1)  && !us->us_oactive){
			wdstart(us, addr, 1);
		}
	}
	else if( IF_QFULL(&ifp->if_snd) ){
		IF_DROP(&ifp->if_snd);
		m_freem(m);
		error = ENOBUFS;
	}
	else {
		IF_ENQUEUE(&ifp->if_snd, m);
	}
	splx(s);
	return(error);
		
dump:	m_freem(m0);
	return(error);
}


/*-------------------------------------------------------------------
 * function: 	wdput, copy packet from mbuf chain to xmit buffer
 * parameters:	us	- address of global data
 *		unit	- number of device
 *		m	- linked-list of mbufs (data packet)
 *		xbuff	- which xmit buffer to use.
 *
 * description:	Copy the mbuf chain to the indicated transmit buffer. 
 *		Check for excessive packet size and mark the buffer in
 *		use.
 *-------------------------------------------------------------------*/
int wdput(us, unit, m, xbuff)
struct wd_softc *us;
int unit;
struct mbuf *m;
int xbuff;
{
	register char *outp = us->us_xbuf[xbuff];
	register int cnt=0;
	register struct mbuf *mp;

 	for( mp=m; mp; mp = mp->m_next ){
		if((cnt += mp->m_len) > MAX_DPDU ){
			printf("wd%d: oversized packet ", unit);
			printf("transmission - packet dropped\n");
			m_freem(m);
			return(0);
		}
		bcopy(mtod(mp, char *), outp, mp->m_len);
		outp += mp->m_len;
	}
	
	us->us_xlen[xbuff]  = ((cnt < MIN_DPDU) ? MIN_DPDU : cnt);
	us->us_xfull[xbuff] = 1;

	m_freem(m);
	return(1);
}



/*-------------------------------------------------------------------
 * function: 	wdreset, reset the network interface
 * parameters:	unit	- device number
 *
 * description:	Mark the device "down" and then call wdinit.
 *-------------------------------------------------------------------*/
int wdreset(unit)
int unit;
{
	if( unit < NWD  &&  wdinfo[unit]->iod_alive ){
		wd_softc[unit].us_if.if_flags &= ~IFF_RUNNING;
		return( wdinit(unit) );
	}

	return(0);
}


/*-------------------------------------------------------------------
 * function:	wdstop, disable the controller.
 * parameters:	unit	- unit number of device
 *
 * description:	stop all interrupts and disable controller.
 *-------------------------------------------------------------------*/
int wdstop(unit)
int unit;
{
	struct wddevice *base = (struct wddevice *)wdinfo[unit]->iod_addr;
	int cnt=0;

	IOOUT( &base->cr, STP );

	while( !(IOIN(&base->isr)&RST) && ++cnt < 5000 );
	if( cnt == 5000 ){
		printf("wd%d: controller would not reset!\n",unit);
		return(0);
	}

	return(1);
}


/*-------------------------------------------------------------------
 * function:	wdstart, enable transmit interrupts.
 * parameters:	us	- pointer to global data structure
 *		addr	- base address of device
 *
 * description:	Kick start ethernet transmission when output has been 
 *		placed in an output buffer and output is not already 
 *		active. (check us_oactive prior to calling)
 *-------------------------------------------------------------------*/
void wdstart(us, addr, xbuff)
struct wd_softc *us;
struct wddevice *addr;
int xbuff;
{
	us->us_oactive = 1;
	us->us_xcurr   = xbuff;

	IOOUT(&addr->tpsr,  us->us_tpsr[xbuff]);
	IOOUT(&addr->tbcr0, (char)(us->us_xlen[xbuff] & 0xff) );
	IOOUT(&addr->tbcr1, (char)((us->us_xlen[xbuff] >> 8) & 0xff) );
	IOOUT(&addr->cr,    IOIN(&addr->cr)|TXP);	
}


/*-------------------------------------------------------------------
 * function: 	wdint, interrupt handler.	
 * parameters:	unit  - the unit that interrupted.	
 *
 * description:	Field the interrupt for this device. Receive interrupts
 *		result in wdrint() being called while transmit 
 *		interrupts result in wdxint() being called. Zero is 
 * 		returned to indicate this interrupt was ours.
 *-------------------------------------------------------------------*/
int wdint( unit )
register int unit;
{
	register struct wd_softc *us = &wd_softc[unit];
	register struct wddevice *base = 
		 (struct wddevice *)wdinfo[unit]->iod_addr;
	register int status;

	IOOUT( &base->imr, (char)0 );

	while( ( status = (IOIN(&base->isr)&(OVW|TXE|RXE|PTX|PRX) ) ) ){
		IOOUT(&base->isr, (char)status);

		if( status&OVW  ||  status&PRX   ||  status&RXE )
			wdrint(unit,us,base);

		if( status&PTX  ||  status&TXE)
			wdxint(unit,us,base);
	} 

 	IOOUT( &base->imr, (char)(PRXE|PTXE|RXEE|TXEE|OVWE) );
	return(0);
}



/*-----------------------------------------------------------------------
 * function: 	wdrint, receive interrupt handler.	
 * parameters:	unit	- unit causing interrupt
 *		us	- global software status of this interface
 *		base	- base address for this controller.
 *
 * description:	grab incoming packets. This function is repeated until
 *		the receive buffers are empty, indicated when the next
 *		page pointer (where we read from) is equal to the 
 *		current page pointer (where the cotroller writes to).	
 *-----------------------------------------------------------------------*/
void wdrint(unit,us,base)
register int  unit;
register struct wd_softc *us;
register struct wddevice *base;
{
	register char *inp;
	register struct mbuf *mhead;
	register struct ether_header *eh;
	register struct ifqueue *inq=0;
	register unsigned type, totlen, rstat, next;
	register int  s, residual, offset;

	while( us->us_nextpage != getCURR(base) ){
		us->us_if.if_ipackets++;

		inp  = (char *)( ((unsigned long)us->us_rbuf & 0xffff0000) +
	    	  ((unsigned long)us->us_nextpage << 8) );

		rstat   = MM_IN( inp++ );
		next    = MM_IN( inp++ );
		totlen  = MM_IN(inp++);
		totlen += ((unsigned)(MM_IN(inp++)) << 8);

		if( !(rstat&PRXS) ){
			us->us_if.if_ierrors++;
			goto dump;
		}

		/*
	 	 * Read the ethernet packet header.
	 	 */
		eh      = (struct ether_header *)inp;
		totlen -= sizeof(struct ether_header);
		inp    += sizeof(struct ether_header);
		type   =  ntohs( (u_short)MM_INW(&eh->ether_type) );

		/*
		 * If this is a "trailer" packet determine the actual 
		 * length and type.  Offset is set to indicate the offset
		 * of the NPDU header offset. (after ether header)
		 * wdget() will move the trailer to the header location.
		 */
		if( type < ETHER_TRAILER_LOW  ||  type >= ETHER_TRAILER_HIGH ){
			offset = 0;
		}
		else {
			if((offset = (type-ETHERTYPE_TRAIL)*512) > ETHERMTU)
				goto dump;

			type     =  getshortatoff(us,eh,offset);
			residual =  getshortatoff(us,eh,offset+2);

			if( (offset + residual) > totlen )
				goto dump;
			totlen = offset+residual;
		}

		/*
		 * Get the packet and pass it to the next layer.
		 */
		switch( type ){
#ifdef INET
		case ETHERTYPE_IP:
			if( (mhead = wdget(base, inp, totlen, us, offset)) ){
				schednetisr(NETISR_IP);
				s   = splimp();
				inq = &ipintrq;
				if( IF_QFULL(inq) ){
					IF_DROP(inq);
					m_freem(mhead);
				}
				else {
					IF_ENQUEUE(inq,mhead); 
				}
				splx(s);	
			}
			break;
#endif INET
#ifdef NS
		case ETHERTYPE_NS:
			schednetisr(NETISR_NS);
			inq = &nsintrq;
			break;
#endif NS
		case ETHERTYPE_ARP:
			if( (mhead = wdget(base, inp, totlen, us, offset)) ){
				arpinput(&us->us_ac, mhead);
			}
		}

		/* 
		 * Update controller status to show removed packet.
		 */
dump:	us->us_nextpage = next;
		IOOUT( &base->bndry, (char)(next-1) ); 
		if( IOIN(&base->bndry) < us->us_pstart )
			IOOUT( &base->bndry, (char)(us->us_pstop - 1) );

	}
}


/*-------------------------------------------------------------------
 * function:	getshortatoff
 * parameters:	us	- address of software status
 *		eh	- address of ether header
 *		offset	- offset from ether header
 *
 * description:	When "trailer" packets are received the actual type and
 *		length are stored in .... you guessed it.. the trailer.
 *		This means we need to read a word at an offset from the
 *		ether header.  This may wrap in the receive buffer, so
 *		this routine handles that problem.
 *-------------------------------------------------------------------*/
u_short getshortatoff(us,eh,offset)
register struct wd_softc *us;
register struct ether_header *eh;
register int offset;
{
	register char *limit = us->us_rbuf + WD_NUMRBUFS * WD_RBUFSIZE;
	register char *p     = (char *)(eh + 1) + offset;

	if( p >= limit )
		p -= WD_NUMRBUFS * WD_RBUFSIZE;
	
	return( ntohs( (u_short)MM_INW(p)) );
}


/*-------------------------------------------------------------------
 * function:	getCURR, read the current page pointer 
 * parameters:	base   - base address of controller
 *
 * description:	wdrint() uses this function to determine if all the 
 *		received packets have been processed. The current page
 *		pointer (CURR) indicates where the controller will put
 *		the next incoming packet.
 *-------------------------------------------------------------------*/
char getCURR(base)
struct wddevice *base;
{
	register char ret;

	IOOUT( &base->cr, IOIN(&base->cr) | PS0 );
	ret = IOIN(&base->curr);
	IOOUT( &base->cr, IOIN(&base->cr) & ~PS0 );	

	return( ret );
}


/*-------------------------------------------------------------------
 * function: 	wdget.
 * parameters:	base	- controller base address
 *		rbuf	- buffer address to begin reading from
 *		totlen	- length of packet
 *		us	- address of global structure 
 *		offset	- offset to header
 *
 * description:	Copy a packet from the controller's buffer to a string
 *		of mbufs.  
 *-------------------------------------------------------------------*/
struct mbuf *wdget(base, rbuf, totlen, us, offset)
register struct wddevice *base;
register char *rbuf;
register int  totlen;
register struct wd_softc *us;
register int offset;
{
	struct ifnet *ifp =  &us->us_if;
	struct mbuf *mhead=0;
	register char *inp = rbuf;
	register char *outp; 
	register char *limit = us->us_rbuf + WD_NUMRBUFS*WD_RBUFSIZE;
	register struct mbuf  *m; 
	register struct mbuf  **mp = &mhead;
	register int n, len, off = offset;

	while( totlen > 0 ){
		MGET(m, M_DONTWAIT, MT_DATA);
		if( m == 0 ){
			if( mhead != 0 ){
				m_freem(mhead);
			}
			return(0);
		}

		if( off ){
			len = totlen - off;
			inp = rbuf + off;
		}
		else {
			len = totlen;
		}

		len = MIN( MLEN, (len + (ifp == 0 ? 0 : sizeof(ifp))) );

		m->m_len  = len;
		m->m_off  = MMINOFF;
		m->m_next = 0;

		outp = mtod(m, char * );

		if( ifp != 0 ){
			bcopy( (char *)&ifp, outp, sizeof(ifp) ); 
			outp += sizeof(ifp);
			len  -= sizeof(ifp);
			ifp  = 0;
		}


		if( inp+len <= limit ){
			bcopy(inp,outp,len);
		}
		else if( inp >= limit ){
			inp -= WD_NUMRBUFS*WD_RBUFSIZE;
			bcopy(inp,outp,len);
		}
		else {
			n = limit - inp;
			bcopy(inp,outp,n);
			outp += n;
			bcopy(us->us_rbuf,outp,len-n);
		}

		inp += len;

		*mp = m;
		mp  = &m->m_next;

		if( off == 0 ){
			totlen -= len;
		}
		else if( (off += len) == totlen ){
			inp = rbuf;
			offset = 0;
			totlen = offset;
		}
	}

	return(mhead);
}


/*-------------------------------------------------------------------
 * function:	wdxint, catch transmit interrupts	
 * parameters:	unit	- unit number for this device.
 *		us	- global software status of this interface
 *		base	- base address for this controller.
 *
 * description:	A packet has been transmitted.  Check the transmit status
 *		and update the iterface accounting information.  If the
 *		other transmit buffer is ready begin sending it.  If a 
 *		packet is ready on the output queue place it in the 
 *		buffer that was just emptied.
 *-------------------------------------------------------------------*/
void wdxint(unit,us,base)
int unit;
struct wd_softc *us;
struct wddevice *base;
{
	register struct mbuf  *m;
	register int next_xmit, status;

	status = IOIN( &base->tsr ) & (PTXS|COL|CRS|OWC);

	if( status&PTXS )
		us->us_if.if_opackets++;
	else
		us->us_if.if_oerrors++;

	if( status&COL )
		us->us_if.if_collisions += IOIN(&base->ncr);
	if( status&OWC )
		us->us_if.if_collisions++;

	if( status&CRS ){
		printf("wd%d: No carrier. Interface connected?\n",unit);
		us->us_if.if_oerrors++;
	}

	if( --(us->us_missed) < 0 )
		us->us_missed = 0;

	us->us_oactive	= 0;
	us->us_xfull[us->us_xcurr] = 0;

	next_xmit = 1 - us->us_xcurr;

	if( us->us_xfull[ next_xmit ] ){
		wdstart(us, base, next_xmit);

		if( us->us_if.if_snd.ifq_head ){
			IF_DEQUEUE(&us->us_if.if_snd, m);		
			wdput(us, unit, m, 1-next_xmit);
		}
	}
}


#endif  ATR
#endif  IBMRTPC 
#endif  NWD 
