/*
 * mmap() is used in two situations in nvi: when opening a file
 * (rec_open.c) and when searching for tags (ex_tag.c).  In the first
 * case, an alternative method will be used if mmap() doesn't work.
 *
 * The tag search code, however, depends on mmap() because it maps the 
 * tags file into memory and does a binary search for the tag.  In this 
 * case, we provide a replacement called 'mmap_hack()' that allocates 
 * memory and then reads the "mapped" file into it.  This only makes 
 * sense if the mapped region is read-only and relatively small, both 
 * of which are generally true for the tags file.
 *
 */

#include <sys/types.h>
#include <sys/mman.h>
#include <errno.h>
#include <string.h>
#include "compat.h"


#define MMAP_ERROR	((char *) -1)


/*
 * This function emulates mmap() by reading `len' bytes from the
 * file descriptor `fd' and returning a pointer to that memory.
 * The "mapped" region can later be deallocated with munmap().
 *
 * Note that `fd' must be seekable and that the `addr', `flags', and
 * `off', arguments are ignored.  Any modes other than `PROT_READ'
 * return an error with ENOSYS (unimplemented function).
 */
char *
mmap_hack(addr, len, prot, flags, fd, off)
	char 	*addr;
	size_t	len;
	int	prot;
	int	flags;
	int	fd;
	off_t	off;
{
	char	*ptr;
	off_t	pos;
	long	n;

	if (prot != PROT_READ) {
		errno = ENOSYS;
		return(MMAP_ERROR);
	}

	ptr = (char *) malloc(len);
	if (ptr == NULL) {
		return(MMAP_ERROR);
	}

	pos = lseek(fd, 0L, SEEK_CUR);
	n = read(fd, ptr, len);
	lseek(fd, pos, SEEK_SET);

	if (n < 0) {
		free(ptr);
		return((char *) ptr);
	}

	if (n < len) {
		memset((ptr + n), 0, (len - n));
	}

	return(ptr);
}


int
munmap_hack(addr, len)
	char	*addr;
	size_t	len;
{
	free(addr);
	return(0);
}


#ifdef __MACHTEN__
/*
 * MachTen 2.1.1 Note:
 *   mmap() doesn't work as expected on MachTen 2.1.1.  It sometimes
 *   returns an address other than ((char *) -1), leading the caller
 *   to expect that this is a valid address, but then the mapped region
 *   contains garbage or accessing it causes a segmentation violation.
 *   Here we provide a saner version that always returns ((char *) -1).
 */

char *
mmap(addr, len, prot, flags, fd, off)
	char 	*addr;
	size_t	len;
	int	prot;
	int	flags;
	int	fd;
	off_t	off;
{
	errno = ENOSYS;
	return(MMAP_ERROR);
}

int
munmap(addr, len)
	char	*addr;
	size_t	len;
{
	return(0);
}
#endif /* __MACHTEN__ */
