.\"
.\" Must use  --  tbl -- for this one
.\"
.\" @(#)rpcgen.ms	1.2 90/06/21 4.0NFSSRC
.NH 1
\&\fBrpcgen\fP Programming Guide
.NH 2
\&The \fBrpcgen\fP Protocol Compiler
..IX rpcgen "" \fIrpcgen\fP "" PAGE MAJOR
.LP
..IX RPC "" "" \fIrpcgen\fP
The details of programming applications to use Remote Procedure Calls
can be tedious.  One of the more difficult areas is writing XDR
routines to convert procedure arguments and results into
their network format and vice-versa.
.LP
Fortunately,
.I rpcgen(1) 
exists to help programmers write RPC applications simply and directly.
.I rpcgen 
does most of the dirty work, allowing programmers to debug
the main features of their application, instead of requiring them to
spend most of their time on their network interface code.
.LP
.I rpcgen 
is a compiler.  It accepts a remote program interface definition written in
a language, called RPC Language, which is similar to C.  It produces a C
language output for RPC programs.  This output
includes skeleton versions of the client routines, a server skeleton, XDR
filter routines for both parameters and results, a header file that
contains common definitions and, optionally, dispatch tables which the
server can use to check authorizations and then invoke service routines.
The client skeletons' interface with the RPC library and effectively hide the
network from their callers.  The server skeleton similarly hides the network
from the server procedures that are to be invoked by remote clients.
.I rpcgen 's
output files can be compiled and linked in the usual way.  The server code
generated by
.I rpcgen 
has support for
.I inetd 
i.e. the server can be started via
.I inetd 
or at the command line.
.LP
The developer writes server procedures\(emin any language that observes 
system calling conventions\(emand links them with the server skeleton 
produced by
.I rpcgen 
to get an executable server program.  To use a remote program, a programmer
writes an ordinary main program that makes local procedure calls to the
client skeletons.
Linking this program with
.I rpcgen 's
skeletons creates an executable program.
.I rpcgen 
options can be used to suppress skeleton generation and to specify the
transport to be used by the server skeleton.
.LP
Like all compilers,
.I rpcgen 
reduces development time
that would otherwise be spent coding and debugging low-level routines.
All compilers, including
.I rpcgen ,
do this at a small cost in efficiency and flexibility.  However, many 
compilers allow escape hatches for programmers to mix low-level code with 
high-level code.
.I rpcgen 
is no exception.  In speed-critical applications, hand-written routines
can be linked with the
.I rpcgen 
output without any difficulty.  Also, one may proceed by using
.I rpcgen 
output as a starting point, and then rewriting it as necessary.
(For a discussion of RPC programming without
.I rpcgen ,
see the next chapter, the \fIRemote Procedure Call
Programming Guide\fP.)  
.NH 3
\&Converting Local Procedures into Remote Procedures
..IX rpcgen "local procedures" \fIrpcgen\fP
..IX rpcgen "remote procedures" \fIrpcgen\fP
.LP
Assume an application that runs on a single machine, one which we want
to convert to run over the network.  Here we will demonstrate such a
conversion by way of a simple example\(ema program that prints a
message to the console:

.DS
.ft CW
.ft I
/*
 * printmsg.c: print a message on the console
 */
.ft CW
#include <stdio.h>
.sp .5
main(argc, argv)
	int argc;
	char *argv[];
{
	char *message;
.sp .5
	if (argc != 2) {
		fprintf(stderr, "usage: %s <message>\en", argv[0]);
		exit (1);
	}
	message = argv[1];
.sp .5
	if (!printmessage(message)) {
		fprintf(stderr,\|"%s:\|couldn't\|print\|your\|message\en",
			argv[0]);
		exit (1);
	}
	printf("Message Delivered!\en");
	exit (0);
}
.ft I
/*
 * Print a message to the console.
 * Return a boolean indicating whether the message was actually printed.
 */
.ft CW
printmessage(msg)
	char *msg;
{
	FILE *f;
.sp .5
	f = fopen("/dev/console", "w");
	if (f == NULL) {
		return (0);
	}
	fprintf(f, "%s\en", msg);
	fclose(f);
	return (1);
}
.DE

.LP
And then, of course:

.DS
.ft CW
example%  \fBcc printmsg.c -o printmsg\fP
example%  \fBprintmsg "Hello, there."\fP
Message delivered!
example%
.DE

.LP
If
.I printmessage() 
was turned into a remote procedure,
then it could be called from anywhere in the network.
Ideally, one would just like to stick a keyword like
.I remote 
in front of a
procedure to turn it into a remote procedure.  Unfortunately,
we have to live within the constraints of the C language, since
it existed long before RPC did.  But even without language
support, it's not very difficult to make a procedure remote.
.LP
In general, it's necessary to figure out what the types are for
all procedure inputs and outputs.  In this case, we have a
procedure
.I printmessage() 
which takes a string as input, and returns an integer
as output.  Knowing this, we can write a protocol specification in RPC
language that describes the remote version of
.I printmessage() .
Here it is:

.DS
.ft CW
.ft I
/*
 * msg.x: Remote message printing protocol
 */
.ft CW
.sp .5
program MESSAGEPROG {
	version MESSAGEVERS {
		int PRINTMESSAGE(string) = 1;
	} = 1;
} = 0x20000099;
.DE

.LP
Remote procedures are part of remote programs, so we actually declared
an entire remote program here which contains the single procedure
.I PRINTMESSAGE .
By convention, all RPC services provide for procedure 0.  It is normally used
for pinging purposes.
The above procedure was declared to be in version 1 of the
remote program.  No null procedure (procedure 0) is necessary in the
protocol definition because
.I rpcgen 
generates it automatically and the user is not concerned with it.
.LP
Notice that everything is declared with all capital letters.  This is
not required, but is a good convention to follow.
.LP
Notice also that the argument type is ``string'' and not ``char *''.  This
is because a ``char *'' in C is ambiguous.  Programmers usually intend it
to mean a null-terminated string of characters, but it could also
represent a pointer to a single character or a pointer to an array of
characters.  In RPC language, a null-terminated string is
unambiguously called a ``string''.
.LP
There are just two more things to write.  First, there is the remote
procedure itself.  Here's the definition of a remote procedure
to implement the
.I PRINTMESSAGE 
procedure we declared above.

.DS
.ft CW
.ft I
/*
 * msg_proc.c: implementation of the remote procedure "printmessage"
 */
.ft CW
.sp .5
#include <stdio.h>
#include <rpc/rpc.h> 	/* \fIalways needed\fP  */
#include "msg.h"	 	/* \fImsg.h will be generated by rpcgen\fP */
.sp .5
.ft I
/*
 * Remote verson of "printmessage"
 */
.ft CW
int *		
printmessage_1(msg)
	char **msg;	
{
	static int result; /* \fImust be static!\fP */
	FILE *f;
.sp .5
	f = fopen("/dev/console", "w");
	if (f == NULL) {
		result = 0;
		return (&result);
	}
	fprintf(f, "%s\en", *msg);
	fclose(f);
	result = 1;
	return (&result);
}
.DE

.LP
Notice here that the declaration of the remote procedure
.I printmessage_1() 
differs from that of the local procedure
.I printmessage() 
in three ways:
.IP  1.
It takes a pointer to a string instead of a string itself.  This
is true of all remote procedures: they always take pointers to their
arguments rather than the arguments themselves.
If there are no arguments, specify
.I void .
.IP  2.
It returns a pointer to an integer instead of an integer itself.  This is
also characteristic of remote procedures \(em they return pointers to 
their results.
Therefore it is important to have the result declared as a
.I static .
If there are no arguments, specify
.I void .
.IP  3.
It has an ``_1'' appended to its name.  In general, all remote
procedures called by
.I rpcgen 
are named by the following rule: the name in the procedure definition
(here
.I PRINTMESSAGE )
is converted to all
lower-case letters, an underbar (``_'') is appended to it, and
finally the version number (here 1) is appended.
.LP
The last thing to do is declare the main client program that will call
the remote procedure.  Here it is:

.DS
.ft CW
.ft I
/*
 * rprintmsg.c: remote version of "printmsg.c"
 */
.ft CW
#include <stdio.h>
#include <rpc/rpc.h>     /* \fIalways needed\fP  */
#include "msg.h"         /* \fImsg.h will be generated by rpcgen\fP */
.sp .5
main(argc, argv)
	int argc;
	char *argv[];
{
	CLIENT *cl;
	int *result;
	char *server;
	char *message;
.sp .5
	if (argc != 3) {
		fprintf(stderr, 
		"usage: %s host message\en", argv[0]);
		exit(1);
	}
.sp .5
	server = argv[1];
	message = argv[2];
.sp .5
.ft I
	/*
	 * Create client "handle" used for calling \fIMESSAGEPROG\fP on the
	 * server designated on the command line.  We tell the RCP package
	 * to use the "tcp" protocol when contacting the server.
	 */
.ft CW
	cl = clnt_create(server, MESSAGEPROG, MESSAGEVERS, 
	  "tcp");
	if (cl == NULL) {
.ft I
		/*
		 * Couldn't establish connection with server.
		 * Print error message and die.
		 */
.ft CW
		clnt_pcreateerror(server);
		exit(1);
	}
	
.ft I
	/*
	 * Call the remote procedure "printmessage" on the server
	 */
.ft CW
	result = printmessage_1(&message, cl);
	if (result == NULL) {
.ft I
		/*
		 * An error occurred while calling the server.
		 * Print error message and die.
		 */
.ft CW
		clnt_perror(cl, server);
		exit(1);
	}
.sp .5
.ft I
	/*
	 * Okay, we successfully called the remote procedure.
	 */
.ft CW
	if (*result == 0) {
.ft I
		/*
		 * Server was unable to print our message.
		 * Print error message and die.
		 */
.ft CW
	fprintf(stderr, "%s: %s couldn't print your message\en", 
		argv[0], server);    
		exit(1);
	}
.sp .5
.ft I
	/*
	 * The message got printed on the server's console
	 */
.ft CW
	printf("Message delivered to %s!\en", server);
	exit(0);
}
.DE

There are a few points worth noting here:
.IP  1.
..IX "client handle, used by rpcgen" "" "client handle, used by \fIrpcgen\fP"
First a client ``handle'' is created using the RPC library routine
.I clnt_create() .
This client handle will be passed to the skeleton routines
which call the remote procedure.
.IP  2.
The last parameter to
.I clnt_create 
is ``tcp'', the transport on which you want your application to run on.
It could also have been ``udp'', as an alternate transport.
For more information on transport selection see the section
\fITransport Selection\fP in the chapter
\fIIntroduction to Remote Procedure Calls\fP.  
.IP  3.
The remote procedure
.I printmessage_1() 
is called exactly the same way as it is declared in
.I msg_proc.c 
except for the inserted client handle as the second argument.
.IP  4.
The remote procedure call can fail in two ways.  The RPC mechanism itself
can fail or, alternatively, there can be an error in the execution of the
actual remote procedure.  In the former case, the remote procedure (in this
case
.I print_message_1() )
returns with a
.I NULL .
In the later case, however, the details of error reporting are application 
dependent.  Here, the error is being reported via
.I "*result" .
.LP
Here's how to put all of the pieces together:

.DS
.ft CW
example%  \fBrpcgen msg.x\fP
example%  \fBcc rprintmsg.c msg_clnt.c -o rprintmsg\fP
example%  \fBcc msg_proc.c msg_svc.c -o msg_server\fP
.DE

Two programs were compiled here: the client program
.I rprintmsg 
and the server program
.I msg_server .
Before doing this though,
.I rpcgen 
was used to fill in the missing pieces.
.LP
Here is what
.I rpcgen 
(called without any flags) did with the input file
.I msg.x :
.IP  1.
It created a header file called
.I msg.h 
that contained
.I #define 's
for
.I MESSAGEPROG ,
.I MESSAGEVERS 
and
.I PRINTMESSAGE 
for use in the other modules.  This file should be included by 
.I both
the client and the server modules.
.IP  2.
It created the client ``skeleton'' routines in the
.I msg_clnt.c 
file.  In this case there is only one, the
.I printmessage_1() 
that was referred from the
.I printmsg 
client program.  If the name of the input file is 
.I FOO.x ,
the client skeletons output file is called
.I FOO_clnt.c .
.IP  3.
It created the server program in 
.I msg_svc.c 
which calls
.I printmessage_1() 
from
.I msg_proc.c .
The rule for naming the server output file is similar to the
previous one: for an input file called
.I FOO.x ,
the output server file is named
.I FOO_svc.c .
.LP
(Note that, given the
.I -T 
argument,
.I rpcgen 
creates an additional output file which contains index information used 
for the dispatching of service routines).
.LP
Now we're ready to have some fun.  First, copy the server to a remote 
machine and run it.  For this example, the machine is called ``moon''.  
.LP
Note that servers generated by
.I rpcgen 
can be invoked with portmonitors like
.I inetd 
as well as from the command line, if they are invoked with the
.I -I 
option.
.LP

.DS
.ft CW
moon% \fBmsg_server &\fP
.DE

Then on our local machine (``sun'') we can print a message on ``moon''s
console.

.DS
.ft CW
sun% \fBrprintmsg moon "Hello, moon."\fP
.DE

The message will get printed on ``moon''s console.  You can print a
message on anybody's console (including your own) with this program if
you can copy the server to their machine and run it.
.NH 3
\&An Advanced Example
..IX RPC "generating XDR routines"
..IX RPC "an advanced example"
.LP
The previous example only demonstrated the automatic generation of
client and server RPC code.
.I rpcgen 
may also be used to generate XDR routines, that is, the routines
necessary to convert local data
structures into network format and vice-versa.  This next example is
more advanced in that it presents
a complete RPC service\(ema remote directory listing service, which uses
.I rpcgen 
not only to generate skeleton routines, but also to generate the XDR
routines.  Here is the protocol description file.

.DS
.ft CW
.ft I
/*
 * dir.x: Remote directory listing protocol
 */
.ft CW
const MAXNAMELEN = 255;		/* \fImaximum length of a directory entry\fP */
.sp .5
typedef string nametype<MAXNAMELEN>;	/* \fIa directory entry\fP */
.sp .5
typedef struct namenode *namelist;		/* \fIa link in the listing\fP */
.sp .5
.ft I
/*
 * A node in the directory listing
 */
.ft CW
struct namenode {
	nametype name;		/* \fIname of directory entry\fP */
	namelist next;		/* \fInext entry\fP */
};
.sp .5
.ft I
/*
 * The result of a READDIR operation.
 */
.ft CW
union readdir_res switch (int errno) {
case 0:
	namelist list;	/* \fIno error: return directory listing\fP */
default:
	void;		/* \fIerror occurred: nothing else to return\fP */
};
.sp .5
.ft I
/*
 * The directory program definition
 */
.ft CW
program DIRPROG {
	version DIRVERS {
		readdir_res
		READDIR(nametype) = 1;
	} = 1;
} = 0x20000076;
.DE

.SH
NOTE
.LP
Types (like
.I readdir_res 
in the example above) can be defined using
the ``struct'', ``union'' and ``enum'' keywords, but those keywords
should not be used in subsequent declarations of variables of those types.
For example, if you define a union ``foo'', you should declare using
only ``foo'' and not ``union foo''.  In fact,
.I rpcgen 
compiles
RPC unions into C structures and it is an error to declare them using the
``union'' keyword.
.LP
.sp
Running
.I rpcgen 
on
.I dir.x 
creates four output files.  First are the basic three itemized above:
those containing the header file, client skeleton routines and server
skeleton.  The fourth contains the XDR routines necessary for converting
the data types we declared into XDR format and vice-versa.  These are
output in the file
.I dir_xdr.c .
For each data type used in the
.I .x 
file, 
.I rpcgen 
assumes that the RPC/XDR library has a routine defined with the name of that
data type prepended by
.I xdr_ 
(e.g.
.I xdr_int ).
If the data type was defined in the
.I .x 
file, then
.I rpcgen 
will generate the required xdr routine.  If there are no such data types,
then the file (e.g.
.I dir_xdr.c )
will not be generated.  If the data types were used but not defined, then
the user has to provide that xdr routine.  This is a way for users
to provide their own customized xdr routines.
.LP
Here is the implementation of the
.I READDIR 
procedure.

.DS
.ft CW
.ft I
/*
 * dir_proc.c: remote readdir implementation
 */
.ft CW
#include <rpc/rpc.h>		/* \fIAlways needed\fP */
#include <sys/dir.h>
#include "dir.h"			/* \fICreated by\fP rpcgen */
.sp .5
extern int errno;
extern char *malloc();
extern char *strdup();
.sp .5
readdir_res *
readdir_1(dirname)
	nametype *dirname;
{
	DIR *dirp;
	struct direct *d;
	namelist nl;
	namelist *nlp;
	static readdir_res res; /* \fImust be static\fP! */
	
.ft I
	/*
	 * Open directory
	 */
.ft CW
	dirp = opendir(*dirname);
	if (dirp == NULL) {
		res.errno = errno;
		return (&res);
	}
.sp .5
.ft I
	/*
	 * Free previous result
	 */
.ft CW
	xdr_free(xdr_readdir_res, &res);
.sp .5
.ft I
	/*
	 * Collect directory entries.
	 * Memory allocated here will be freed by \fIxdr_free\fP
	 * next time \fIreaddir_1\fP is called
	 */
.ft CW
	nlp = &res.readdir_res_u.list;
	while (d = readdir(dirp)) {
		nl = *nlp = (namenode *) malloc(sizeof(namenode));
		nl->name = strdup(d->d_name);
		nlp = &nl->next;
	}
	*nlp = NULL;
.sp .5
.ft I
	/*
	 * Return the result
	 */
.ft CW
	res.errno = 0;
	closedir(dirp);
	return (&res);
}
.DE

.KS
.LP
Finally, there is the client side program to call the server:

.DS
.ft CW
.ft I
/*
 * rls.c: Remote directory listing client
 */
.ft CW
#include <stdio.h>
#include <rpc/rpc.h>	/* \fIalways need this\fP */
#include "dir.h"		/* \fIwill be generated by rpcgen\fI */
.sp .5
extern int errno;
.sp .5
main(argc, argv)
	int argc;
	char *argv[];
{
	CLIENT *cl;
	char *server;
	char *dir;
	readdir_res *result;
	namelist nl;
.sp .5
	if (argc != 3) {
		fprintf(stderr, "usage: %s host directory\en", 
		  argv[0]);
		exit(1);
	}
.sp .5
	server = argv[1];
	dir = argv[2];
.sp .5
.ft I
	/*
	 * Create client "handle" used for calling \fIDIRPROG\fP on the
	 * server designated on the command line.  Use the tcp protocol when
	 * contacting the server.
	 */
.ft CW
	cl = clnt_create(server, DIRPROG, DIRVERS, "tcp");
	if (cl == NULL) {
.ft I
		/*
		 * Couldn't establish connection with server.
		 * Print error message and die.
		 */
.ft CW
		clnt_pcreateerror(server);
		exit(1);
	}
	
.ft I
	/*
	 * Call the remote procedure \fIreaddir\fP on the server
	 */
.ft CW
	result = readdir_1(&dir, cl);
	if (result == NULL) {
.ft I
		/*
		 * An RPC error occurred while calling the server.
	 	 * Print error message and die.
		 */
.ft CW
		clnt_perror(cl, server);
		exit(1);
	}
.sp .5
.ft I
	/*
	 * Okay, we successfully called the remote procedure.
	 */
.ft CW
	if (result->errno != 0) {
.ft I
		/*
		 * A remote system error occurred.
		 * Print error message and die.
		 */
.ft CW
		errno = result->errno;
		perror(dir);
		exit(1);
	}
.sp .5
.ft I
	/*
	 * Successfully got a directory listing.
	 * Print it out.
	 */
.ft CW
	for (nl = result->readdir_res_u.list; nl != NULL;
	  nl = nl->next) {
		printf("%s\en", nl->name);
	}
	exit(0);
}
.DE
.KE

Compile everything, and run.

.DS
.ft CW
sun%  \fBrpcgen dir.x\fP
sun%  \fBcc -c dir_xdr.c
sun%  \fBcc rls.c dir_clnt.c dir_xdr.o -o rls\fP
sun%  \fBcc dir_svc.c dir_proc.c dir_xdr.o -o dir_svc\fP
.sp .5
sun%  \fBdir_svc &\fP
.sp .5
moon%  \fBrls sun /usr/pub\fP
\&.
\&..
ascii
eqnchar
greek
kbd
marg8
tabclr
tabs
tabs4
moon%
.DE

.LP
.I rpcgen 
generated client code does not release the memory allocated for the results
of the RPC call.  Users can call
.I xdr_free 
to free up the memory once they are done with it. It is quite similar to calling
.I free() 
except that here one also has to pass the xdr routine for the result.
In this example, after printing the list, the user could have called
.DS
.ft CW
xdr_free(xdr_readdir_res, result);
.DE
..IX "debugging with rpcgen" "" "debugging with \fIrpcgen\fP"
.NH 3
\&Debugging Applications
.LP
It is often difficult to debug distributed applications like these
because the client and the server are two different processes.
To simplify the testing and debugging process,
the client program and the server procedure can be tested together
as a single program by simply linking them with each other rather
than with the client and server skeletons.
This could be done in the previous example by doing:
.DS
.ft CW
cc rls.c dir_clnt.c dir_proc.c dir_xdr.c -o rls
.DE
The procedure calls will be
executed as ordinary local procedure calls and the program can be
debugged with a local debugger such as
.I dbxtool .
When the program is working, the client program can be linked to
the client skeleton produced by
.I rpcgen 
and the server procedures can be linked to the server skeleton produced
by
.I rpcgen .
.SH
NOTE
.LP
If you do this, you will have to comment out calls to client create RPC library
routines (e.g. 
.I clnt_create() ).
.LP
.LP
There are two kinds of errors which can happen in an RPC call.
The first kind of error is caused if there is some problem with the actual
mechanism of the remote procedure calls.  This could happen in such cases
as the procedure is not available, the remote server is not responding, the
remote server is unable to decode the arguments, and so on.
In the previous example, an RPC error has occurred if 
.I result 
is
.I NULL .
The reason for the failure can be printed by using
.I clnt_perror() ,
or an error string can be returned through
.I clnt_sperror() .
.LP
The second type of error is due to the server itself.  In the previous
example, an error was reported if
.I opendir() 
fails.  Now you can see why
.I readdir_res 
is of type
.I union .
The handling of these types of errors are the responsibility of the programmer.
.NH 3
\&The C-Preprocessor
..IX rpcgen "C-preprocessor" \fIrpcgen\fP
.LP
The C-preprocessor, 
.I cpp ,
is run on all input files before they are compiled, so all the 
preprocessor directives are legal within a ``.x'' file.  
Five macro identifiers 
may have been defined, depending upon which output file is getting 
generated.  They are:
.TS
box tab (&);
lfI lfI
l l .
Identifier&Usage
_
RPC_HDR&For header-file output
RPC_XDR&For XDR routine output	
RPC_SVC&For server-skeleton output
RPC_CLNT&For client skeleton output
RPC_TBL&For index table output
.TE	
.LP
Also,
.I rpcgen 
does some additional preprocessing of the input file.  Any line that
begins with a percent sign is passed directly into the output file,
without any interpretation of the line.  Here is a simple example that
demonstrates this processing feature.

.DS
.ft CW
.ft I
/*
 * time.x: Remote time protocol
 */
.ft CW
program TIMEPROG {
		version TIMEVERS {
				unsigned int TIMEGET(void) = 1;
		} = 1;
} = 44;
.sp .5
#ifdef RPC_SVC
%int *
%timeget_1()
%{
%		static int thetime;
%
%		thetime = time(0);
%		return (&thetime);
%}
#endif
.DE

When using the '%' feature, there is no guarantee that 
.I rpcgen 
will place the output where you intended.  If you have problems of this
type, we recommend you to not use this feature.
.NH 3
\&\fBrpcgen\fP Programming Notes
..IX rpcgen "programming notes" \fIrpcgen\fP
.sp 2
.NH 4
\&Network Types
..IX rpcgen "network types" \fIrpcgen\fP
.LP
By default
.I rpcgen 
generates server code for both UDP & TCP transports.  The
.I -s 
flag creates a server which responds to requests on the specified transport.
The following example creates a udp server:
.DS
.ft CW
example#  \fBrpcgen -s udp_n proto.x\fP
.DE
.sp 2
.NH 4
\&User-Provided Define Statements
..IX "define statements" "" "\fIdefine\fP statements"
.LP
.I rpcgen 
also provides a means of defining symbols and assigning values to 
them.  These defined symbols are passed on to the C preprocessor
when it is invoked.  This facility is useful when the user wants 
to, for example, invoke debugging code which is enabled only when
the 
.I DEBUG 
symbol is defined.  For example:

.DS
.ft CW
example%  \fBrpcgen -DDEBUG proto.x\fP
.DE

.sp 2
.NH 4
\&Inetd Support
..IX "rpcgen Inetd support" "" "\fIrpcgen\fP Inetd support"
.LP
.I rpcgen 
can also be used to create RPC servers which can be invoked by
.I inetd 
when a request for that service comes in.

.DS
.ft CW
example%  \fBrpcgen -I proto.x\fP
.DE

.LP
The server code in
.I proto_svc.c 
has the required support for
.I inetd .
For more information on how to setup the entry for RPC services in
.I /etc/inetd.conf ,
please see the section \fIUsing Inetd\fP in the chapter
\fIRemote Procedure Call Programming Guide\fP.  
.LP
In many applications, it is useful
for services to wait after satisfying a servicing request, on the
chance that another will follow.  However, if there is no call within the
specified time, the server will exit and the portmonitor will continue to
monitor requests for its services.  By default, services wait of 120
seconds after servicing a request before exiting.  The user can, however,
change that interval with the \fI-K\fP flag.

.DS
.ft CW
example%  \fBrpcgen -I -K 20 proto.x\fP
.DE

Here the server will wait only for 20 seconds before exiting.  If you want
the server to exit immediately, \fI-K 0\fP can be used, while if the server
is intended to stay around forever (a normal server) the appropriate
argument is \fI-K -1\fP.
.NH 4
\&Dispatch Tables
..IX rpcgen "dispatch tables" \fIrpcgen\fP
.LP
There are a number of cases when dispatch tables are useful.  
For example, the server dispatch routine may need to check
authorization and then invoke the service routine;
or a client library may want to deal with the details of
storage management and XDR data conversion.

.DS
.ft CW
example%  \fBrpcgen -T proto.x\fP
.DE

.LP
Here
.I rpcgen 
generates RPC dispatch tables for each program defined in the protocol
description file,
.I proto.x ,
in the file 
.I proto_tbl.i .
(The suffix \fI.i\fP stands for ``index'').
See below for how to use this file when compiling programs.  Each entry 
in the table is a 
.I "struct rpcgen_table"
defined in the header file 
.I proto.h 
as follows:

.DS
.ft CW
struct rpcgen_table {
    char        *(*proc)();
    xdrproc_t   xdr_arg;
    unsigned    len_arg;
    xdrproc_t   xdr_res;
    unsigned    len_res;
};
.DE

where
.IP 
\fIproc\fP is a pointer to the service routine,
.IP 
\fIxdr_arg\fP is a pointer to the input (argument) xdr_routine,
.IP 
\fIlen_arg\fP is the length in bytes of the input argument,
.IP 
\fIxdr_res\fP is a pointer to the output (result) xdr_routine, and
.IP 
\fIlen_res\fP is the length in bytes of the output result.
.LP
The table, named
.I dirprog_1_table ,
is indexed by procedure number.  The variable 
.I dirprog_1_nproc 
contains the number of entries in the table.
.LP
An example of how to locate an procedure in the dispatch 
tables is demonstrated by the routine 
.I find_proc :

.DS
.ft CW
struct rpcgen_table *
find_proc(proc)
	long	proc;
{
	if (proc >= dirprog_1_nproc) 
		/* \fIerror\fP */
	else
		return (&dirprog_1_table[proc]);
}
.DE

.LP
Each entry in the dispatch table contains a pointer to the corresponding
service routine.  However, the service routine is not defined
in the client code.  To avoid generating unresolved external references,
and to require only one source file for the dispatch table, the actual 
service routine initializer is 
.I "RPCGEN_ACTION(proc_ver)" .
.LP
This way, the same dispatch table can be included in both the client and 
the server.  Use the following define when compiling the client:

.DS
.ft CW
#define RPCGEN_ACTION(routine)	0
.DE

and use this define when compiling the server:

.DS
.ft CW
#define RPCGEN_ACTION(routine)	routine
.DE

.NH 3
\&Client Programming Notes
..IX rpcgen "other operations" \fIrpcgen\fP
..IX rpcgen "client programming" \fIrpcgen\fP
.sp 2
.NH 4
\&Timeout Changes
..IX rpcgen "timeout changes" \fIrpcgen\fP
.LP
RPC sets a default timeout of 25 seconds for RPC calls when
.I clnt_create() 
is used.
This means RPC will wait for 25 seconds to get the results from the server.
If it does not hear within that time period, then perhaps the server isn't
running or the remote machine crashed or the network is unreachable.
There are many possibilities of why no answer is heard.  In such cases
the function will return NULL and the error can be printed using
.I clnt_perrno() .
.LP
There are cases when the user wants to change the timeout value to accommodate 
the application needs or the fact that the server is slow and quite far away.
The timeout can be changed using
.I clnt_control() .
Here is a small code fragment to demonstrate use of
.I clnt_control() :

.DS
.ft CW
struct timeval tv;
CLIENT *cl;
.sp .5
cl = clnt_create("somehost", SOMEPROG, SOMEVERS, "tcp");
if (cl == NULL) {
	exit(1);
}
tv.tv_sec = 60;	/* \fIchange timeout to 1 minute\fP */
tv.tv_usec = 0; /* \fIthis should always be set\fP */
clnt_control(cl, CLSET_TIMEOUT, &tv);	
.DE

.NH 4
\&Client Authentication
..IX rpcgen "client authentication" \fIrpcgen\fP
.LP
The client create routines do not, by default, have any facilities
for client authentication, but the client may sometimes want to
authenticate itself to the server.
For more information on how to perform authentication, see the
\fIAuthentication\fP section of the chapter
\fIRemote Procedure Call Programming Guide\fP.  
Doing so is trivial, and looks like this:

.DS
.ft CW
CLIENT *cl;
.sp .5
cl = client_create("somehost", SOMEPROG, SOMEVERS, "udp");
if (cl != NULL) {
	/* \fITo set UNIX style authentication\fP */
	cl->cl_auth = authunix_create_default();
}
.DE

.NH 3
\&Server Programming Notes
..IX rpcgen "server programming" \fIrpcgen\fP
.sp 2
.NH 4
\&Handling Broadcast on the Server Side
..IX "broadcast RPC"
..IX rpcgen "broadcast RPC" \fIrpcgen\fP
.LP
Clients may sometimes broadcast to find out whether a particular server
exists on the network or just to find out about all the servers for a
particular program and version number.  These calls are made via
.I clnt_broadcast() .
Note that there is no rpcgen support for that.  Please see
the section \fIBroadcast RPC Synopsis\fP in the chapter
\fIRemote Procedure Call Programming Guide\fP.  
.LP
When a procedure is known to be called via broadcast RPC,
it is usually wise for the server to not reply unless it can provide
some useful information to the client.  This prevents the network
from getting flooded by useless replies.
.LP
To prevent the server from replying, a remote procedure can
return NULL as its result, and the server code generated by
.I rpcgen 
will detect this and not send out a reply.
.LP
Here is an example of a procedure that replies only if it
thinks it is an NFS server:

.DS
.ft CW
void *
reply_if_nfsserver()
{
	char notnull;	/* \fIjust here so we can use its address\fP */
.sp .5
	if (access("/etc/exports", F_OK) < 0) {
		return (NULL);	/* \fIprevent RPC from replying\fP */
	}
.ft I
	/*
	 * return non-null pointer so RPC will send out a reply
	 */
.ft CW
	return ((void *)&notnull);
}
.DE

Note that if procedure returns type ``void *'', they must return a non-NULL
pointer if they want RPC to reply for them.
.NH 4
\&Other Information Passed to Server Procedures
.LP
Server procedures will often want to know more about an RPC call
than just its arguments.  For example, getting authentication information
is important to procedures that want to implement some level of security.
This extra information is actually supplied to the server procedure as a
second argument.  (For details see the structure of 
.I svc_req ,
in the
\fIAuthentication\fP section of the chapter
\fIRemote Procedure Call Programming Guide\fP.  
Here is an example to demonstrate its use.  What we've
done here is rewrite the previous
.I printmessage_1() 
procedure to only allow root users to print a message to the console.

.DS
.ft CW
int *
printmessage_1(msg, rqstp)
	char **msg;
	struct svc_req	*rqstp;
{
	static int result;	/* \fIMust be static\fP */
	FILE *f;
	struct authunix_parms *aup;
.sp .5
	aup = (struct authunix_parms *)rqstp->rq_clntcred;
	if (aup->aup_uid != 0) {
		result = 0;
		return (&result);
	}
.sp
.ft I
	/*
	 * Same code as before.
	 */
.ft CW
}
.DE

.NH 3
\&RPC Language
..IX RPCL
..IX rpcgen "RPC Language" \fIrpcgen\fP
.LP
RPC language is an extension of XDR language.  The sole extension is
the addition of the
.I program 
and 
.I version 
types.  For a complete description of the XDR language syntax, see the
\fIExternal Data Representation Standard:  Protocol Specification\fP 
chapter.  For a description of the RPC extensions to the XDR language,
see the
\fIRemote Procedure Calls:  Protocol Specification\fP
chapter.
.LP
However, XDR language is so close to C that if you know C, you know most
of it already.  We describe here the syntax of the RPC language,
showing a few examples along the way.  We also show how the various
RPC and XDR type definitions get compiled into C type definitions in
the output header file.
.NH 4
\&Definitions
..IX rpcgen definitions \fIrpcgen\fP
.LP
An RPC language file consists of a series of definitions.
.DS
.ft CW
	definition-list:
		definition ";"
		definition ";" definition-list
.DE
It recognizes the following types of definitions.
.DS
.ft CW
	definition:
		enum-definition
		typedef-definition
		const-definition
		declaration-definition
		struct-definition
		union-definition
		program-definition
.DE
.NH 4
\&Enumerations
..IX rpcgen enumerations \fIrpcgen\fP
.LP
XDR enumerations have the same syntax as C enumerations.
.DS
.ft CW
	enum-definition:
		"enum" enum-ident "{"
			enum-value-list
		"}"
.sp .5
	enum-value-list:
		enum-value
		enum-value "," enum-value-list
.sp .5
	enum-value:
		enum-value-ident
		enum-value-ident "=" value
.DE
Here is a short example of an XDR enum, and the C enum that it gets
compiled into.
.DS
.ft CW
     enum colortype {      enum colortype {
          RED = 0,              RED = 0,
          GREEN = 1,   -->      GREEN = 1,
          BLUE = 2              BLUE = 2,
     };                    };
                           typedef enum colortype colortype;
.DE
.NH 4
\&Typedefs
..IX rpcgen typedef \fIrpcgen\fP
.LP
XDR typedefs have the same syntax as C typedefs.
.DS
.ft CW
	typedef-definition:
		"typedef" declaration
.DE
Here is an example that defines a
.I fname_type 
used for declaring
file name strings that have a maximum length of 255 characters.
.DS
.ft CW
typedef string fname_type<255>; --> typedef char *fname_type;
.DE
.NH 4
\&Constants
..IX rpcgen constants \fIrpcgen\fP
.LP
XDR constants may be used wherever a
integer constant is used, for example, in array size specifications.
.DS
.ft CW
	const-definition:
		"const" const-ident "=" integer
.DE
For example, the following defines a constant
.I DOZEN 
equal to 12.
.DS
.ft CW
	const DOZEN = 12;  -->  #define DOZEN 12
.DE
.NH 4
\&Declarations
..IX rpcgen declarations \fIrpcgen\fP
.LP
In XDR, there are only four kinds of declarations.
.DS
.ft CW
	declaration:
		simple-declaration
		fixed-array-declaration
		variable-array-declaration
		pointer-declaration
.DE
\fB1) Simple declarations\fP are just like simple C declarations.
.DS
.ft CW
	simple-declaration:
		type-ident variable-ident
.DE
Example:
.DS
.ft CW
	colortype color;	--> colortype color;
.DE
\fB2) Fixed-length Array Declarations\fP are just like C array declarations:
.DS
.ft CW
	fixed-array-declaration:
		type-ident variable-ident "[" value "]"
.DE
Example:
.DS
.ft CW
	colortype palette[8];	--> colortype palette[8];
.DE
\fB3) Variable-Length Array Declarations\fP have no explicit syntax
in C, so XDR invents its own using angle-brackets.
.DS
.ft CW
variable-array-declaration:
	type-ident variable-ident "<" value ">"
	type-ident variable-ident "<" ">"
.DE
The maximum size is specified between the angle brackets.  The size may
be omitted, indicating that the array may be of any size.
.DS
.ft CW
	int heights<12>;	/* \fIat most 12 items\fP */
	int widths<>;		/* \fIany number of items\fP */
.DE
Since variable-length arrays have no explicit syntax in C, these
declarations are actually compiled into ``struct''s.  For example, the
``heights'' declaration gets compiled into the following struct:
.DS
.ft CW
	struct {
		u_int heights_len;	/* \fI# of items in array\fP */
		int *heights_val;	/* \fIpointer to array\fP */
	} heights;
.DE
Note that the number of items in the array is stored in the ``_len''
component and the pointer to the array is stored in the ``_val''
component.  The first part of each of these component's names is the
same as the name of the declared XDR variable.
.LP
\fB4) Pointer Declarations\fP are made in
XDR exactly as they are in C.  You can't
really send pointers over the network, but you can use XDR pointers
for sending recursive data types such as lists and trees.  The type is
actually called ``optional-data'', not ``pointer'', in XDR language.
.DS
.ft CW
pointer-declaration:
	type-ident "*" variable-ident
.DE
Example:
.DS
.ft CW
	listitem *next;  -->  listitem *next;
.DE
.NH 4
\&Structures
..IX rpcgen structures \fIrpcgen\fP
.LP
An XDR struct is declared almost exactly like its C counterpart.  It
looks like the following:
.DS
.ft CW
	struct-definition:
		"struct" struct-ident "{"
			declaration-list
		"}"
.sp .5
	declaration-list:
		declaration ";"
		declaration ";" declaration-list
.DE
As an example, here is an XDR structure to a two-dimensional
coordinate, and the C structure that it gets compiled into in the
output header file.
.DS
.ft CW
   struct coord {             struct coord {
        int x;       -->           int x;
        int y;                     int y;
   };                         };
                              typedef struct coord coord;
.DE
The output is identical to the input, except for the added
.I typedef 
at the end of the output.  This allows one to use ``coord'' instead of
``struct coord'' in declarations.
.NH 4
\&Unions
..IX rpcgen unions \fIrpcgen\fP
.LP
XDR unions are discriminated unions, and look quite different from C
unions.  They are more analogous to Pascal variant records than they
are to C unions.
.DS
.ft CW
	union-definition:
		"union"\|union-ident\|"switch"\|"("\|simple\|declaration\|")"\|"{"
			case-list
		"}"
.sp .5
	case-list:
		"case" value ":" declaration ";"
		"case" value ":" declaration ";" case-list
		"default" ":" declaration ";"
.DE
Here is an example of a type that might be returned as the result of a
``read data'' operation.  If there is no error, return a block of data.
Otherwise, don't return anything.
.DS
.ft CW
	union read_result switch (int errno) {
	case 0:
		opaque data[1024];
	default:
		void;
	};
.DE
It gets compiled into the following:
.DS
.ft CW
	struct read_result {
		int errno;
		union {
			char data[1024];
		} read_result_u;
	};
	typedef struct read_result read_result;
.DE
Notice that the union component of the output struct has the same name as
the structure type name, except for the trailing ``_u''.
.NH 4
\&Programs
..IX rpcgen programs \fIrpcgen\fP
.LP
RPC programs are declared using the following syntax:
.DS
.ft CW
program-definition:
	"program" program-ident "{"
		version-list
	"}" "=" value
.sp .5
version-list:
	version ";"
	version ";" version-list
.sp .5
version:
	"version" version-ident "{"
		procedure-list
	"}" "=" value
.sp .5
procedure-list:
	procedure ";"
	procedure ";" procedure-list
.sp .5
procedure:
	type-ident procedure-ident "(" type-ident ")" "=" value
.DE
For example, here is the time protocol, revisited:
.DS
.ft CW
.ft I
/*
 * time.x: Get or set the time.  Time is represented as number of seconds
 * since 0:00, January 1, 1970.
 */
.ft CW
program TIMEPROG {
	version TIMEVERS {
		unsigned int TIMEGET(void) = 1;
		void TIMESET(unsigned) = 2;
	} = 1;
} = 44;		
.DE
This file compiles into these #defines in the output header file:
.DS
.ft CW
#define TIMEPROG 44
#define TIMEVERS 1
#define TIMEGET 1
#define TIMESET 2
.DE
.NH 4
\&Special Cases
..IX rpcgen "special cases" \fIrpcgen\fP
.LP
There are a few exceptions to the rules described above.
.LP
.B Booleans:
C has no built-in boolean type.  However, the RPC library has a
boolean type called
.I bool_t 
that is either
.I TRUE 
or
.I FALSE .
Things declared as type
.I bool 
in XDR language are compiled into
.I bool_t 
in the output header file.
.LP
Example:
.DS
.ft CW
	bool married;  -->  bool_t married;
.DE
.B Strings:
C has no built-in string type, but instead uses the null-terminated
``char *'' convention.  In XDR language, strings are declared using the
``string'' keyword, and compiled into ``char *''s in the output header
file.  The maximum size contained in the angle brackets specifies the
maximum number of characters allowed in the strings (not counting the
.I NULL 
character).  The maximum size may be left off, indicating a string
of arbitrary length.
.LP
Examples:
.DS
.ft CW
	string name<32>;  -->  char *name;	
	string longname<>;  -->  char *longname;	
.DE
.B "Opaque Data:"
Opaque data is used in RPC and XDR to describe untyped data, that is,
just sequences of arbitrary bytes.  It may be declared either as a
fixed or variable length array.
.DS
.ft CW
Examples:
	opaque diskblock[512];  -->  char diskblock[512];
.sp .5
	opaque filedata<1024>;  -->  struct {
									u_int filedata_len;
									char *filedata_val;
								} filedata;
.DE
.B Voids:
In a void declaration, the variable is not named.  The declaration is
just ``void'' and nothing else.  Void declarations can only occur in two
places: union definitions and program definitions (as the argument or
result of a remote procedure).
