/* @(#)Pod.cc 1.2 91/05/15 SMI */

/*
 * Copyright (c) 1988, Sun Microsystems, Inc.  All Rights Reserved. Sun
 * considers its source code as an unpublished, proprietary trade secret, and
 * it is available only under strict license provisions.  This copyright
 * notice is placed here only to protect Sun in the event the source is
 * deemed a published work.  Dissassembly, decompilation, or other means of
 * reducing the object code to human readable form is prohibited by the
 * license agreement under which this code is provided to the user or company
 * in possession of this copy.
 * 
 * RESTRICTED RIGHTS LEGEND: Use, duplication, or disclosure by the Government
 * is subject to restrictions as set forth in subparagraph (c)(1)(ii) of the
 * Rights in Technical Data and Computer Software clause at DFARS 52.227-7013
 * and in similar clauses in the FAR and NASA FAR Supplement.
 * 
 * (C) Copyright 1987 Sun Microsystems, Inc. All rights reserved. This material
 * contains trade secret information of Sun Microsystems, Inc. Use,
 * reproduction, or disclosure is prohibited, except under a written license
 * by Sun Microsystems, Inc. Use of copyright notice is precautionary only
 * and does not imply publication or disclosure.
 */

/*
 * This module (class) contains the code and data for the Pod. The Pod (there
 * is only one) is at the root of the shared/global memory data hierarchy.
 * 
 * It contains information on the associated semaphore, shared memory, and all
 * of the Threads.
 */

/*
 * FUNCTIONS IN THIS MODULE:
 *
 * addr_Pod			- get the address of the shared pod structure.
 *				  if none is found, create it or release the
 *				  connection to the pod
 * lock_Pod			- lock or unlock the Pod structure, if semaphore
 *				  is not available, create it
 */

#define Pod_c
#include <stdio.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <rpc/rpc.h>
#include <errno.h>
#include "config.h"
#include "TEP.h"
#include "Thread.h"
#include "Pod.h"
#include <xti.h>


/*
 * Function Name: addr_Pod
 * Description:  get the address of the pod.  Find the pod and attach it
 *		 if it is not attached.  Create a pod if none exist.
 *		 OR release the pod.
 * Return Values: address of the pod structure.
 *		  if there is a system error in allocating, attaching, etc.,
 *		  return a NULL pointer.
 */

Pod *
addr_Pod(op)
int op;
{
    static Pod *pod = (Pod *)NULL;
    static int shm_id = -1;	  	   /* shared memory id */
    static key_t shm_key = 0x5854492E;     /* key is "XTI." */
    short new_pod = 0;
    int inx;

    switch(op)
    {
    case GET_ADDR:
	if(pod == (Pod *)NULL)
	{
	    if ((shm_id = shmget(shm_key, sizeof(Pod), 0666 )) < 0)
	    {
		/* not found, create it */
		if(lock_Pod(LOCK_POD) < 0)
		{
		    return (Pod *)NULL;
		}
		if((shm_id = shmget(shm_key, sizeof(Pod), 0666 | IPC_CREAT)) < 0)
		{
		    int local_errno = errno;

		    lock_Pod(UNLOCK_POD);
		    errno = local_errno;
		    return (Pod *)NULL;
		}
		++new_pod;
	    }
	    if(shm_id >= 0)
	    {
		/* map it in */
		if((pod = (Pod *) shmat(shm_id, (char *)0, 0)) < (Pod *)0)
		{
		    return (Pod *)NULL;
		}
	    }
	    else
	    {
		return (Pod *)NULL;
	    }

	    /* if a new pod, initialize it then unlock it */
	    if(new_pod)
	    {
		--new_pod;
		for(inx=0; inx < XTI_MAX_THREADS; inx++)
		{
		    pod->threads[inx].active = 0;
		}
		if(lock_Pod(UNLOCK_POD) < 0)
		{
		    return (Pod *)NULL;
		}
	    }
	}

	return pod;
	break;
    
    case FREE_ADDR:
	if(pod != (Pod *)NULL)
	{
	    if(shmdt(pod) == 0)
	    {
		Pod *pod_ptr = pod;

		pod = (Pod *)NULL;
		shm_id = -1;
		return pod_ptr;
	    }
	}
	return (Pod *)NULL;
	break;
    }


}


/*
 * Function Name: lock_Pod
 * Description:   Lock or unlock the pod.  If, when locking, there is no 
 *		  semaphore, create one.
 *		  The pod is locked for the following reasons:
 *			1. The pod is being created.
 *			2. A thread is being cloned into the pod for a
 *			   child that is about to be spawned.
 * Return Value: 0 for a successful lock, -1 on error
 */
lock_Pod(op)
char op;
{
    static lock_status = UNLOCKED;
    static mask = 0;
    static key_t sem_key = 0x5854492E;
    static sem_id = -1;
    static struct sembuf op_lock[2] = {
        0, 0, 0,
        0, 1, SEM_UNDO
    };
    static struct sembuf op_unlock[1] = {
        0, -1, SEM_UNDO
    };

    switch(op)
    {
    case LOCK_POD:	/* lock the pod */
	    if(lock_status != LOCKED)
	    {
		if(sem_id == -1)
		{
		    if((sem_id = semget(sem_key, 1, 0666)) < 0)
		    {
			/* there is no semaphore, so create one */
			if((sem_id = semget(sem_key, 1, 0666 | IPC_CREAT)) < 0)
			{
			    return -1;
			}
		    }
		}
		mask = sigblock(0xffffffff);
		/*
		 * Process is blocked until shared/global memory is free, then blocks
		 * other processes.
		 */
		if(semop(sem_id, &op_lock[0], 2) < 0)
		{
		    return -1;
		}
		lock_status = LOCKED;
	    }
	    break;

    case UNLOCK_POD:	/* unlock the pod */
	    if(lock_status != UNLOCKED)
	    {
		if(semop(sem_id, &op_unlock[0], 1) < 0)
		{
		    return -1;
		}
		sigsetmask(mask);
		lock_status = UNLOCKED;
	    }
	    break;

    default:
	    return -1;
    }
    return 0;
}
