/*
 * Functions that manage Transport Endpoints (TEPs).
 *
 */

/*
 * FUNCTIONS IN THIS MODULE:
 *
 * init_address		- initialize an internet socket address structure
 * init_TEP		- initialize a Transport Endpoint structure
 * get_tep		- get the address of the TEP associated with the fd
 * accept_TEP		- Perform t_accept() semantics
 * open_TEP		- establish a transport endpoint
 * bind_TEP		- bind a socket address to the TPV
 * close_TEP		- close a TEP
 * connect_TEP		- connect to a another transport user
 * getinfo_TEP		- get information about TEP
 * getprotaddr_TEP	- malloc a netbuf, and initialize it
 * getstate_TEP		- get the current state of the TEP
 * listen_TEP		- listen for a connect indication
 * look_TEP		- look at the current event on a TEP
 * optmgmt_TEP		- manage options for a transport endpoint
 * rcv_TEP		- receive data sent over a connection
 * snd_TEP		- send data sent over a connection
 * rcvconnect_TEP	- receive the confirmation from a connect request
 * rcvudata_TEP		- receive a data unit
 * sndudata_TEP		- send a data unit
 * rcvuderr_TEP		- receive a data unit error indication
 * sync_TEP		- synchronise transport library
 * unbind_TEP		- disable a transport endpoint
 * snddis_TEP		- send user initiated disconnect request
 * rcvdis_TEP		- retrieve information from disconnect
 * sndrel_TEP		- initiate an orderly release
 * rcvrel_TEP		- acknowledge receipt of an orderly release indication
 * serialize_TEP	- Format the TEP structures
 *
 */

#define TEP_c
#include <string.h>
#include <sys/param.h>
#include <sys/types.h>

#ifdef sun
#include <sys/fcntl.h>
#endif

#include <stdio.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/ioctl.h>
#include <netinet/tcp.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <rpc/rpc.h>
#include <sys/syslog.h>
#include <memory.h>
#include "config.h"
#include "TEP.h"
#include "Thread.h"
#include "Pod.h"
#include <xti.h>

#ifdef u370
extern char    *calloc();

#endif

/*
 * This is the "loop-back" socket type. It has the semantics of SOCK_DGRAM.
 */

#define SOCK_LODG 99


/*
 * Function Name: init_address
 * Description: initialize an internet socket address structure.
 * Return values: pointer to the initialized structure.
 */

static struct sockaddr_in *
init_address(addr)
    struct sockaddr_in *addr;
{
    bzero((char *) addr, sizeof(struct sockaddr_in));
    addr->sin_family = AF_INET;
    addr->sin_addr.s_addr = htonl(INADDR_ANY);
    addr->sin_port = htons(0);
    return (addr);
}


/*
 * Function Name: init_TEP
 * Description: initialize a Transport Endpoint structure.
 * Return values: pointer to the initialized structure.
 */

TEP            *
init_TEP(tep, name, fd)
    TEP            *tep;
    char           *name;
    int             fd;
{
    TEP            *result = tep;

    if (!strncmp("/dev/tcp", name, 8))
    {
	tep->socket_type = (u_char) SOCK_STREAM;
    } else if (!strncmp("/dev/udp", name, 8))
    {
	tep->socket_type = (u_char) SOCK_DGRAM;
    } else if (!strncmp("/dev/lodg", name, 9))
    {
	tep->socket_type = (u_char) SOCK_LODG;
    } else
    {
	tep->socket_type = (u_char) 255;
	result = (TEP *) 0;
    }
    tep->state = (u_char) T_UNINIT;
    tep->event = (u_char) 0;
    tep->qlen = (u_char) 0;
    tep->fd = (u_char) fd;
    tep->clnt_fd = (u_char) 255;
    return (result);
}


/*
 * Function Name: get_tep
 * Description: get the address of the TEP associated with the fd.
 * Return values: pointer to the TEP structure, or a NULL pointer if fd is
 *                invalid.
 */

TEP *
get_tep(fd)
    int             fd;
{
    Thread         *thread = addr_Thread(GET_ADDR);

#ifdef XTI_DEBUG
    fprintf(stderr,"get_tep: pid: %d fd: %d\n", getpid(), fd);
#endif

    if (fd >= 0 && fd < XTI_NOFILE)
    {
	return (&(thread->teps[fd]));
    }
    else
    {
	return (TEP *)NULL;
    }
}


/*
 * Function Name: accept_TEP
 * Description: Perform t_accept() semantics. The accept was done when the
 *		listen was done - so what is done here is to undo the
 *		"application" t_open()/t_bind().
 * Return values: operation result code.
 */

int
accept_TEP(tep, client_fd, call)
    TEP            *tep;
    int             client_fd;
    struct t_call  *call;
{
    int             result;
    TEP            *client_tep;

    result = 0;
    call = call;

#ifdef XTI_DEBUG
    fprintf(stderr, "accept pid: %d state: %d fd: %d client_fd: %d\n",
	    getpid(), tep->state, tep->fd, client_fd);
#endif

    if (tep->state == (u_char) T_INCON ||
	    (tep->fd != (u_char) client_fd && tep->state == (u_char) T_IDLE))
    {
	if (tep->fd == (u_char) client_fd)
	{
	    /*
	     * We don't accept on the same file descriptor that we listen on.
	     */
	    t_errno = TINDOUT;
	    result = -1;
	} else if (tep->clnt_fd > 0 && (u_char) client_fd > 0)
	{
	    /*
	     * Duplicate the "lib opened" fd into the "app opened" fd.
	     * 
	     * The dup2() call closes the client_fd file descriptor.
	     */
	    dup2((int) tep->clnt_fd, client_fd);
	    /*
	     * Get the address of the client TEP structure.
	     */
	    client_tep = get_tep(client_fd);
	    /*
	     * The client fd in ready to transfer data.
	     */
	    client_tep->state = (u_char) T_DATAXFER;
	    /*
	     * We're done with the "lib opened" file descriptor.
	     */
	    close((int)tep->clnt_fd);
	    /*
	     * We're done with this client instance, so let's init the
	     * descriptor.
	     */
	    tep->clnt_fd = (u_char) 255;
	} else
	{
	    /*
	     * The fds weren't valid.
	     */
	    t_errno = TBADF;
	    result = -1;
	}
    } else
    {
	/*
	 * We are in the wrong state.
	 */
	result = -1;
	t_errno = TOUTSTATE;
    }
    return (result);
}


/*
 * Function Name:  open_TEP
 * Description: establish a transport endpoint.
 * Return values: fd for success and -1 for error.
 */

int
open_TEP(tep, oflag, info)
    TEP            *tep;	/* This is a tmp tep. */
    int             oflag;
    struct t_info  *info;
{
    int             osize;
    int             rcv_size;
    int             snd_size;
    int             result;
    int             socket_type;

    result = -1;
    if (tep->state == (u_char) T_UNINIT)
    {
	if (tep->socket_type == (u_char) SOCK_DGRAM ||
		tep->socket_type == (u_char) SOCK_LODG)
	{
	    socket_type = SOCK_DGRAM;
	} else
	{
	    socket_type = SOCK_STREAM;
	}
	if ((result = socket(AF_INET, socket_type, 0)) < 0)
	{
	    t_errno = TSYSERR;
	} else
	{
	    tep->fd = (u_char) result;
	    /* increment count of open TEPs */
	    tep->state = (u_char) T_UNBND;
	}
    } else
    {
	t_errno = TOUTSTATE;
    }

#ifdef XTI_DEBUG
    fprintf(stderr, "open pid: %d state: %d fd: %d \n",
	    getpid(), tep->state, tep->fd);
#endif

    if (info)
    {
	info->addr = sizeof(struct sockaddr_in);
	info->connect = -2;
	info->discon = -2;
	info->flags = T_SENDZERO;
	if (tep->socket_type == (u_char) SOCK_STREAM)
	{
	    info->options = sizeof(struct tt_soopt);
	    info->tsdu = 0;
	    info->etsdu = -1;
	    info->servtype = T_COTS;
	} else if (tep->socket_type == (u_char) SOCK_DGRAM ||
		   tep->socket_type == (u_char) SOCK_LODG)
	{
	    info->options = -2;
	    if (result >= 0)
	    {
		osize = sizeof(rcv_size);
		getsockopt((int) tep->fd, SOL_SOCKET, SO_RCVBUF, (char *) &rcv_size,
			   &osize);
		osize = sizeof(snd_size);
		getsockopt((int) tep->fd, SOL_SOCKET, SO_SNDBUF, (char *) &snd_size,
			   &osize);
		info->tsdu = MIN(rcv_size, snd_size);
	    }
	    info->etsdu = -2;
	    info->servtype = T_CLTS;
	} else
	{
	    t_errno = TBADNAME;
	}
    }
    return (result);
}


/*
 * Function Name:  bind_TEP
 * Description: bind a socket address to the TPV.
 * Return values: 0 for success and -1 for error.
 */

int
bind_TEP(tep, req, ret)
    TEP            *tep;
    struct t_bind  *req;
    struct t_bind  *ret;
{
    int             result;
    int             retlen;
    struct sockaddr_in local_addr;
    struct sockaddr_in *local_addr_ptr = &local_addr;
    int             addr_size;
    struct servent *service_entry;
    char           *inetstr;
    char           *service_name;
    int             service_port;
    long            inetaddr;
    struct hostent *host_entry;
    struct in_addr *inet_address;


#ifdef XTI_DEBUG
    fprintf(stderr, "bind pid: %d state: %d fd: %d\n",
	    getpid(), tep->state, tep->fd);
#endif

    addr_size = sizeof(struct sockaddr_in);
    if (tep->state == (u_char) T_UNBND)
    {
	/*
	 * Case where caller doesn't care to which address the TEP is bound.
	 */
	if (req == (struct t_bind *) 0)
	{
	    tep->qlen = (u_char) 0;
	    init_address(local_addr_ptr);
	} else
	{
	    /*
	     * Same as previous case, except the qlen is specified.
	     */
	    if (req->addr.len == 0)
	    {
		init_address(local_addr_ptr);
	    }
	    /*
	     * The bind address is provided by the caller.
	     */
	    else if (req->addr.len > 0 && req->addr.len <= addr_size)
	    {
		/*
		 * If it is a loop-back socket then the address in string
		 * format - so convert it to sockaddr_in.
		 * 
		 * Syntax is: address : type1 | type2;   type1 : hostname '.'
		 * service_name ; type2 : inet_addr '.' service_name ;
		 * inet_addr : net_id '.' host_id ; net_id : upper_byte '.'
		 * lower_byte ; host_id : upper_byte '.' lower_byte ;
		 */
		if (tep->socket_type == (u_char) SOCK_LODG)
		{
		    inetstr = (char *) calloc(1, req->addr.len);
		    strncpy(inetstr, req->addr.buf, req->addr.len);
		    if ((service_name = strrchr(inetstr, '.')) != (char *) 0)
		    {
			/*
			 * separate the host and service parts.
			 */
			*service_name = '\0';
			service_name++;
			/*
			 * get the service entry so we know the port.
			 */
			if ((service_entry = getservbyname(service_name, NULL)) !=
				(struct servent *) 0)
			{
			    service_port = service_entry->s_port;
			    local_addr_ptr->sin_port = htons(service_port);
			    /*
			     * get the inet address from the hostname.
			     */
			    if ((host_entry = gethostbyname(inetstr)) >
				    (struct hostent *) 0)
			    {
				inet_address = (struct in_addr *) host_entry->h_addr_list[0];
				local_addr_ptr->sin_addr = *inet_address;

#ifdef XTI_DEBUG
				fprintf(stderr,
					"\thost: %s service: %s port: %d addr: %s\n",
					inetstr, service_name, service_port,
					inet_ntoa(*inet_address));
#endif
			    }
			    /*
			     * if that didn't work see if it is in inet
			     * string format.
			     */
			    else if ((inetaddr = inet_addr(inetstr)) >= 0)
			    {
				local_addr_ptr->sin_addr.s_addr = inetaddr;
			    }
			}
		    }
		    free(inetstr);
		} else
		{
		    local_addr_ptr = (struct sockaddr_in *) req->addr.buf;
		}
	    }
	    tep->qlen = (u_char) req->qlen;
	}


	if ((result = bind((int) tep->fd, (struct sockaddr *) local_addr_ptr,
			   sizeof(struct sockaddr_in))) == -1)
	{
	    switch (errno)
	    {
	    case EADDRINUSE:
	    case EADDRNOTAVAIL:
		t_errno = TNOADDR;
		break;
	    default:
		t_errno = TSYSERR;
	    }
	} else
	{
	    retlen = addr_size;
	    tep->state = (u_char) T_IDLE;
	    tep->event = (u_char) T_LISTEN;
	    if (tep->qlen > (u_char) 0 &&
		    (!(tep->socket_type == (u_char) SOCK_DGRAM ||
		       tep->socket_type == (u_char) SOCK_LODG)))
	    {
		if ((result = listen((int) tep->fd, tep->qlen)) < 0)
		{
		    t_errno = TSYSERR;
		}
	    }
	    result = getsockname((int) tep->fd, local_addr_ptr, &retlen);
	}
	if (ret != (struct t_bind *) 0)
	{
	    if (addr_size <= ret->addr.maxlen)
	    {
		ret->addr.len = addr_size;
		memcpy(ret->addr.buf, (char *) local_addr_ptr, addr_size);
	    } else
	    {
		result = -1;
		t_errno = TBUFOVFLW;
	    }
	}
    } else
    {
	result = -1;
	t_errno = TOUTSTATE;
    }
    return (result);
}



/*
 * Function Name:  close_TEP
 * Description: close a TEP.
 * Return values: 0 for success and -1 for error.
 */

int
close_TEP(tep)
    TEP            *tep;
{
    int             result;

#ifdef XTI_DEBUG
    fprintf(stderr, "close_TEP pid: %d state: %d fd: %d\n",
	    getpid(), tep->state, tep->fd);
#endif

    if (tep->state != (u_char) T_UNINIT)
    {
	if ((result = close((int) tep->fd)) >= 0)
	{
	    tep->fd = (u_char) 255;
	    tep->clnt_fd = (u_char) 255;
	    tep->state = (u_char) T_UNINIT;

	} else
	{
	    result = -1;
	}
    } else
    {
	result = -1;
    }
    return (result);
}


/*
 * Function Name:  connect_TEP
 * Description: connect to a another transport user.
 * Return values: 0 for success and -1 for error.
 */

int
connect_TEP(tep, sndcall, rcvcall)
    TEP            *tep;
    struct t_call  *sndcall;
    struct t_call  *rcvcall;
{
    int             result = -1;
    struct sockaddr paddr;
    int             plen = sizeof(struct sockaddr);;

#ifdef XTI_DEBUG
    fprintf(stderr, "connect pid: %d state: %d fd: %d\n",
	    getpid(), tep->state, tep->fd);
#endif

    if (tep->state == (u_char) T_IDLE && sndcall)
    {
	if ((result = connect((int) tep->fd, sndcall->addr.buf, sndcall->addr.len))
		== -1)
	{
	    switch (result)
	    {
	    case EINPROGRESS:
		t_errno = TNODATA;
		tep->state = (u_char) T_OUTCON;
		break;
	    case EINTR:
		t_errno = TSYSERR;
		tep->state = (u_char) T_OUTCON;
		break;
	    default:
		t_errno = TSYSERR;
	    }
	} else
	{
	    tep->state = (u_char) T_DATAXFER;
	    if (rcvcall != (struct t_call *) 0)
	    {
		if ((result = getpeername((int) tep->fd, &paddr, &plen)) >= 0)
		{
		    if (plen <= rcvcall->addr.maxlen)
		    {
			rcvcall->addr.len = plen;
			memcpy(rcvcall->addr.buf, (char *) &paddr, plen);
		    }
		}
	    }
	}
    } else
    {
	t_errno = TOUTSTATE;
    }
    return (result);
}


/*
 * Function Name:  getinfo_TEP
 * Description: get information about TEP.
 * Return values: if the fd is good, fd is returned, -1 for error.
 */

int
getinfo_TEP(tep, info)
    TEP            *tep;
    struct t_info  *info;
{
    int             osize;
    int             rcv_size;
    int             snd_size;
    int             result = (int) tep->fd;


#ifdef XTI_DEBUG
    fprintf(stderr, "getinfo pid: %d state: %d fd: %d\n",
	    getpid(), tep->state, tep->fd);
#endif

    if (info != (struct t_info *) 0)
    {
	info->addr = sizeof(struct sockaddr_in);
	info->connect = -2;
	info->discon = -2;
	info->flags = T_SENDZERO;
	if (tep->socket_type == (u_char) SOCK_STREAM)
	{
	    info->options = sizeof(struct tt_soopt);
	    info->tsdu = 0;
	    info->etsdu = -1;
	    info->servtype = T_COTS;
	} else if (tep->socket_type == (u_char) SOCK_DGRAM ||
		   tep->socket_type == (u_char) SOCK_LODG)
	{
	    info->options = -2;
	    if (result >= 0)
	    {
		osize = sizeof(rcv_size);
		getsockopt((int) tep->fd, SOL_SOCKET, SO_RCVBUF, (char *) &rcv_size,
			   &osize);
		osize = sizeof(snd_size);
		getsockopt((int) tep->fd, SOL_SOCKET, SO_SNDBUF, (char *) &snd_size,
			   &osize);
		info->tsdu = MIN(rcv_size, snd_size);
	    }
	    info->etsdu = -2;
	    info->servtype = T_CLTS;
	} else
	{
	    t_errno = TBADNAME;
	    result = -1;
	}
    }
    return (result);
}


/*
 * Function Name:   getprotaddr_TEP
 * Description: malloc a netbuf, and initialize it.
 * Return values: 0 for success and -1 for error.
 */

int
getprotaddr_TEP(tep, boundaddr, peeraddr)
    TEP            *tep;
    struct t_bind  *boundaddr;
    struct t_bind  *peeraddr;
{
    struct sockaddr paddr;
    int             plen;
    struct sockaddr baddr;
    int             blen;
    int             result;

#ifdef XTI_DEBUG
    fprintf(stderr, "getprotaddr pid: %d state: %d fd: %d\n",
	    getpid(), tep->state, tep->fd);
#endif

    if ((result = getsockname((int) tep->fd, &baddr, &blen)) >= 0 &&
	    (result = getpeername((int) tep->fd, &paddr, &plen)) >= 0)
    {
	switch ((u_char) tep->state)
	{
	case T_UNBND:
	    if (boundaddr != (struct t_bind *) 0)
	    {
		boundaddr->addr.len = 0;
	    }
	    if (peeraddr != (struct t_bind *) 0)
	    {
		peeraddr->addr.len = 0;
	    }
	    break;
	case T_DATAXFER:
	    if (peeraddr != (struct t_bind *) 0)
	    {
		if (peeraddr->addr.maxlen >= plen)
		{
		    memcpy(peeraddr->addr.buf, (char *) &paddr, plen);
		    peeraddr->addr.len = plen;
		} else
		{
		    t_errno = TBUFOVFLW;
		    result = -1;
		}
	    }
	    break;
	default:
	    if (peeraddr != (struct t_bind *) 0)
	    {
		peeraddr->addr.len = 0;
	    }
	    if (boundaddr != (struct t_bind *) 0)
	    {
		if (boundaddr->addr.maxlen >= blen)
		{
		    memcpy(boundaddr->addr.buf, (char *) &baddr, blen);
		    boundaddr->addr.len = blen;
		} else
		{
		    t_errno = TBUFOVFLW;
		    result = -1;
		}
	    }
	}
    } else
    {
	t_errno = TSYSERR;
    }
    return (result);
}


/*
 * Function Name:   getstate_TEP
 * Description: get the current state of the TEP.
 * Return values: state for success and -1 for error.
 */

int
getstate_TEP(tep)
    TEP            *tep;
{
    int             result;

#ifdef XTI_DEBUG
    fprintf(stderr, "getstate pid: %d state: %d fd: %d\n",
	    getpid(), tep->state, tep->fd);
#endif

    if (tep->fd != (u_char) 255)
    {
	result = (int) tep->state;
    } else
    {
	t_errno = TBADF;
	result = -1;
    }
    return (result);
}


/*
 * Function Name:   listen_TEP
 * Description: listen for a connect indication.
 * Return values: 0 for success and -1 for error.
 */

int
listen_TEP(tep, call)
    TEP            *tep;
    struct t_call  *call;
{
    int             result;
    sockaddr_in     from;
    int             fromlen;
    int             sockaddr_len;

#ifdef XTI_DEBUG
    fprintf(stderr, "listen pid: %d state: %d fd: %d\n",
	    getpid(), tep->state, tep->fd);
#endif

    result = 0;
    sockaddr_len = sizeof(struct sockaddr_in);
    if (tep->state == (u_char) T_IDLE || tep->state == (u_char) T_INCON)
    {
	fromlen = sizeof(struct sockaddr_in);
	result = accept((int) tep->fd, &from, &fromlen);
	tep->clnt_fd = (u_char) result;
	if (result < 0)
	{
	    result = -1;
	    if (errno == EWOULDBLOCK)
	    {
		t_errno = TNODATA;
	    } else
	    {
		t_errno = TSYSERR;
	    }
	} else if (call->addr.maxlen >= sockaddr_len)
	{
	    call->addr.len = sockaddr_len;
	    memcpy(call->addr.buf, (char *) &from, sockaddr_len);
	}
	tep->state = (u_char) T_INCON;
	tep->event = (u_char) T_LISTEN;
    } else
    {
	result = -1;
	t_errno = TOUTSTATE;
    }
    return (result);
}


/*
 * Function Name:   look_TEP
 * Description: look at the current event on a TEP.
 * Return values: event for success and -1 for error.
 */

int
look_TEP(tep)
    TEP            *tep;
{
    int             result;

#ifdef XTI_DEBUG
    fprintf(stderr, "look pid: %d state: %d fd: %d\n",
	    getpid(), tep->state, tep->fd);
#endif

    if (tep->fd != (u_char) 255)
    {
	result = (int) tep->event;
    } else
    {
	t_errno = TBADF;
	result = -1;
    }
    return (result);
}


/*
 * Function Name:   optmgmt_TEP
 * Description: manage options for a transport endpoint.
 * Return values: 0 for success and -1 for error.
 */

int
optmgmt_TEP(tep, req, ret)
    TEP            *tep;
    struct t_optmgmt *req;
    struct t_optmgmt *ret;
{
    int             result;
    struct tt_soopt *req_sockopts;
    struct tt_soopt *ret_sockopts;
    int             size = sizeof(int);;

#ifdef XTI_DEBUG
    fprintf(stderr, "optmgmt pid: %d state: %d fd: %d\n",
	    getpid(), tep->state, tep->fd);
#endif

    result = 0;
    /*
     * We do nothing if we are not in the T_IDLE state.
     */
    if (tep->state == (u_char) T_IDLE)
    {  
        /*
         * If there is no req sturcture there is nothing to do.
         */
        if (req != (struct t_optmgmt *) 0)
        {
            /*
             * We have a request - let's do it.
             */
            switch (req->flags)
            {  
	    case T_DEFAULT:
			/*
			 * Only set the broadcast option since that it the
			 * only option currently used by the tirpc library.
			 */
			getsockopt((int) tep->fd, SOL_SOCKET, SO_BROADCAST,
			     (char *) &(ret_sockopts->tts_broadcast), &size);
			break;
            case T_NEGOTIATE:
                if ((req_sockopts = (struct tt_soopt *) req->opt.buf))
                {
                    if(req_sockopts->tts_broadcast)
                    {
                        if ((result = setsockopt((int) tep->fd, SOL_SOCKET,
                                             SO_BROADCAST,
                                             (char *) &req_sockopts->tts_broadcast,
                                             sizeof(int))) == 0)
                        {
                            if ((ret_sockopts = (struct tt_soopt *) ret->opt.buf))
                            {
                                getsockopt((int) tep->fd, SOL_SOCKET,
                                           SO_BROADCAST,
                                           (char *) &(ret_sockopts->tts_broadcast),
                                           &size);
                            }
                        } else
                        {
                            result = -1;
                            t_errno = TSYSERR;
                        }
                        break;
                    }    
                }
            default:
                result = -1;
                t_errno = TNOTSUPPORT;
            }
        } else
        {

            result = -1;                      
            t_errno = TPROTO;
        }
    } else
    {
        result = -1;
        t_errno = TOUTSTATE;
    }
    return (result);
}


/*
 * Function Name:   rcv_TEP
 * Description: receive data sent over a connection.
 * Return values: number of bytes received for success and -1 for error.
 */

int
rcv_TEP(tep, buf, nbytes, flags)
    TEP            *tep;
    char           *buf;
    unsigned int    nbytes;
    int            *flags;
{
    int             result;

    result = 0;
    *flags = 0;

#ifdef XTI_DEBUG
    fprintf(stderr, "rcv pid: %d state: %d fd: %d\n",
	    getpid(), tep->state, tep->fd);
#endif

    if (tep->state == (u_char) T_DATAXFER || tep->state == (u_char) T_OUTREL)
    {
	while ((result = recv((int) tep->fd, buf, nbytes, 0)) < 0)
	{
	    if (errno == EINTR)
	    {
		continue;
	    } else if (errno == EWOULDBLOCK)
	    {
		*flags = T_MORE;
		t_errno = TNODATA;
		break;
	    } else
	    {
		t_errno = TSYSERR;
		break;
	    }
	}
	tep->event = (u_char) T_DATA;
    } else
    {
	result = -1;
	t_errno = TOUTSTATE;
    }
    return (result);
}


/*
 * Function Name:   snd_TEP
 * Description: send data sent over a connection.
 * Return values: number of bytes sent for success and -1 for error.
 */

int
snd_TEP(tep, buf, nbytes, flags)
    TEP            *tep;
    char           *buf;
    unsigned int    nbytes;
    int             flags;
{
    int             result;

#ifdef XTI_DEBUG
    fprintf(stderr, "snd pid: %d state: %d fd: %d\n",
	    getpid(), tep->state, tep->fd);
#endif

    result = 0;
    flags = flags;
    if (tep->state == (u_char) T_DATAXFER || tep->state == (u_char) T_INREL)
    {
	while ((result = send((int) tep->fd, buf, (int) nbytes, 0)) < 0)
	{
	    if (errno == EINTR)
	    {
		continue;
	    } else if (errno == EWOULDBLOCK)
	    {
		t_errno = TFLOW;
		break;
	    } else
	    {
		t_errno = TSYSERR;
		break;
	    }
	}
    } else
    {
	result = -1;
	t_errno = TOUTSTATE;
    }
    return (result);
}


/*
 * Function Name:   rcvconnect_TEP
 * Description: receive the confirmation from a connect request.
 * Return values: 0 for success and -1 for error.
 */

int
rcvconnect_TEP(tep, call)
    TEP            *tep;
    struct t_call  *call;
{
    int             result;
    struct sockaddr baddr;
    int             blen;
    struct sockaddr paddr;
    int             plen;

#ifdef XTI_DEBUG
    fprintf(stderr, "rcvconnect pid: %d state: %d fd: %d\n",
	    getpid(), tep->state, tep->fd);
#endif

    result = 0;
    if (tep->state == (u_char) T_OUTCON)
    {
	if (((result = getsockname((int) tep->fd, &baddr, &blen)) >= 0) &&
		(result = connect((int) tep->fd, &baddr, blen))
		== -1)
	{
	    switch (result)
	    {
	    case EINPROGRESS:
		t_errno = TNODATA;
		tep->state = (u_char) T_OUTCON;
		break;
	    case EINTR:
		t_errno = TSYSERR;
		tep->state = (u_char) T_OUTCON;
		break;
	    default:
		t_errno = TSYSERR;
	    }
	} else if (result >= 0)
	{
	    tep->event = (u_char) T_CONNECT;
	    tep->state = (u_char) T_DATAXFER;
	    if (call != (struct t_call *) 0)
	    {
		if ((result = getpeername((int) tep->fd, &paddr, &plen)) >= 0)
		{
		    if (plen <= call->addr.maxlen)
		    {
			call->addr.len = plen;
			memcpy(call->addr.buf, (char *) &paddr, plen);
		    }
		}
	    }
	}
    } else
    {
	result = -1;
	t_errno = TOUTSTATE;
    }
    return (result);
}


/*
 * Function Name:   rcvudata_TEP
 * Description: receive a data unit.
 * Return values: number of bytes received for success and -1 for error.
 */

int
rcvudata_TEP(tep, unitdata, flags)
    TEP            *tep;
    struct t_unitdata *unitdata;
    int            *flags;
{
    int             result;
    sockaddr_in    *from;
    int             fromlen;
    char           *buf;
    int             nbytes;

    struct sockaddr_in local_addr;
    struct sockaddr_in *local_addr_ptr = &local_addr;
    struct servent *service_entry;
    char           *inetstr;
    char           *service_name;
    int             service_port;
    long            inetaddr;
    struct hostent *host_entry;
    struct in_addr *inet_address;

#ifdef XTI_DEBUG
    fprintf(stderr, "rcvudata pid: %d state: %d fd: %d\n",
	    getpid(), tep->state, tep->fd);
#endif

    *flags = 0;
    if (tep->state == (u_char) T_IDLE)
    {
	nbytes = unitdata->udata.maxlen;
	buf = unitdata->udata.buf;
	fromlen = unitdata->addr.maxlen;
	if (tep->socket_type == (u_char) SOCK_LODG)
	{
	    init_address(local_addr_ptr);
	    inetstr = (char *) calloc(1, unitdata->addr.len);
	    strncpy(inetstr, unitdata->addr.buf, unitdata->addr.len);
	    if ((service_name = strrchr(inetstr, '.')) != (char *) 0)
	    {
		/*
		 * separate the host and service parts.
		 */
		*service_name = '\0';
		service_name++;
		/*
		 * get the service entry so we know the port.
		 */
		if ((service_entry = getservbyname(service_name, NULL)) !=
			(struct servent *) 0)
		{
		    service_port = service_entry->s_port;
		    local_addr_ptr->sin_port = htons(service_port);
		    /*
		     * get the inet address from the hostname.
		     */
		    if ((host_entry = gethostbyname(inetstr)) >
			    (struct hostent *) 0)
		    {
			inet_address = (struct in_addr *) host_entry->h_addr_list[0];
			local_addr_ptr->sin_addr = *inet_address;

#ifdef XTI_DEBUG
			fprintf(stderr,
			       "\thost: %s service: %s port: %d addr: %s\n",
				inetstr, service_name, service_port,
				inet_ntoa(*inet_address));
#endif
		    }
		    /*
		     * if that didn't work see if it is in inet string
		     * format.
		     */
		    else if ((inetaddr = inet_addr(inetstr)) >= 0)
		    {
			local_addr_ptr->sin_addr.s_addr = inetaddr;
		    }
		}
	    }
	    free(inetstr);
	    from = local_addr_ptr;
	} else
	{
	    from = (sockaddr_in *) unitdata->addr.buf;
	}
	while ((result = recvfrom((int) tep->fd,
		   buf, nbytes, 0, (struct sockaddr *) from, &fromlen)) < 0)
	{
	    if (errno == EINTR)
	    {
		continue;
	    } else if (errno == EWOULDBLOCK)
	    {
		*flags = T_MORE;
		t_errno = TNODATA;
		break;
	    } else
	    {
		t_errno = TSYSERR;
		break;
	    }
	}
	unitdata->udata.len = result;
	unitdata->addr.len = fromlen;
	tep->event = (u_char) T_DATA;
    } else
    {
	result = -1;
	t_errno = TOUTSTATE;
    }
    return (result);
}


/*
 * Function Name:   sndudata_TEP
 * Description: send a data unit.
 * Return values: number of bytes sent for success and -1 for error.
 */

int
sndudata_TEP(tep, unitdata)
    TEP            *tep;
    struct t_unitdata *unitdata;
{
    int             result;
    sockaddr_in    *to;
    int             tolen;
    char           *buf;
    int             nbytes;

    struct sockaddr_in local_addr;
    struct sockaddr_in *local_addr_ptr = &local_addr;
    struct servent *service_entry;
    char           *inetstr;
    char           *service_name;
    int             service_port;
    long            inetaddr;
    struct hostent *host_entry;
    struct in_addr *inet_address;

#ifdef XTI_DEBUG
    fprintf(stderr, "sndudata pid: %d state: %d fd: %d\n",
	    getpid(), tep->state, tep->fd);
#endif

    if (tep->state == (u_char) T_IDLE)
    {
	nbytes = unitdata->udata.len;
	buf = unitdata->udata.buf;
	tolen = unitdata->addr.len;
	if (tep->socket_type == (u_char) SOCK_LODG)
	{
	    init_address(local_addr_ptr);
	    inetstr = (char *) calloc(1, unitdata->addr.len);
	    strncpy(inetstr, unitdata->addr.buf, unitdata->addr.len);
	    if ((service_name = strrchr(inetstr, '.')) != (char *) 0)
	    {
		/*
		 * separate the host and service parts.
		 */
		*service_name = '\0';
		service_name++;
		/*
		 * get the service entry so we know the port.
		 */
		if ((service_entry = getservbyname(service_name, NULL)) !=
			(struct servent *) 0)
		{
		    service_port = service_entry->s_port;
		    local_addr_ptr->sin_port = htons(service_port);
		    /*
		     * get the inet address from the hostname.
		     */
		    if ((host_entry = gethostbyname(inetstr)) >
			    (struct hostent *) 0)
		    {
			inet_address = (struct in_addr *) host_entry->h_addr_list[0];
			local_addr_ptr->sin_addr = *inet_address;

#ifdef XTI_DEBUG
			fprintf(stderr,
			       "\thost: %s service: %s port: %d addr: %s\n",
				inetstr, service_name, service_port,
				inet_ntoa(*inet_address));
#endif
		    }
		    /*
		     * if that didn't work see if it is in inet string
		     * format.
		     */
		    else if ((inetaddr = inet_addr(inetstr)) >= 0)
		    {
			local_addr_ptr->sin_addr.s_addr = inetaddr;
		    }
		}
	    }
	    free(inetstr);
	    to = local_addr_ptr;
	    tolen = sizeof(struct sockaddr_in);
	} else
	{
	    to = (sockaddr_in *) unitdata->addr.buf;
	}
#ifdef XTI_DEBUG
	fprintf(stderr, "sendto: to address = %x\n", to->sin_addr);
#endif
	while ((result = sendto((int) tep->fd, buf, nbytes, 0,
				(struct sockaddr *) to, tolen)) < 0)
	{
	    if (errno == EINTR)
	    {
		continue;
	    } else if (errno == EWOULDBLOCK)
	    {
		t_errno = TFLOW;
		break;
	    } else
	    {
		t_errno = TSYSERR;
		break;
	    }
	}
	if (result > 0)
	{
	    result = 0;
	}
    } else
    {
	result = -1;
	t_errno = TOUTSTATE;
    }
    return (result);
}


/*
 * Function Name:   rcvuderr_TEP
 * Description: receive a data unit error indication.
 * Return values: 0 for success and -1 for error.
 */

int
rcvuderr_TEP(tep, uderr)
    TEP            *tep;
    struct t_uderr *uderr;
{
    int             result;

#ifdef XTI_DEBUG
    fprintf(stderr, "rcvuderr pid: %d state: %d fd: %d\n",
	    getpid(), tep->state, tep->fd);
#endif

    result = -1;
    uderr = uderr;
    if (tep->state == (u_char) T_IDLE)
    {
	t_errno = TNOTSUPPORT;
    } else
    {
	t_errno = TOUTSTATE;
    }
    return (result);
}



/*
 * Function Name:   sync_TEP
 * Description: synchronise transport library.
 * Return values: state of tep for success and -1 for error.
 */

int
sync_TEP(tep)
    TEP            *tep;
{
    Thread	   *thread = addr_Thread(GET_ADDR);
    int             result = 0;
    TEP            *my_tep;

#ifdef XTI_DEBUG
    fprintf(stderr, "sync pid: %d state: %d fd: %d\n",
	    getpid(), tep->state, tep->fd);
#endif

    if (thread != (Thread *) 0)
    {
	if (tep->state != (u_char) T_UNINIT)
	{
	    result = (int) tep->state;
	    if ((my_tep = get_tep((int) tep->fd)) > (TEP *) 0)
	    {
		memcpy((char *) my_tep, (char *) tep, sizeof(TEP));
	    }
	} else
	{
	    result = -1;
	    t_errno = TOUTSTATE;
	}
    } else
    {
	result = -1;
    }
    return (result);
}


/*
 * Function Name:   unbind_TEP
 * Description: disable a transport endpoint.
 * Return values: 0 for success and -1 for error.
 */

int
unbind_TEP(tep)
    TEP            *tep;
{
    int             result;

#ifdef XTI_DEBUG
    fprintf(stderr, "unbind pid: %d state: %d fd: %d\n",
	    getpid(), tep->state, tep->fd);
#endif

    result = -1;
    if (tep->state == (u_char) T_IDLE)
    {
	t_errno = TNOTSUPPORT;
    } else
    {
	t_errno = TOUTSTATE;
    }
    tep->state = (u_char) T_UNBND;
    return (result);
}


/*
 * Function Name:   snddis_TEP
 * Description: send user initiated disconnect request.
 * Return values: 0 for success and -1 for error.
 */

int
snddis_TEP(tep, call)
    TEP            *tep;
    struct t_call  *call;
{
    int             result;

#ifdef XTI_DEBUG
    fprintf(stderr, "snddis pid: %d state: %d fd: %d\n",
	    getpid(), tep->state, tep->fd);
#endif

    result = -1;
    call = call;
    if (tep->state == (u_char) T_DATAXFER ||
	    tep->state == (u_char) T_OUTCON ||
	    tep->state == (u_char) T_OUTREL ||
	    tep->state == (u_char) T_INREL ||
	    tep->state == (u_char) T_INCON)
    {
	t_errno = TNOTSUPPORT;
    } else
    {
	t_errno = TOUTSTATE;
    }
    tep->state = (u_char) T_IDLE;
    return (result);
}


/*
 * Function Name:   rcvdis_TEP
 * Description: retrieve information from disconnect.
 * Return values: 0 for success and -1 for error.
 */

int
rcvdis_TEP(tep, discon)
    TEP            *tep;
    struct t_discon *discon;
{
    int             result;

#ifdef XTI_DEBUG
    fprintf(stderr, "rcvdis pid: %d state: %d fd: %d\n",
	    getpid(), tep->state, tep->fd);
#endif

    result = -1;
    discon = discon;
    if (tep->state == (u_char) T_DATAXFER ||
	    tep->state == (u_char) T_OUTCON ||
	    tep->state == (u_char) T_OUTREL ||
	    tep->state == (u_char) T_INREL ||
	    tep->state == (u_char) T_INCON)
    {
	t_errno = TNOTSUPPORT;
	if (tep->state != (u_char) T_INCON)
	{
	    tep->state = (u_char) T_IDLE;
	}
    } else
    {
	t_errno = TOUTSTATE;
    }
    return (result);
}


/*
 * Function Name:   sndrel_TEP
 * Description: initiate an orderly release.
 * Return values: 0 for success and -1 for error.
 */

int
sndrel_TEP(tep)
    TEP            *tep;
{
    int             result;

#ifdef XTI_DEBUG
    fprintf(stderr, "sndrel pid: %d state: %d fd: %d\n",
	    getpid(), tep->state, tep->fd);
#endif

    result = -1;
    if (tep->state == (u_char) T_DATAXFER ||
	    tep->state == (u_char) T_INREL)
    {
	t_errno = TNOTSUPPORT;
	if (tep->state == (u_char) T_DATAXFER)
	{
	    tep->state = (u_char) T_OUTREL;
	} else
	{
	    tep->state = (u_char) T_IDLE;
	}
    } else
    {
	t_errno = TOUTSTATE;
    }
    return (result);
}


/*
 * Function Name:   rcvrel_TEP
 * Description: acknowledge receipt of an orderly release indication.
 * Return values: 0 for success and -1 for error.
 */

int
rcvrel_TEP(tep)
    TEP            *tep;
{
    int             result;

#ifdef XTI_DEBUG
    fprintf(stderr, "rcvrel pid: %d state: %d fd: %d\n",
	    getpid(), tep->state, tep->fd);
#endif

    result = -1;
    if (tep->state == (u_char) T_DATAXFER ||
	    tep->state == (u_char) T_OUTREL)
    {
	t_errno = TNOTSUPPORT;
	if (tep->state == (u_char) T_DATAXFER)
	{
	    tep->state = (u_char) T_INREL;
	} else
	{
	    tep->state = (u_char) T_IDLE;
	}
    } else
    {
	t_errno = TOUTSTATE;
    }
    return (result);
}
