/*
 * Functions for managing TEPs associated with process threads
 */

/*
 * FUNCTIONS IN THIS MODULE:
 *
 * addr_Thread		    - get the address of the thread, if not available
 *			      create it
 * init_Thread		    - Initailize a newly allocated thread structure
 * alloc_Thread		    - Allocate a thread to pass to a child process
 *
 */

#define Proc_c
#include <stdio.h>
#include <sys/types.h>
#include <sys/param.h>

#ifdef sun
#include <sys/syscall.h>
#endif

#include <rpc/rpc.h>
#include <errno.h>
#include <memory.h>
#include "TEP.h"
#include "config.h"
#include "Thread.h"
#include "Pod.h"
#include <xti.h>


/*
 * Function Name: addr_Thread
 * Description:  get the address of the thread.  Find the pod and get it
 *		 if it is not defined.  Create a pod and a thread if
 *		 none exist.
 * Return Values: address of thread structure.
 *		  if there is a system error in allocating, attaching, etc.,
 *		  return a NULL pointer.
 */

Thread *
addr_Thread(op)
int op;
{
    static Thread *thread;   /* Thread for this process */
    Pod *pod;
    int inx;


    switch(op)
    {
    case GET_ADDR:
	if(thread == (Thread *)0)
	{
	/* check to see if there is an xti  shared memory segment
	 * if yes, then attach it
	 * else make one and attach it
	 */

	    /* allocate space for our thread */
	    if((thread = alloc_Thread()) <= (Thread *)NULL)
	    {
		return (Thread *)NULL;
	    }

	    if((pod = addr_Pod(GET_ADDR)) == (Pod *)NULL)
	    {
		return (Thread *)NULL;
	    }

	    /* look for our thread in the pod */
	    for(inx=0; inx < XTI_MAX_THREADS; inx++)
	    {
		if(pod->threads[inx].pid == getpid())
		{
		    memcpy((char *)thread, (char *)(&(pod->threads[inx])),
			   sizeof(Thread));
		    pod->threads[inx].active = 0;

#ifdef XTI_DEBUG
		    fprintf(stderr, "process %d: thread copied from pod\n",
			    getpid());
#endif
		    break;
		}
	    }

	    /* detach the pod */
	    /* we don't need it anymore, unless we spawn a child */
	    if(addr_Pod(FREE_ADDR) < 0)
	    {
		return (Thread *)NULL;
	    }
	    else
	    {
		return thread;
	    }
	}
	else
	{
	    return thread;
	}
        break;

    case FREE_ADDR:
	{
	    Thread *thread_ptr = thread;

	    if(free((char *)thread) < 0)
	    {
		return (Thread *)NULL;
	    }
	    thread = (Thread *)NULL;
	    return(thread_ptr);
	    break;
    
	}
    }
}


/*
 * Function Name: init_Thread
 * Description:  Initailize a newly allocated thread structure.
 * Return values: The thread pointer.
 */

Thread         *
init_Thread(thread)
    Thread         *thread;
{
    int             inx;

    thread->active = 1;
    thread->pid = getpid();
    for (inx = 0; inx < XTI_NOFILE; inx++)
    {
	(void)init_TEP(&(thread->teps[inx]), "", 255);
    }
    return (thread);
}


/*
 * Function Name: alloc_Thread
 * Description: Allocate a thread to pass to a child process.
 * Return values: Pointer to the thread to use or a NULL pointer on error.
 */

Thread         *
alloc_Thread()
{
    Thread         *thread = (Thread *)NULL;

    if((thread = (Thread *)calloc(1, sizeof(Thread))) != (Thread *)NULL)
    {
	thread = init_Thread(thread);
    }
    return (thread);
}
