/*
 * Copyright (c) 1988, Sun Microsystems, Inc.  All Rights Reserved. Sun
 * considers its source code as an unpublished, proprietary trade secret, and
 * it is available only under strict license provisions.  This copyright
 * notice is placed here only to protect Sun in the event the source is
 * deemed a published work.  Dissassembly, decompilation, or other means of
 * reducing the object code to human readable form is prohibited by the
 * license agreement under which this code is provided to the user or company
 * in possession of this copy.
 * 
 * RESTRICTED RIGHTS LEGEND: Use, duplication, or disclosure by the Government
 * is subject to restrictions as set forth in subparagraph (c)(1)(ii) of the
 * Rights in Technical Data and Computer Software clause at DFARS 52.227-7013
 * and in similar clauses in the FAR and NASA FAR Supplement.
 * 
 * (C) Copyright 1987 Sun Microsystems, Inc. All rights reserved. This material
 * contains trade secret information of Sun Microsystems, Inc. Use,
 * reproduction, or disclosure is prohibited, except under a written license
 * by Sun Microsystems, Inc. Use of copyright notice is precautionary only
 * and does not imply publication or disclosure.
 */

/*
 * This module defines the api for the library. All functions that require
 * the TEP (Transport Endpoint) as an argument are in the TEP.c module.
 */

/*
 * FUNCTIONS IN THIS MODULE:
 *
 * get_t_errno          - get the value of the errno variable for this Thread
 * set_t_errno          - set the value of the errno variable for this Thread
 * t_accept		- clean-up after listen()/accept
 * alloc_netbuf		- malloc a netbuf, and initialize it
 * t_alloc		- allocate a library structure
 * t_free		- free a structure
 * t_bind		- bind a socket address to the TPV
 * t_close		- close a TEP
 * t_connect		- connect to a another transport user
 * t_error		- print an error message to stdout
 * t_getinfo		- get information about TEP
 * t_getprotaddr	- malloc a netbuf, and initialize it
 * t_getstate		- get the current state of the TEP
 * t_listen		- listen for a connect indication
 * t_look		- look at the current event on a TEP
 * t_open		- establish a transport endpoint
 * t_optmgmt		- manage options for a transport endpoint
 * t_rcv		- receive data sent over a connection
 * t_snd		- send data sent over a connection
 * t_rcvconnect		- receive the confirmation from a connect request
 * t_rcvdis		- retrieve information from disconnect
 * t_rcvrel		- acknowledge receipt of an orderly release indication
 * t_rcvudata		- receive a data unit
 * t_sndudata		- send a data unit
 * t_rcvuderr		- receive a data unit error indication
 * t_snddis		- send user initiated disconnect request
 * t_sndrel		- initiate an orderly release
 * t_strerror		- produce an error message string
 * t_sync		- synchronize transport library
 * t_unbind		- disable a transport endpoint
 * FORK			- Fork a child process with its own thread
 *
 */

#define xti_c
#include <stdio.h>
#include <errno.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/syslog.h>
#include <fcntl.h>
#include <malloc.h>
#include <rpc/rpc.h>
#include <netinet/in.h>
#include <memory.h>
#include "config.h"
#include "TEP.h"
#include "Thread.h"
#include "Pod.h"
#include "xti.h"


/*
 * The number of valid xti library errors.
 */
#define XTI_NUM_ERRORS 30
int             t_nerr = XTI_NUM_ERRORS;

int             t_errno;

/*
 * error message array.
 */

char           *t_errlist[] =
{
    "",
    "incorrect addr format",
    "incorrect option format",
    "incorrect permissions",
    "illegal transport fd",
    "couldn't allocate addr",
    "out of state",
    "bad call sequence number",
    "system error",
    "event requires attention",
    "illegal amount of data",
    "buffer not large enough",
    "flow control",
    "no data",
    "discon_ind not found on queue",
    "unitdata error not found",
    "bad flags",
    "no ord rel found on queue",
    "primitive/action not supported",
    "state is in process of changing",
    "unsupported struct-type requested",
    "invalid transport provider name",
    "qlen is zero",
    "address in use",
    "outstanding connection indications",
    "transport provider mismatch",
    "resfd specified to accept w/qlen > 0",
    "resfd not bound to same addr as fd",
    "incoming connection queue full",
    "XTI protocol error"
};


/*
 * Function Name: get_t_errno
 * Description: get the value of the errno variable for this Thread.
 * Return values: Thread specific t_errno.
 */

int
get_t_errno(pid)
int             pid;
{

    return (t_errno);
}

/*
 * Function Name: set_t_errno
 * Description: set the value of the errno variable for this Thread.
 * Return values: Thread specific t_errno.
 */
 
int
set_t_errno(errnum)
int             errnum;
{
 
    t_errno = errnum;

    return (t_errno); 
}

/*
 * Function Name:   t_accept
 * Description: clean-up after listen()/accept.
 * Return values: 0 for success and -1 for error.
 */

int
t_accept(fd, resfd, call)
    int             fd;
    int             resfd;
    struct t_call  *call;
{
    TEP            *tep;
    int             result;

    /*
     * get the address of the TEP associated with the fd.
     */
    if ((tep = get_tep(fd)) > (TEP *) 0)
    {
	/*
	 * Send accept message to TEP instance 'tep'.
	 */
	result = accept_TEP(tep, resfd, call);
    } else
    {
	/*
	 * There isn't a TEP for this fd - so assume the fd is bad.
	 */
	result = (int) -1;
	t_errno = TBADF;
    }
    return (result);
}


/*
 * Function Name:   alloc_netbuf
 * Description: malloc a netbuf, and initialize it.
 * Return values: result of calloc.
 */

static char    *
alloc_netbuf(buffer, size, max)
    struct netbuf  *buffer;
    int             size;
    int             max;
{
    if (max != 0)
    {
	buffer->buf = (char *) calloc((unsigned) 1, (unsigned) size);
	if (buffer->buf)
	{
	    buffer->len = size;
	    buffer->maxlen = max;
	}
    }
    return (buffer->buf);
}


/*
 * Function Name: t_alloc
 * Description: allocate a library structure. netbuf size for T_ADDR is
 *		sizeof(sockaddr_in); size of address. netbuf size for
 * 		T_OPT is sizeof(int); getsockopt() returns int. netbuf
 *		size for T_UDATA is 0; connect/disconnect/tsdu not supported.
 * Return values: struct pointer, or (char*)0 if "calloc" fails.
 */

char           *
t_alloc(fd, struct_type, fields)
    int             fd;
    int             struct_type;
    int             fields;
{
    char           *result;
    int             sockaddr_size;
    int             bufsize;
    struct t_info   info;

    fd = fd;
    sockaddr_size = sizeof(struct sockaddr_in);
    switch (struct_type)
    {
    case T_BIND:
	result = (char *) calloc((unsigned) 1, (unsigned) sizeof(struct t_bind));
	if (result && ((fields & T_ADDR) || (fields & T_ALL)))
	{
	    alloc_netbuf(&((struct t_bind *) result)->addr, sockaddr_size,
			 sockaddr_size);
	}
	break;
    case T_CALL:
	result = (char *) calloc((unsigned) 1, (unsigned) sizeof(struct t_call));
	if (result && ((fields & T_ADDR) || (fields & T_ALL)))
	{
	    alloc_netbuf(&((struct t_call *) result)->addr, sockaddr_size,
			 sockaddr_size);
	}
	if (result && ((fields & T_OPT) || (fields & T_ALL)))
	{
	    alloc_netbuf(&((struct t_call *) result)->opt, sizeof(int),
			 sizeof(int));
	}
	if (result && ((fields & T_UDATA) || (fields & T_ALL)))
	{
	     /* This netbuf is never allocated because there is no unit data. */ ;
	}
	break;
    case T_OPTMGMT:
	result = (char *) calloc((unsigned) 1, (unsigned) sizeof(struct t_optmgmt));
	if (result && ((fields & T_OPT) || (fields & T_ALL)))
	{
	    alloc_netbuf(&((struct t_optmgmt *) result)->opt,
			 sizeof(struct tt_soopt),
			 sizeof(struct tt_soopt));
	}
	break;
    case T_DIS:
	result = (char *) calloc((unsigned) 1, (unsigned) sizeof(struct t_discon));
	if (result && ((fields & T_UDATA) || (fields & T_ALL)))
	{
	     /* This netbuf is never allocated because there is no unit data. */ ;
	}
	break;
    case T_UNITDATA:
	result = (char *) calloc((unsigned) 1, (unsigned) sizeof(struct t_unitdata));
	if (result && ((fields & T_ADDR) || (fields & T_ALL)))
	{
	    alloc_netbuf(&((struct t_unitdata *) result)->addr, sockaddr_size,
			 sockaddr_size);
	}
	if (result && ((fields & T_OPT) || (fields & T_ALL)))
	{
	    alloc_netbuf(&((struct t_call *) result)->opt, sizeof(int),
			 sizeof(int));
	}
	if (result && ((fields & T_UDATA) || (fields & T_ALL)))
	{
	    t_getinfo(fd, &info);
	    bufsize = info.tsdu;
	    alloc_netbuf(&((struct t_call *) result)->udata, bufsize, bufsize);
	}
	break;
    case T_UDERROR:
	result = (char *) calloc((unsigned) 1, (unsigned) sizeof(struct t_uderr));
	if (result && ((fields & T_ADDR) || (fields & T_ALL)))
	{
	    alloc_netbuf(&((struct t_uderr *) result)->addr, sockaddr_size,
			 sockaddr_size);
	}
	if (result && ((fields & T_OPT) || (fields & T_ALL)))
	{
	    alloc_netbuf(&((struct t_call *) result)->opt, sizeof(int),
			 sizeof(int));
	}
	break;
    case T_INFO:
	result = (char *) calloc((unsigned) 1, (unsigned) sizeof(struct t_info));
	break;
    default:
	t_errno = TNOSTRUCTYPE;
	result = (char *) -1;
    }
    return (result);
}


/*
 * Function Name:   t_free
 * Description: free a structure.
 * Return values: 0 for success and -1 for error.
 */

int
t_free(ptr, struct_type)
    char           *ptr;
    int             struct_type;
{
    int             result = 0;
    struct t_unitdata *unitdata_ptr;
    struct t_discon *discon_ptr;
    struct t_uderr *uderr_ptr;
    struct t_bind  *bind_ptr;
    struct t_call  *call_ptr;
    struct t_optmgmt *optmgmt_ptr;

    switch (struct_type)
    {
    case T_BIND:
	bind_ptr = (struct t_bind *) ptr;
	if (bind_ptr->addr.buf)
	{
	    free((char *) (bind_ptr->addr.buf));
	}
	break;
    case T_CALL:
	call_ptr = (struct t_call *) ptr;
	if (call_ptr->addr.buf)
	{
	    free((char *) (call_ptr->addr.buf));
	}
	if (call_ptr->opt.buf)
	{
	    free((char *) call_ptr->opt.buf);
	}
	if (call_ptr->udata.buf)
	{
	    free((char *) call_ptr->udata.buf);
	}
	break;
    case T_OPTMGMT:
	optmgmt_ptr = (struct t_optmgmt *) ptr;
	if (optmgmt_ptr->opt.buf)
	{
	    free((char *) optmgmt_ptr->opt.buf);
	}
	break;
    case T_DIS:
	discon_ptr = (struct t_discon *) ptr;
	if (discon_ptr->udata.buf)
	{
	    free((char *) discon_ptr->udata.buf);
	}
	break;
    case T_UNITDATA:
	unitdata_ptr = (struct t_unitdata *) ptr;
	if (unitdata_ptr->addr.buf)
	{
	    free((char *)unitdata_ptr->addr.buf);
	}
	if (unitdata_ptr->opt.buf)
	{
	    free((char *)unitdata_ptr->opt.buf);
	}
	if (unitdata_ptr->udata.buf)
	{
	    free((char *)unitdata_ptr->udata.buf);
	}
	break;
    case T_UDERROR:
	uderr_ptr = (struct t_uderr *) ptr;
	if (uderr_ptr->addr.buf)
	{
	    free((char *)uderr_ptr->addr.buf);
	}
	if (uderr_ptr->opt.buf)
	{
	    free((char *)uderr_ptr->opt.buf);
	}
	break;
    case T_INFO:
	break;
    default:
	t_errno = TNOSTRUCTYPE;
	return (-1);
	;
    }
    free((char *) ptr);
    return (result);
}


/*
 * Function Name:   t_bind
 * Description: bind a socket address to the TPV.
 * Return values: 0 for success and -1 for error.
 */

int
t_bind(fd, req, ret)
    int             fd;
    struct t_bind  *req;
    struct t_bind  *ret;
{
    TEP            *tep;
    int             result;

    /*
     * get the address of the TEP associated with the fd.
     */
    if ((tep = get_tep(fd)) > (TEP *) 0)
    {
	/*
	 * Send bind message to TEP instance 'tep'.
	 */
	result = bind_TEP(tep, req, ret);
    } else
    {
	/*
	 * There isn't a TEP for this fd - so assume the fd is bad.
	 */
	result = (int) -1;
	t_errno = TBADF;
    }
    return (result);
}


/*
 * Function Name:   t_close
 * Description: close a TEP.
 * Return values: 0 for success and -1 for error.
 */

int
t_close(fd)
    int             fd;
{
    TEP            *tep;
    int             result;

    /*
     * get the address of the TEP associated with the fd.
     */
    if ((tep = get_tep(fd)) > (TEP *) 0)
    {
	/*
	 * Send close message to TEP instance 'tep'.
	 */
	result = close_TEP(tep);
    } else
    {
	/*
	 * There isn't a TEP for this fd - so assume the fd is bad.
	 */
	result = (int) -1;
	t_errno = TBADF;
    }
    return (result);
}


/*
 * Function Name:   t_connect
 * Description: connect to a another transport user.
 * Return values: 0 for success and -1 for error.
 */

int
t_connect(fd, sndcall, rcvcall)
    int             fd;
    struct t_call  *sndcall;
    struct t_call  *rcvcall;
{
    TEP            *tep;
    int             result;

    /*
     * get the address of the TEP associated with the fd.
     */
    if ((tep = get_tep(fd)) > (TEP *) 0)
    {
	/*
	 * Send connect message to TEP instance 'tep'.
	 */
	result = connect_TEP(tep, sndcall, rcvcall);
    } else
    {
	/*
	 * There isn't a TEP for this fd - so assume the fd is bad.
	 */
	result = (int) -1;
	t_errno = TBADF;
    }
    return (result);
}


/*
 * Function Name:   t_error
 * Description: print an error message to stdout.
 * Return values: 0.
 */

int
t_error(errmsg)
    char           *errmsg;
{
    int             result = 0;

    if (strlen(errmsg))
    {
	fprintf(stderr, "%s: ", errmsg);
    }
    if (t_errno != TSYSERR)
    {
	fprintf(stderr, "\n");
    } else if (errno <= t_nerr)
    {
	fprintf(stderr, "%s\n", t_errlist[errno]);
    }
    return (result);
}


/*
 * Function Name:   t_getinfo
 * Description: get information about TEP.
 * Return values: if the fd is good, fd is returned, -1 for error.
 */

int
t_getinfo(fd, info)
    int             fd;
    struct t_info  *info;
{
    TEP            *tep;
    int             result;

    /*
     * get the address of the TEP associated with the fd.
     */
    if ((tep = get_tep(fd)) > (TEP *) 0)
    {
	/*
	 * Send getinfo message to TEP instance 'tep'.
	 */
	result = getinfo_TEP(tep, info);
    } else
    {
	/*
	 * There isn't a TEP for this fd - so assume the fd is bad.
	 */
	result = (int) -1;
	t_errno = TBADF;
    }
    return (result);
}


/*
 * Function Name:   t_getprotaddr
 * Description: malloc a netbuf, and initialize it.
 * Return values: 0 for success and -1 for error.
 */

int
t_getprotaddr(fd, boundaddr, peeraddr)
    int             fd;
    struct t_bind  *boundaddr;
    struct t_bind  *peeraddr;
{
    TEP            *tep;
    int             result;

    /*
     * get the address of the TEP associated with the fd.
     */
    if ((tep = get_tep(fd)) > (TEP *) 0)
    {
	/*
	 * Send getprotaddr message to TEP instance 'tep'.
	 */
	result = getprotaddr_TEP(tep, boundaddr, peeraddr);
    } else
    {
	/*
	 * There isn't a TEP for this fd - so assume the fd is bad.
	 */
	result = (int) -1;
	t_errno = TBADF;
    }
    return (result);
}


/*
 * Function Name:   t_getstate
 * Description: get the current state of the TEP.
 * Return values: state for success and -1 for error.
 */

int
t_getstate(fd)
    int             fd;
{
    TEP            *tep;
    int             result;

    /*
     * get the address of the TEP associated with the fd.
     */
    if ((tep = get_tep(fd)) > (TEP *) 0)
    {
	/*
	 * Send close getstate to TEP instance 'tep'.
	 */
	result = getstate_TEP(tep);
    } else
    {
	/*
	 * There isn't a TEP for this fd - so assume the fd is bad.
	 */
	result = (int) -1;
	t_errno = TBADF;
    }
    return (result);
}


/*
 * Function Name:   t_listen
 * Description: listen for a connect indication.
 * Return values: 0 for success and -1 for error.
 */

int
t_listen(fd, call)
    int             fd;
    struct t_call  *call;
{
    TEP            *tep;
    int             result;

    /*
     * get the address of the TEP associated with the fd.
     */
    if ((tep = get_tep(fd)) > (TEP *) 0)
    {
	/*
	 * Send listen message to TEP instance 'tep'.
	 */
	result = listen_TEP(tep, call);
    } else
    {
	/*
	 * There isn't a TEP for this fd - so assume the fd is bad.
	 */
	result = (int) -1;
	t_errno = TBADF;
    }
    return (result);
}


/*
 * Function Name:   t_look
 * Description: look at the current event on a TEP.
 * Return values: event for success and -1 for error.
 */

int
t_look(fd)
    int             fd;
{
    TEP            *tep;
    int             result;

    /*
     * get the address of the TEP associated with the fd.
     */
    if ((tep = get_tep(fd)) > (TEP *) 0)
    {
	/*
	 * Send look message to TEP instance 'tep'.
	 */
	result = look_TEP(tep);
    } else
    {
	/*
	 * There isn't a TEP for this fd - so assume the fd is bad.
	 */
	result = (int) -1;
	t_errno = TBADF;
    }
    return (result);
}


/*
 * Function Name:   t_open
 * Description: establish a transport endpoint.
 * Return values: fd for success and -1 for error.
 */

int
t_open(name, oflag, info)
    char           *name;
    int             oflag;
    struct t_info  *info;
{
    Thread         *thread = addr_Thread(GET_ADDR);
    int             newFd = -1;
    TEP             localTEP;
    TEP            *tep;
    TEP            *local_tep;


    if (!strncmp("/dev/tcp", name, 8) || !strncmp("/dev/udp", name, 8) ||
	    !strncmp("/dev/lodg", name, 9))
    {
	if (thread != (Thread *)NULL)
	{
	    /*
	     * Let's allocate a tep.
	     */
	    local_tep = &localTEP;
	    /*
	     * We created a tep, now initialize it and open the socket.
	     */
	    if (init_TEP(local_tep, name, 255) > (TEP *) 0)
	    {
		if ((newFd = open_TEP(local_tep, oflag, info)) >= 0)
		{
#ifdef XTI_DEBUG
		fprintf(stderr, "t_open: fd %d opened in thread\n",
				newFd);
#endif
		    tep = &(thread->teps[newFd]);
		    memcpy((char *) tep, (char *) local_tep, sizeof(TEP));
		} else
		{
		    newFd = -1;
		    t_errno = TSYSERR;
		}
	    } else
	    {
		newFd = -1;
		t_errno = TBADNAME;
	    }
	}
    }
    /*
     * We've got a bad TPV name.
     */
    else
    {
	t_errno = TBADNAME;
    }
return (newFd);
}


/*
 * Function Name:   t_optmgmt
 * Description: manage options for a transport endpoint.
 * Return values: 0 for success and -1 for error.
 */

int
t_optmgmt(fd, req, ret)
    int             fd;
    struct t_optmgmt *req;
    struct t_optmgmt *ret;
{
    TEP            *tep;
    int             result;

    /*
     * get the address of the TEP associated with the fd.
     */
    if ((tep = get_tep(fd)) > (TEP *) 0)
    {
	/*
	 * Send optmgmt message to TEP instance 'tep'.
	 */
	result = optmgmt_TEP(tep, req, ret);
    } else
    {
	/*
	 * There isn't a TEP for this fd - so assume the fd is bad.
	 */
	result = (int) -1;
	t_errno = TBADF;
    }
    return (result);
}


/*
 * Function Name:   t_rcv
 * Description: receive data sent over a connection.
 * Return values: number of bytes received for success and -1 for error.
 */

int
t_rcv(fd, buf, nbytes, flags)
    int             fd;
    char           *buf;
    unsigned int    nbytes;
    int            *flags;
{
    TEP            *tep;
    int             result;

    /*
     * get the address of the TEP associated with the fd.
     */
    if ((tep = get_tep(fd)) > (TEP *) 0)
    {
	/*
	 * Send close rcv to TEP instance 'tep'.
	 */
	result = rcv_TEP(tep, buf, nbytes, flags);
    } else
    {
	/*
	 * There isn't a TEP for this fd - so assume the fd is bad.
	 */
	result = (int) -1;
	t_errno = TBADF;
    }
    return (result);
}


/*
 * Function Name:   t_snd
 * Description: send data sent over a connection.
 * Return values: number of bytes sent for success and -1 for error.
 */

int
t_snd(fd, buf, nbytes, flags)
    int             fd;
    char           *buf;
    unsigned int    nbytes;
    int             flags;
{
    TEP            *tep;
    int             result;

    /*
     * get the address of the TEP associated with the fd.
     */
    if ((tep = get_tep(fd)) > (TEP *) 0)
    {
	/*
	 * Send snd message to TEP instance 'tep'.
	 */
	result = snd_TEP(tep, buf, nbytes, flags);
    } else
    {
	/*
	 * There isn't a TEP for this fd - so assume the fd is bad.
	 */
	result = (int) -1;
	t_errno = TBADF;
    }
    return (result);
}


/*
 * Function Name:   t_rcvconnect
 * Description: receive the confirmation from a connect request.
 * Return values: 0 for success and -1 for error.
 */

int
t_rcvconnect(fd, call)
    int             fd;
    struct t_call  *call;
{
    TEP            *tep;
    int             result;

    /*
     * get the address of the TEP associated with the fd.
     */
    if ((tep = get_tep(fd)) > (TEP *) 0)
    {
	/*
	 * Send revconnect message to TEP instance 'tep'.
	 */
	result = rcvconnect_TEP(tep, call);
    } else
    {
	/*
	 * There isn't a TEP for this fd - so assume the fd is bad.
	 */
	result = (int) -1;
	t_errno = TBADF;
    }
    return (result);
}


/*
 * Function Name:   t_rcvdis
 * Description: retrieve information from disconnect.
 * Return values: 0 for success and -1 for error.
 */

int
t_rcvdis(fd, discon)
    int             fd;
    struct t_discon *discon;
{
    TEP            *tep;
    int             result;

    /*
     * get the address of the TEP associated with the fd.
     */
    if ((tep = get_tep(fd)) > (TEP *) 0)
    {
	/*
	 * Send rcvdis message to TEP instance 'tep'.
	 */
	result = rcvdis_TEP(tep, discon);
    } else
    {
	/*
	 * There isn't a TEP for this fd - so assume the fd is bad.
	 */
	result = (int) -1;
	t_errno = TBADF;
    }
    return (result);
}


/*
 * Function Name:   t_rcvrel
 * Description: acknowledge receipt of an orderly release indication.
 * Return values: 0 for success and -1 for error.
 */

int
t_rcvrel(fd)
    int             fd;
{
    TEP            *tep;
    int             result;

    /*
     * get the address of the TEP associated with the fd.
     */
    if ((tep = get_tep(fd)) > (TEP *) 0)
    {
	/*
	 * Send rcvrel message to TEP instance 'tep'.
	 */
	result = rcvrel_TEP(tep);
    } else
    {
	/*
	 * There isn't a TEP for this fd - so assume the fd is bad.
	 */
	result = (int) -1;
	t_errno = TBADF;
    }
    return (result);
}


/*
 * Function Name:   t_rcvudata
 * Description: receive a data unit.
 * Return values: number of bytes received for success and -1 for error.
 */

int
t_rcvudata(fd, unitdata, flags)
    int             fd;
    struct t_unitdata *unitdata;
    int             flags;
{
    TEP            *tep;
    int             result;

    /*
     * get the address of the TEP associated with the fd.
     */
    if ((tep = get_tep(fd)) > (TEP *) 0)
    {
	/*
	 * Send rcvudata message to TEP instance 'tep'.
	 */
	result = rcvudata_TEP(tep, unitdata, flags);
    } else
    {
	/*
	 * There isn't a TEP for this fd - so assume the fd is bad.
	 */
	result = (int) -1;
	t_errno = TBADF;
    }
    return (result);
}


/*
 * Function Name:   t_sndudata
 * Description: send a data unit.
 * Return values: number of bytes sent for success and -1 for error.
 */

int
t_sndudata(fd, unitdata)
    int             fd;
    struct t_unitdata *unitdata;
{
    TEP            *tep;
    int             result;

    /*
     * get the address of the TEP associated with the fd.
     */
    if ((tep = get_tep(fd)) > (TEP *) 0)
    {
	/*
	 * Send sndudata message to TEP instance 'tep'.
	 */
	result = sndudata_TEP(tep, unitdata);
    } else
    {
	/*
	 * There isn't a TEP for this fd - so assume the fd is bad.
	 */
	result = (int) -1;
	t_errno = TBADF;
    }
    return (result);
}


/*
 * Function Name:   t_rcvuderr
 * Description: receive a data unit error indication.
 * Return values: 0 for success and -1 for error.
 */

int
t_rcvuderr(fd, uderr)
    int             fd;
    struct t_uderr *uderr;
{
    TEP            *tep;
    int             result;

    /*
     * get the address of the TEP associated with the fd.
     */
    if ((tep = get_tep(fd)) > (TEP *) 0)
    {
	/*
	 * Send rcvuderr message to TEP instance 'tep'.
	 */
	result = rcvuderr_TEP(tep, uderr);
    } else
    {
	/*
	 * There isn't a TEP for this fd - so assume the fd is bad.
	 */
	result = (int) -1;
	t_errno = TBADF;
    }
    return (result);
}


/*
 * Function Name:   t_snddis
 * Description: send user initiated disconnect request.
 * Return values: 0 for success and -1 for error.
 */

int
t_snddis(fd, call)
    int             fd;
    struct t_call  *call;
{
    TEP            *tep;
    int             result;

    /*
     * get the address of the TEP associated with the fd.
     */
    if ((tep = get_tep(fd)) > (TEP *) 0)
    {
	/*
	 * Send snddis message to TEP instance 'tep'.
	 */
	result = snddis_TEP(tep, call);
    } else
    {
	/*
	 * There isn't a TEP for this fd - so assume the fd is bad.
	 */
	result = (int) -1;
	t_errno = TBADF;
    }
    return (result);
}


/*
 * Function Name:   t_sndrel
 * Description: initiate an orderly release.
 * Return values: 0 for success and -1 for error.
 */

int
t_sndrel(fd)
    int             fd;
{
    TEP            *tep;
    int             result;

    /*
     * get the address of the TEP associated with the fd.
     */
    if ((tep = get_tep(fd)) > (TEP *) 0)
    {
	/*
	 * Send sndrel message to TEP instance 'tep'.
	 */
	result = sndrel_TEP(tep);
    } else
    {
	/*
	 * There isn't a TEP for this fd - so assume the fd is bad.
	 */
	result = (int) -1;
	t_errno = TBADF;
    }
    return (result);
}


/*
 * Function Name:   t_strerror
 * Description: produce an error message string.
 * Return values: pointer to error message for success and -1 for error.
 */

char           *
t_strerror(errnum)
    int             errnum;
{
    char            msg[80];

    if (errnum <= t_nerr)
    {
	sprintf(msg, "%d: %s", errnum, t_errlist[errnum]);
    } else
    {
	sprintf(msg, "%d: %s", errnum, "error unknown");
    }
    return (&msg[0]);
}


/*
 * Function Name:   t_sync
 * Description: synchronise transport library.
 * Return values: state of tep for success and -1 for error.
 */

int
t_sync(fd)
    int             fd;
{
    TEP            *tep;
    int             result;

    if ((tep = get_tep(fd)) > (TEP *) 0)
    {
	/*
	 * Send sync message to TEP instance 'tep'.
	 */
	result = sync_TEP(tep);
    } else
    {
	/*
	 * There isn't a TEP for this fd - so assume the fd is bad.
	 */
	result = (int) -1;
	t_errno = TBADF;
    }
    return (result);
}


/*
 * Function Name:   t_unbind
 * Description: disable a transport endpoint.
 * Return values: 0 for success and -1 for error.
 */

int
t_unbind(fd)
    int             fd;
{
    TEP            *tep;
    int             result;

    /*
     * get the address of the TEP associated with the fd.
     */
    if ((tep = get_tep(fd)) > (TEP *) 0)
    {
	/*
	 * Send unbind message to TEP instance 'tep'.
	 */
	result = unbind_TEP(tep);
    } else
    {
	/*
	 * There isn't a TEP for this fd - so assume the fd is bad.
	 */
	result = (int) -1;
	t_errno = TBADF;
    }
    return (result);
}


/*
 * Function Name: FORK
 * Description: Fork a child process with its own thread (a copy of its parent's
 *		thread).
 * Return values: If parent, pid of child or -1 if error.
 *		  If child, 0.
 */

int
FORK()
{
    Thread         *parent_thread = addr_Thread(GET_ADDR);
    Thread	   *thread = (Thread *) 0;
    int             new_pid;
    int             inx;

    if (parent_thread != (Thread *)NULL)
    {
	/*
	 * Let's allocate a thread for the child. Having allocated the thread
	 * - let's initialize it.
	 */
	if ((thread = alloc_Thread()) > (Thread *) 0)
	{
	    memcpy((char *) thread, (char *) parent_thread, sizeof(Thread));
	    /* reset the thread index for this thread */
	    /* and clear any TEPs on which the parent is listening with
	       a queue length greater than 0 */
	    for (inx = 0; inx < XTI_NOFILE; inx++)
	    {
		if((&(thread->teps[inx]))->state == T_INCON
		    && (&(thread->teps[inx]))->event == T_LISTEN
		    && (&(thread->teps[inx]))->qlen > 0)
		{
		    /* the parent is listening on this TEP, don't pass it
		       to the child (tli compatability) */
		    (void) init_TEP(&(thread->teps[inx]), "", 255);
		}
	    }
	} else
	{
	    /* We couldn't allocate a Thread ... now what? */
	    /* The child can't inherit the parent's TEPs */
	    syslog(LOG_ERR, "couldn't alloc thread for child in fork, pid: %d",
		   getpid());
	    fprintf(stderr, "couldn't alloc thread for child in fork, pid: %d\n",
		   getpid());
	}
    }
    switch(new_pid = fork())
    {
    case 0:
	{
	    /* this is the child */
	    Pod *pod;

	    /* find an unused thread in the pod and copy the child's thread to it */
	    /* this way, if there is an exec the child can locate its thread */
	    if(lock_Pod(LOCK_POD) < 0)
	    {
		return(-1);
	    }
	    if((pod = addr_Pod(GET_ADDR)) == (Pod *)NULL)
	    {
		return(-1);
	    }
	    for(inx = 0; inx < XTI_MAX_THREADS; ++inx)
	    {
		if(pod->threads[inx].active == 1)
		{
		    /* check to see if there is a process for this thread */
		    int kill_return = kill(pod->threads[inx].pid, 0);

		    if (kill_return == -1 && errno == ESRCH)
		    {
			/* no such process */
			/* clear the active flag and use this thread */
		        pod->threads[inx].active = 0;
		    }
		}
		if(pod->threads[inx].active == 0)
		{
		    memcpy((char *) (&(pod->threads[inx])),
			   (char *) thread, sizeof(Thread));
		    pod->threads[inx].active = 1;
		    pod->threads[inx].pid = getpid();
		    addr_Thread(FREE_ADDR);
		    break;
		}
	    }
	    if(lock_Pod(UNLOCK_POD) < 0)
	    {
		return(-1);
	    }
	    if(addr_Pod(FREE_ADDR) == (Pod *)NULL)
	    {
		return(-1);
	    }

	    if(inx == XTI_MAX_THREADS)
	    {
		/* there are too many processes forking at once */
		return(-1);
	    }

#ifdef XTI_DEBUG
	    fprintf(stderr, "child process %d: thread written to pod\n", new_pid);
#endif

	    return (0);
	}

    case -1:
	(void)free((char *)thread);
	return (-1);

    default:
	/* This is the parent. */
	(void)free((char *)thread);
	return (new_pid);
    }
}
