/* 
 * Copyright (c) 1987, 1988 
 *	Carnegie-Mellon University.   All rights reserved.
 *
 * This code is derived from software contributed by Carnegie-Mellon 
 * University from the Mach Operating System.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	@(#)pmap.c	2.10 (CMU) 12/20/88
 */

/*
 *	Manages physical address maps.
 *
 *	In addition to hardware address maps, this
 *	module is called upon to provide software-use-only
 *	maps which may or may not be stored in the same
 *	form as hardware maps.  These pseudo-maps are
 *	used to store intermediate results from copy
 *	operations to and from address spaces.
 *
 *	Since the information managed by this module is
 *	also stored by the logical address mapping module,
 *	this module may throw away valid virtual-to-physical
 *	mappings at almost any time.  However, invalidations
 *	of virtual-to-physical mappings must be done as
 *	requested.
 *
 *	In order to cope with hardware architectures which
 *	make virtual-to-physical map invalidates expensive,
 *	this module may delay invalidate or reduced protection
 *	operations until such time as they are actually
 *	necessary.  This module is given full information as
 *	to which processors are currently using which maps,
 *	and to when physical maps must be made correct.
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/proc.h>
#ifndef	SIMULATOR
#include <sys/malloc.h>
#endif
#include <sys/user.h>

#include <vm/vm.h>
#include <vm/vm_kern.h>
#include <vm/vm_page.h>

#include <machine/cpu.h>

			/* XXX use Mach version of linked lists for now */
#include "rt/rt/mach_queue.h"

#ifdef PMAPSTATS
struct {
	int collectscans;
	int collectpages;
	int kpttotal;
	int kptinuse;
	int kptmaxuse;
} kpt_stats;
struct {
	int kernel;	/* entering kernel mapping */
	int user;	/* entering user mapping */
	int ptpneeded;	/* needed to allocate a PT page */
	int nochange;	/* no change at all */
	int pwchange;	/* no mapping change, just wiring or protection */
	int wchange;	/* no mapping change, just wiring */
	int pchange;	/* no mapping change, just protection */
	int mchange;	/* was mapped but mapping to different page */
	int managed;	/* a managed page */
	int firstpv;	/* first mapping for this PA */
	int secondpv;	/* second mapping for this PA */
	int ci;		/* cache inhibited */
	int unmanaged;	/* not a managed page */
	int flushes;	/* cache flushes */
} enter_stats;
struct {
	int calls;
	int removes;
	int pvfirst;
	int pvsearch;
	int ptinvalid;
	int uflushes;
	int sflushes;
} remove_stats;
struct {
	int calls;
	int changed;
	int alreadyro;
	int alreadyrw;
} protect_stats;
struct chgstats {
	int setcalls;
	int sethits;
	int setmiss;
	int clrcalls;
	int clrhits;
	int clrmiss;
} changebit_stats[16];
#endif

#ifdef DEBUG
int debugmap = 0;
int pmapdebug = 0x2000;
#define PDB_FOLLOW	0x0001
#define PDB_INIT	0x0002
#define PDB_ENTER	0x0004
#define PDB_REMOVE	0x0008
#define PDB_CREATE	0x0010
#define PDB_PTPAGE	0x0020
#define PDB_CACHE	0x0040
#define PDB_BITS	0x0080
#define PDB_COLLECT	0x0100
#define PDB_PROTECT	0x0200
#define PDB_SEGTAB	0x0400
#define PDB_MULTIMAP	0x0800
#define PDB_PARANOIA	0x2000
#define PDB_WIRING	0x4000
#define PDB_PVDUMP	0x8000
#define	DDISPLAY(x)
#else
#define	DDISPLAY(x)
#endif

struct pmap	kernel_pmap_store;
vm_map_t	st_map, pt_map;

vm_offset_t    	avail_start;	/* PA of first available physical page */
vm_offset_t	avail_end;	/* PA of last available physical page */
vm_size_t	mem_size;	/* memory size in bytes */
vm_offset_t	virtual_avail;  /* VA of first avail page (after kernel bss)*/
vm_offset_t	virtual_end;	/* VA of last avail page (end of kernel AS) */
vm_offset_t	vm_first_phys;	/* PA of first managed page */
vm_offset_t	vm_last_phys;	/* PA just past last managed page */
boolean_t	pmap_initialized = FALSE;	/* Has pmap_init completed? */
char		*pmap_attributes;	/* reference and modify bits */

int		ppages_per_vmpage=2;	/* 2K pages physical, 4K vm page */
int		rose_page_size=2048;	/* 2K page size */
int		rose_page_shift=11;

lock_data_t	pmap_lock;	/* read/write lock for pmap system */
pmap_t		active_pmap;	/* XXX should be an array for multiprocessors*/
struct pmap_list_entry *pmap_list_entries; /* one ppage_list_entry per page */
extern unsigned char	*is_in;	/* bitvector for physical pages (in locore) */
vm_offset_t	min_user_pa;	/* Minimum user physical page */
extern int	holestart, holeEnd; /* start and end of the memory hole */

/*
 *	Bit array for in-use segments. 
 */
unsigned char 	pmap_seg_tab[(ROMP_NSEGS + 7) / 8];
short		max_free_seg = ROMP_NSEGS - 1;

#define	HACK_SEG 7
/*
 * Note to future hackers: Several places in this code we want to have rosetta
 * do a virtual address translation on an address which isn't in the currently
 * active pmap.  To do this, we munge the segment register represented by
 * HACK_SEG and then coerce the virtual address to be translated into this
 * segment.  If you should, for some strange reason, decide to change the
 * value of HACK_SEG, you should not use segments e (where the kernel lives),
 * f (which is used for IO), and 0 (in case virtual = real is turned on).
 */

/*
 *	Macro:		rtakey
 *	Function:
 *		Map machine-independent VM protection to MMU access key.
 *	Note:
 *		Write capability implies read capability; execute capability
 *		is ignored.  The Mach interface allows these reductions.
 *
 *		Total lack of capability (VM_PROT_NONE) must be handled
 *		separately, as no access code applies.  The page in question
 *		must not be mapped at all.
 *
 *		To make memory read-only to the kernel, it must be readable
 *		by user programs.  This makes it impossible to send kernel
 *		memory copy-on-write from a segment that is accessible by
 *		users.  [XXX THIS PRECAUTION IS NOT CURRENTLY TAKEN.]
 */
#define	rtakey(map, prot)						\
		(((prot) & VM_PROT_WRITE) ?				\
		 (((map) == kernel_pmap) ? ROMP_KEY_KW : ROMP_KEY_UW) :	\
		 ROMP_KEY_URKR)						

/*
 * Declarations of subroutines specific to the RT implementation
 */
       void	pmap_bootstrap __P((vm_offset_t, vm_offset_t *, vm_offset_t *,
			vm_offset_t *, vm_offset_t *));
static void	pmap_make_kernel_text_readable __P(());
static void	pmap_copy_on_write __P((vm_offset_t));
static void	pmap_pr_phys_range __P((int, int, int, vm_offset_t,
			vm_offset_t, pmap_t, int));
static void	pmap_pr_phys __P((pmap_t, vm_offset_t, vm_offset_t, int));
       void	pmap_protect_phys_page __P((vm_offset_t,int));
static short	pmap_alloc_seg __P(());
static void	pmap_dealloc_seg __P((short));
static void	pmap_remove_all __P((vm_offset_t));
static void	pmap_init_seg_alloc __P(());
static void	pmap_add_ppage_entry __P((struct pmap *, vm_offset_t,
			vm_offset_t));
static void	pmap_remove_ppage_entry __P((vm_offset_t));
static long	trans_virt_addr __P((vm_offset_t));
static int	virt_eq_real __P((int));
       void	mapout __P((vm_offset_t, int));
       void	mapin __P((vm_offset_t, vm_offset_t, int, int, int));
       void	remap __P((int, vm_offset_t, vm_offset_t, int));

#ifdef	ROMP_CACHE

#define	pmap_cache_init(ls,av)						\
				cache_init(ls,av)
#define	pmap_cache_remove(pmap,sid,page)				\
				cache_remove(pmap,sid,page)
#define	pmap_cache_remove_range(pmap,s,e) 				\
				cache_remove_range(pmap,s,e)
#define	pmap_cache_protect_range(pmap,s,e,prot) 			\
				cache_protect_range(pmap,s,e,prot)
#define	pmap_cache_remove_page(addr)					\
				cache_remove_page(addr)
#define	pmap_cache_enter(pmap,sid,vp,pp,prot)				\
				cache_enter(pmap,sid,vp,pp,prot)

#else

#define	pmap_cache_init(ls,av)
#define	pmap_cache_remove(pmap,sid,page)		
#define	pmap_cache_remove_range(pmap,s,e)
#define	pmap_cache_protect_range(pmap,s,e,prot) 
#define	pmap_cache_remove_page(addr)		
#define	pmap_cache_enter(pmap,sid,vp,pp,prot)	

#endif	/* ROMP_CACHE */

/*
 * Bootstrap memory allocator. This function allows for early dynamic
 * memory allocation until the virtual memory system has been bootstrapped.
 * After that point, either kmem_alloc or malloc should be used. This
 * function works by stealing pages from the (to be) managed page pool,
 * stealing virtual address space, then mapping the pages and zeroing them.
 *
 * It should be used from pmap_bootstrap until vm_page_startup, afterwards
 * it cannot be used, and will generate a panic if tried. Note that this
 * memory will never be freed, and in essence it is wired down.
 */
void *
pmap_bootstrap_alloc(size)
	int size;
{
}

/*
 *	Initialize the pmap module.
 *	Called by vm_init, to initialize any structures that the pmap
 *	system needs to map virtual memory.
 */
void
pmap_init(phys_start, phys_end)
	vm_offset_t	phys_start, phys_end;
{
	/*
	 * Don't need to do much really, pmap_bootstrap() does most of it.
	 * Just make the kernel text segment readable by all maps.
	 */
	pmap_make_kernel_text_readable();
}

/*
 *	Used to map a range of physical addresses into kernel
 *	virtual address space.
 *
 *	For now, VM is already on, we only need to map the
 *	specified memory.
 */
vm_offset_t
pmap_map(virt, start, end, prot)
	vm_offset_t	virt;
	vm_offset_t	start;
	vm_offset_t	end;
	int		prot;
{
#ifdef DEBUG
	if (pmapdebug & PDB_FOLLOW)
		printf("pmap_map(%x, %x, %x, %x)\n", virt, start, end, prot);
#endif
	while (start < end) {
		pmap_enter(kernel_pmap, virt, start, prot, FALSE);
		virt += PAGE_SIZE;
		start += PAGE_SIZE;
	}
	return(virt);
}

/*
 *	Create and return a physical map.
 *
 *	If the size specified for the map
 *	is zero, the map is an actual physical
 *	map, and may be referenced by the
 *	hardware.
 *
 *	If the size specified is non-zero,
 *	the map will be used in software only, and
 *	is bounded by that size.
 */
pmap_t
pmap_create(size)
	vm_size_t	size;
{
	register pmap_t pmap;

#ifdef DEBUG
	if (pmapdebug & (PDB_FOLLOW|PDB_CREATE))
		printf("pmap_create(%x)\n", size);
#endif
	/*
	 * Software use map does not need a pmap
	 */
	if (size)
		return(NULL);

#ifndef	SIMULATOR
	/* XXX: is it ok to wait here? */
	pmap = (pmap_t) malloc(sizeof *pmap, M_VMPMAP, M_WAITOK);
#else
	pmap = (pmap_t) malloc(sizeof *pmap);
#endif
#ifdef notifwewait
	if (pmap == NULL)
		panic("pmap_create: cannot allocate a pmap");
#endif
	bzero(pmap, sizeof(*pmap));
	pmap_pinit(pmap);
	return (pmap);
}

/*
 * Initialize a preallocated and zeroed pmap structure,
 * such as one in a vmspace structure.
 */
void
pmap_pinit(pmap)
	register struct pmap *pmap;
{
	int			index;
	pmap_statistics_t	stats;

#ifdef DEBUG
	if (pmapdebug & (PDB_FOLLOW|PDB_CREATE))
		printf("pmap_pinit(%x)\n", pmap);
#endif
	/*
	 *	Initialize the sidtab.
	 */
	for (index = 0; index < NUSR_SEGS; index++) {
		pmap->sidtab[index] = 0;
	}

	pmap->ref_count = 1;

	simple_lock_init(&pmap->lock);

	/*
	 *	Initialize statistics.
	 */

	stats = &pmap->stats;
	stats->resident_count = 0;
	stats->wired_count = 0;

	queue_init(&pmap->ppage_list);
	queue_init(&pmap->cache_entries);
	pmap->ppage_count = 0;
	pmap->cache_count = 0;
}

/*
 *	Retire the given physical map from service.
 *	Should only be called if the map contains
 *	no valid mappings.
 */
void
pmap_destroy(pmap)
	register pmap_t pmap;
{
	int count;

#ifdef DEBUG
	if (pmapdebug & PDB_FOLLOW)
		printf("pmap_destroy(%x)\n", pmap);
#endif
	if (pmap == NULL)
		return;

	simple_lock(&pmap->lock);
	count = --pmap->ref_count;
	simple_unlock(&pmap->lock);
	if (count == 0) {
		pmap_release(pmap);
#ifndef	SIMULATOR
		free((caddr_t)pmap, M_VMPMAP);
#else
		free((caddr_t)pmap);
#endif
	}
}

/*
 * Release any resources held by the given physical map.
 * Called when a pmap initialized by pmap_pinit is being released.
 * Should only be called if the map contains no valid mappings.
 */
void
pmap_release(pmap)
	register struct pmap *pmap;
{
	int pl;
	int index;

#ifdef DEBUG
	if (pmapdebug & PDB_FOLLOW)
		printf("pmap_release(%x)\n", pmap);
#endif
	if (pmap->ppage_count != 0)
		panic("pmap_destroy: ppage count not zero");
	if (pmap->cache_count != 0)
		panic("pmap_destroy: cache count not zero");
	pl = splvm();
	for (index = 0; index < NUSR_SEGS; index++) 
		if (pmap->sidtab[index]) 
			pmap_dealloc_seg(pmap->sidtab[index]);
	splx(pl);
}

/*
 *	Add a reference to the specified pmap.
 */
void
pmap_reference(pmap)
	pmap_t	pmap;
{
#ifdef DEBUG
	if (pmapdebug & PDB_FOLLOW)
		printf("pmap_reference(%x)\n", pmap);
#endif
	if (pmap != NULL) {
		simple_lock(&pmap->lock);
		pmap->ref_count++;
		simple_unlock(&pmap->lock);
	}
}

/*
 *	Remove the given range of addresses from the specified map.
 *
 *	It is assumed that the start and end are properly
 *	rounded to the page size.
 */
void
pmap_remove(pmap, sva, eva)
	register pmap_t pmap;
	register vm_offset_t sva, eva;
{
	queue_head_t *list;
	struct pmap_list_entry *entry;
	pmap_list_entry_t next_entry;
	int spl;

#ifdef DEBUG
	if (pmapdebug & (PDB_FOLLOW|PDB_REMOVE|PDB_PROTECT))
		printf("pmap_remove(%x, %x, %x)\n", pmap, sva, eva);
#endif

	if (pmap == NULL)
		return;

#ifdef PMAPSTATS
	remove_stats.calls++;
#endif
	spl = splvm();
    	if (atoseg(sva) != SYS_SEG)
		pmap_cache_remove_range(pmap, sva, eva-1);
	
	list = &pmap->ppage_list;
	entry = (struct pmap_list_entry *) queue_first(list);
	while (!queue_end(list,(queue_entry_t) entry)) {
	    next_entry = (pmap_list_entry_t) queue_next(&entry->pmap_pages);
	    if ((entry->vaddr >= sva) && (entry->vaddr < eva)) {
		/*
		 * remove it from the hatipt
		 * 	(might not be there since the cache fault
		 *	 handler could have remapped the page)
		 * remove it from the TLBs,
		 * remove it from the pmap ppage list,
		 * and then free it.
		 */
#ifdef	notdef
		/* the following should work, plus it should be faster */
		/* it needs the mmu_extract_addrtag macro added to pmap.h*/
		
		if (mmu_extract_addrtag(a_to_rosetta_page(entry->paddr)) ==
			( pmap->sidtab[atoseg(entry->vaddr)] << ROMP_VPAGE_BITS
			  || a_to_rosetta_page(segoffset(entry->vaddr))))
#endif	notdef		    
		if (pmap_extract(pmap,entry->vaddr) != 0) {

		    mapout(a_to_rosetta_page(entry->paddr),ppages_per_vmpage);
		    tlb_invalidate_seg(atoseg(entry->vaddr));
		}
		queue_remove(list,entry,pmap_list_entry_t, pmap_pages);
		pmap->ppage_count--;
		entry->pmap = PMAP_NULL;
		entry->in_use = FALSE;
	    }
	    entry = next_entry;
	}
	splx(spl);
}

/*
 *	pmap_page_protect:
 *
 *	Lower the permission for all mappings to a given page.
 */
void
pmap_page_protect(pa, prot)
	vm_offset_t	pa;
	vm_prot_t	prot;
{
#ifdef DEBUG
	if ((pmapdebug & (PDB_FOLLOW|PDB_PROTECT)) ||
	    prot == VM_PROT_NONE && (pmapdebug & PDB_REMOVE))
		printf("pmap_page_protect(%x, %x)\n", pa, prot);
#endif
	if (pa < vm_first_phys || pa >= vm_last_phys)
		return;
	switch (prot) {
		case VM_PROT_READ:
		case VM_PROT_READ|VM_PROT_EXECUTE:
			pmap_copy_on_write(pa);
			break;
		case VM_PROT_ALL:
			break;
		default:
			pmap_remove_all(pa);
			break;
	}
}

/*
 *	Set the physical protection on the
 *	specified range of this map as requested.
 */
void
pmap_protect(pmap, sva, eva, prot)
	register pmap_t	pmap;
	register vm_offset_t sva, eva;
	vm_prot_t prot;
{
	queue_head_t *list;
	struct pmap_list_entry *entry;
	int mmu_prot = rtakey(pmap, prot);

#ifdef DEBUG
	if (pmapdebug & (PDB_FOLLOW|PDB_PROTECT))
		printf("pmap_protect(%x, %x, %x, %x)\n", pmap, sva, eva, prot);
#endif

	if (pmap == NULL)
		return;

#ifdef PMAPSTATS
	protect_stats.calls++;
#endif
	if ((prot & VM_PROT_READ) == VM_PROT_NONE) {
		pmap_remove(pmap, sva, eva);
		return;
	}
	if (prot & VM_PROT_WRITE)
		return;
	if (atoseg(sva) != SYS_SEG) {
		int     spl;
		spl = splvm();

		pmap_cache_protect_range(pmap, sva, eva, prot);
		splx(spl);
	}
	list = &pmap->ppage_list;
	entry = (struct pmap_list_entry *) queue_first(list);
	while (!queue_end(list, (queue_entry_t) entry)) {
	    if ((entry->vaddr >= sva) && (entry->vaddr < eva)) {
		if (pmap_extract(pmap, entry->vaddr) != 0) {
		    pmap_protect_phys_page(entry->paddr, mmu_prot);
		    tlb_invalidate_seg(atoseg(entry->vaddr));
		}
	    }
	    entry = (pmap_list_entry_t) queue_next(&entry->pmap_pages);
	}
}

/*
 *	Insert the given physical page (pa) at
 *	the specified virtual address (va) in the
 *	target physical map with the protection requested.
 *
 *	If specified, the page will be wired down, meaning
 *	that the related pte can not be reclaimed.
 *
 *	NB:  This is the only routine which MAY NOT lazy-evaluate
 *	or lose information.  That is, this routine must actually
 *	insert this page into the given map NOW.
 */
void
pmap_enter(pmap, va, pa, prot, wired)
	register pmap_t pmap;
	vm_offset_t va;
	register vm_offset_t pa;
	vm_prot_t prot;
	boolean_t wired;
{
	int		sid;
	int		old_sr;
	int		seg = atoseg(va);
	int		pl, this_page;
	int		count;
	vm_offset_t 	vpage =  a_to_rosetta_page(segoffset(va));
	vm_offset_t 	ppage =  a_to_rosetta_page(pa);

#ifdef DEBUG
	if (pmapdebug & (PDB_FOLLOW|PDB_ENTER))
		printf("pmap_enter(%x, %x, %x, %x, %x)\n",
		       pmap, va, pa, prot, wired);
#endif
	if (pmap == NULL)
		return;

#ifdef PMAPSTATS
	if (pmap == kernel_pmap)
		enter_stats.kernel++;
	else
		enter_stats.user++;
#endif
	pl = splvm();

	lock_read(&pmap_lock);
	simple_lock(&pmap->lock);

	if (seg != SYS_SEG) {
		sid = pmap->sidtab[seg];
		if (sid == 0) {
			sid = pmap->sidtab[seg] = pmap_alloc_seg();

			/* If we are entering in the currently active pmap,
			 * physically enter the new segment id in the rosetta
			 * segment register.
			 *
			 * No need for TLB invalidate here (or it
			 * wouldn''t have faulted).
			 */

			if (pmap == active_pmap)
				set_rosetta_seg(seg,sid_to_seg(sid));
		}
	} else {
		sid = ROMP_SID_SYSTEM;
	}

	/*
 	 * This fakes sr0 into pointing at the segment on which we're
	 * working so that the tva will tell us if the virtual address
	 * is REALLY mapped in.
	 */
	 
	old_sr = get_rosetta_seg(HACK_SEG);
	set_rosetta_seg(HACK_SEG,sid_to_seg(sid));
	tlb_invalidate_seg(HACK_SEG);

	prot = rtakey(pmap,prot);
	count = PAGE_SIZE/rose_page_size;

#define	pmap_ppage_in_use(p)	(pmap_list_entries[atop(p)].in_use)

	/*
	 * remove the physical page from the pmap->ppages_list
	 */

	if (pmap_ppage_in_use(pa)) {
	    pmap_remove_ppage_entry(pa);
	    /*
	     * Only add a cache entry if we are sharing this
	     * page with another process
	     */
	    if (sid != ROMP_SID_SYSTEM) {
		pmap_cache_enter(pmap,sid,vpage,ppage,prot);
	    }
	} else {
	    pmap_cache_remove(pmap,sid,vpage);
	}
	
	if (isset(is_in,ppage)) {
		mapout(ppage,ppages_per_vmpage);
		tlb_invalidate_all();
	}

	/*
	 * the vm system assumes that it can remap a virtual address
	 * without first removing an old mapping.
	 */
	if ((this_page = 
	    trans_virt_addr((HACK_SEG << 28) | segoffset(va))) != -1) {
		mapout(a_to_rosetta_page(this_page),count);
		pmap_remove_ppage_entry(this_page);
	}
	
	mapin(ppage, vpage, prot, count, sid);
	pmap_add_ppage_entry(pmap, va, pa);
	
	set_rosetta_seg(HACK_SEG,old_sr);
	tlb_invalidate_seg(HACK_SEG);
	tlb_invalidate_seg(atoseg(va));

	simple_unlock(&pmap->lock);
	lock_read_done(&pmap_lock);
	DDISPLAY(0);
	splx(pl);
}

/*
 *	Routine:	pmap_change_wiring
 *	Function:	Change the wiring attribute for a map/virtual-address
 *			pair.
 *	In/out conditions:
 *			The mapping must already exist in the pmap.
 */
void
pmap_change_wiring(pmap, va, wired)
	register pmap_t	pmap;
	vm_offset_t	va;
	boolean_t	wired;
{
#ifdef DEBUG
	if (pmapdebug & PDB_FOLLOW)
		printf("pmap_change_wiring(%x, %x, %x)\n", pmap, va, wired);
#endif
	if (pmap == NULL)
		return;
}

/*
 *	Routine:	pmap_extract
 *	Function:
 *		Extract the physical page address associated
 *		with the given map/virtual_address pair.
 */

vm_offset_t
pmap_extract(pmap, va)
	register pmap_t	pmap;
	vm_offset_t va;
{
	int	phys;
	int	pl;
	int 	old_sr;

#ifdef DEBUG
	if (pmapdebug & PDB_FOLLOW)
		printf("pmap_extract(%x, %x) -> ", pmap, va);
#endif
	pl = splvm();
	simple_lock(&pmap->lock);
	DDISPLAY(11);

	if (atoseg(va) == SYS_SEG) {
		phys = trans_virt_addr(va);
		DDISPLAY(0);
		simple_unlock(&pmap->lock);
		splx(pl);
		return(phys == -1 ? 0 : phys);
	}
	old_sr = get_rosetta_seg(HACK_SEG);
	set_rosetta_seg(HACK_SEG,sid_to_seg(pmap->sidtab[atoseg(va)]));
	tlb_invalidate_seg(HACK_SEG);
	
	phys = trans_virt_addr((HACK_SEG<<28) | segoffset(va));

	set_rosetta_seg(HACK_SEG,old_sr);
	tlb_invalidate_seg(HACK_SEG);

	DDISPLAY(0);
	simple_unlock(&pmap->lock);
	splx(pl);
	return(phys == -1 ? 0 : phys);
}

/*
 *	Copy the range specified by src_addr/len
 *	from the source map to the range dst_addr/len
 *	in the destination map.
 *
 *	This routine is only advisory and need not do anything.
 */
void pmap_copy(dst_pmap, src_pmap, dst_addr, len, src_addr)
	pmap_t		dst_pmap;
	pmap_t		src_pmap;
	vm_offset_t	dst_addr;
	vm_size_t	len;
	vm_offset_t	src_addr;
{
#ifdef DEBUG
	if (pmapdebug & PDB_FOLLOW)
		printf("pmap_copy(%x, %x, %x, %x, %x)\n",
		       dst_pmap, src_pmap, dst_addr, len, src_addr);
#endif
}

/*
 *	Require that all active physical maps contain no
 *	incorrect entries NOW.  [This update includes
 *	forcing updates of any address map caching.]
 *
 *	Generally used to insure that a thread about
 *	to run will see a semantically correct world.
 */
void pmap_update()
{
#ifdef DEBUG
	if (pmapdebug & PDB_FOLLOW)
		printf("pmap_update()\n");
#endif
}

/*
 *	Routine:	pmap_collect
 *	Function:
 *		Garbage collects the physical map system for
 *		pages which are no longer used.
 *		Success need not be guaranteed -- that is, there
 *		may well be pages which are not referenced, but
 *		others may be collected.
 *	Usage:
 *		Called by the pageout daemon when pages are scarce.
 */
void
pmap_collect(pmap)
	pmap_t		pmap;
{

	if (pmap != kernel_pmap)
		return;

#ifdef DEBUG
	if (pmapdebug & PDB_FOLLOW)
		printf("pmap_collect(%x)\n", pmap);
#endif
#ifdef PMAPSTATS
	kpt_stats.collectscans++;
#endif
}

/*
 *	Routine:	pmap_activate
 *	Function:
 *		Binds the given physical map to the given
 *		processor.  (Since there are only uniprocessor
 *		romp/rosetta's, all this does is remember which
 *		pmap is currently active for pmap_enter.)
 *
 */
void
pmap_activate(pmap, pcbp)
	register pmap_t pmap;
	struct pcb *pcbp;
{
#ifdef DEBUG
	if (pmapdebug & (PDB_FOLLOW|PDB_SEGTAB))
		printf("pmap_activate(%x, %x)\n", pmap, pcbp);
#endif
	active_pmap = pmap;
}

/*
 *	pmap_zero_page zeros the specified (machine independent)
 *	page by mapping the page into virtual memory and using
 *	bzero to clear its contents, one machine dependent page
 *	at a time.
 *
 */
void
pmap_zero_page(phys)
	vm_offset_t phys;
{
	register int s = splvm(); /* Can't interrupt with virt_eq_real on! */
	register int old_real;

#ifdef DEBUG
	if (pmapdebug & PDB_FOLLOW)
		printf("pmap_zero_page(%x)\n", phys);
#endif

	DDISPLAY(13);
	old_real = virt_eq_real(TRUE);
	bzero((caddr_t)phys, PAGE_SIZE);
	set_mod_bit(phys>>0xb,0); /* Zero fill page is clean...*/
	virt_eq_real(old_real);
	DDISPLAY(0);
	splx(s);
}

/*
 *	pmap_copy_page copies the specified (machine independent)
 *	page by mapping the page into virtual memory and using
 *	bcopy to copy the page, one machine dependent page at a
 *	time.
 *
 */
void
pmap_copy_page(src, dst)
	vm_offset_t src, dst;
{
	register int s = splvm();
	register int old_real;

#ifdef DEBUG
	if (pmapdebug & PDB_FOLLOW)
		printf("pmap_copy_page(%x, %x)\n", src, dst);
#endif

	DDISPLAY(14);
	old_real = virt_eq_real(TRUE);
	bcopy((caddr_t)src, (caddr_t)dst, PAGE_SIZE);
	virt_eq_real(old_real);
	splx(s);
	DDISPLAY(0);
}

/*
 *	Routine:	pmap_pageable
 *	Function:
 *		Make the specified pages (by pmap, offset)
 *		pageable (or not) as requested.
 *
 *		A page which is not pageable may not take
 *		a fault; therefore, its page table entry
 *		must remain valid for the duration.
 *
 *		This routine is merely advisory; pmap_enter
 *		will specify that these pages are to be wired
 *		down (or not) as appropriate.
 */
void
pmap_pageable(pmap, sva, eva, pageable)
	pmap_t		pmap;
	vm_offset_t	sva, eva;
	boolean_t	pageable;
{
#ifdef DEBUG
	if (pmapdebug & PDB_FOLLOW)
		printf("pmap_pageable(%x, %x, %x, %x)\n",
		       pmap, sva, eva, pageable);
#endif
}

/*
 *	Routine:	pmap_clear_modify
 *	Function:	Clear the hardware modified ("dirty") bit for one
 *			machine independant page starting at the given
 *			physical address.  phys must be aligned on a machine
 *			independant page boundary.
 */
void
pmap_clear_modify(pa)
	vm_offset_t	pa;
{
	vm_offset_t	addr = pa;

#ifdef DEBUG
	if (pmapdebug & PDB_FOLLOW)
		printf("pmap_clear_modify(%x)\n", pa);
#endif

	DDISPLAY(25);
	for (; addr < PAGE_SIZE + pa; addr += rose_page_size)
		set_mod_bit(addr>>0xb,0);
	DDISPLAY(0);
}

/*
 *	Routine:	pmap_clear_reference
 *	Function:
 *		Clear the hardware referenced bit in the given machine
 *		independant physical page.  If referenced bits are not
 *		available in a particular hardware architecture, do a
 *		pmap_remove_all instead.
 */
void
pmap_clear_reference(pa)
	vm_offset_t	pa;
{
	vm_offset_t	addr = pa;

#ifdef DEBUG
	if (pmapdebug & PDB_FOLLOW)
		printf("pmap_clear_reference(%x)\n", pa);
#endif

	DDISPLAY(27);
#if	1
	for (;addr < PAGE_SIZE + pa; addr += rose_page_size) 
		set_ref_bit(addr>>0xb,0);
#else	0
	pmap_remove_all(pa);
#endif	0
	DDISPLAY(0);
}

/*
 *	pmap_is_referenced:
 *
 *	Return whether or not the specified physical page is referenced
 *	by any physical maps.
 *
 *	XXX actually returns TRUE iff the given physical page has
 *	    been referenced since the last pmap_clear_reference.

 */

boolean_t
pmap_is_referenced(pa)
	vm_offset_t	pa;
{
	register vm_offset_t	addr = pa;

#ifdef HP300_DEBUG
	if (pmapdebug & PDB_FOLLOW) {
		boolean_t rv = pmap_testbit(pa, PG_U);
		printf("pmap_is_referenced(%x) -> %c\n", pa, "FT"[rv]);
		return(rv);
	}
#endif
	for (;addr < PAGE_SIZE + pa; addr += rose_page_size) {
		if (get_ref_bit(addr>>0xb))
			return(TRUE);
	}
	return(FALSE);
}

/*
 *	pmap_is_modified:
 *
 *	Return whether or not the specified physical page is modified
 *	by any physical maps.
 *	XXX actually returns TRUE iff the given physical page has
 *	    been modified since the last pmap_clear_reference.
 */

boolean_t
pmap_is_modified(pa)
	vm_offset_t	pa;
{
	vm_offset_t	addr = pa;

#ifdef HP300_DEBUG
	if (pmapdebug & PDB_FOLLOW) {
		boolean_t rv = pmap_testbit(pa, PG_M);
		printf("pmap_is_modified(%x) -> %c\n", pa, "FT"[rv]);
		return(rv);
	}
#endif
	for (;addr < PAGE_SIZE + pa; addr += rose_page_size) {
		if (get_mod_bit(addr>>0xb))
			return(TRUE);
	}
	return(FALSE);
}

vm_offset_t
pmap_phys_address(ppn)
	int ppn;
{
	return(ptoa(ppn));
}

/********* End of official routines, start of RT support routines *********/

/*
 *	Routine:	pmap_make_kernel_text_readable
 *	Function:
 *		Allow users that have access to the kernel
 *		text segment to execute code from that segment.
 *		[There is no added privilege gained -- this is just
 *		a cheap shared library mechanism.]
 *	Note:
 *		This routine should be placed near the end of the
 *		code file, to avoid lint complaints about re-using
 *		the external symbol "start".  Obviously lint should
 *		be fixed, but that is not a kernel hacker's job.
 */
static void
pmap_make_kernel_text_readable()
{
	extern int start, edata;

	/*
	 * protect all the kernel text as world readable.
	 */
	pmap_pr_phys(kernel_pmap,
		trunc_page(&start),
		round_page(&edata),
		ROMP_KEY_URKW << ROMP_KEY_SHIFT);
}

/*
 *	Routine:	pmap_copy_on_write
 *	Function:
 *		Remove write privileges from all
 *		physical maps for this physical page.
 */
static void 
pmap_copy_on_write(phys)
	vm_offset_t	phys;
{
	int	ppage;
	struct hatipt_coarse *ipte;
	int s = splvm();
	
	DDISPLAY(6);
	/*
	 *	Lock the entire pmap system, since we may be changing
	 *	several maps.
	 */
	lock_write(&pmap_lock);
	for (ppage = phys; ppage < PAGE_SIZE + phys; ppage += rose_page_size)
		if (isset(is_in,ppage>>rose_page_shift)) {
			ipte = &ROMP_HATIPT[ppage>>rose_page_shift];

			/*
			 *	Because a mapping exists, it must already
			 *	have read permission; therefore, it is safe
			 *	to give out read permission here.
			 *
			 *	It is not possible to make kernel memory
			 *	read-only without letting users also read it.
			 *	XXX The kernel should only let users have access
			 *	to segments that don't contain sensitive data!
			 */

			ipte->key_addrtag &= ~(3 << ROMP_KEY_SHIFT);
			ipte->key_addrtag |= ROMP_KEY_URKR << ROMP_KEY_SHIFT;
		}
	pmap_cache_remove_page(phys);

	tlb_invalidate_all();
	lock_write_done(&pmap_lock);
	DDISPLAY(0);
	splx(s);
}

static void 
pmap_pr_phys_range(min_ppage, max_ppage, sid, s, e, map, internal_prot)
	int			min_ppage;
	register int		max_ppage, sid;
	register vm_offset_t	s, e;
	pmap_t			map;
	int			internal_prot;

{
	register int		pagenum;
	struct  hatipt_coarse	*ipte;
	unsigned long		key_addrtag;
	
#ifdef	lint
	map++; /* XXX This routine is internal-use-only, but has unused arguments? */
#endif	lint

	for (pagenum = min_ppage; pagenum < max_ppage; pagenum++) {
		ipte = &ROMP_HATIPT[pagenum];
		key_addrtag = ipte->key_addrtag;
		if ((isset(is_in,pagenum))
		  && (((key_addrtag & ROMP_SID_MASK) >> ROMP_VPAGE_BITS) == sid)
		  && ((key_addrtag & ROMP_VPAGE_MASK) >= s)
		  && ((key_addrtag & ROMP_VPAGE_MASK) < e)) {
			ipte->key_addrtag &= ~(3 << ROMP_KEY_SHIFT);
			ipte->key_addrtag |= internal_prot;
		  }
	     }
}

static void 
pmap_pr_phys(map, start, end, internal_prot)
	pmap_t		map;
	vm_offset_t	start, end;
	int		internal_prot;	
{
	int		pl, seg = atoseg(start);
	int		sid;
	vm_offset_t	s, e;	
	
	if (atoseg(start) != atoseg(end-1))
		panic("pmap_pr_phys (too big)");

	pl = splvm();

	lock_read(&pmap_lock);
	simple_lock(&map->lock);

	DDISPLAY(8);

	if (seg == SYS_SEG)
		sid = ROMP_SID_SYSTEM;
	else
		sid = map->sidtab[seg];

	if (sid == 0) {
		simple_unlock(&map->lock);
		lock_read_done(&pmap_lock);
		splx(pl);
		return;
	}
	/*
	 * Why aren't the following shifts using rose_page_size??
	 */

	if (atoseg(end) == atoseg(start))
		e = segoffset(end) >> 0xb;
	else
		e = (0x10000000) >> 0xb; /*XXX*/

	s = segoffset(start) >> rose_page_shift;		

	if (seg != SYS_SEG)
		pmap_cache_remove_range(map,start,end);

	/*
 	 *	Optimization:  Do not check the kernel text
	 *	if we are removing from a non-system segment (we probably
	 *	do not need to anyway, but just for safety's sake...)
	 */
	pmap_pr_phys_range(((sid == ROMP_SID_SYSTEM) ? 0 : min_user_pa),
		a_to_rosetta_page(holestart),
		sid, s, e, map, internal_prot);

	if (ptoa(physmem) > holestart) 
		pmap_pr_phys_range(a_to_rosetta_page(holeEnd),
			vm_to_rosetta_page(physmem) +
			a_to_rosetta_page(holeEnd - holestart),
			sid, s, e, map, internal_prot);

	tlb_invalidate_seg(seg);
	simple_unlock(&map->lock);
	lock_read_done(&pmap_lock);

	DDISPLAY(0);
	splx(pl);
}

void
pmap_protect_phys_page(phys_addr,prot)
	vm_offset_t	phys_addr;
	int		prot;
{
    	vm_offset_t p;
	struct hatipt_coarse *ipte;
        int     spl;

	spl = splvm();
	
	for (p = phys_addr; p < phys_addr + PAGE_SIZE; p += rose_page_size) {
		ipte = &ROMP_HATIPT[a_to_rosetta_page(p)];
		ipte->key_addrtag &= ~(3 << ROMP_KEY_SHIFT);
		ipte->key_addrtag |= prot << ROMP_KEY_SHIFT;
	}
	splx(spl);	
}

/*	
 *	Allocate a segment number which is currently unused.
 *	Current algorithm allocates by doing a linear scan of an
 *	array.  XXX - Should be done with a free list.
 */
static short
pmap_alloc_seg()
{
	register short	cur_seg;

	cur_seg = max_free_seg;
	while (cur_seg > 0) {
		if (isclr(pmap_seg_tab, cur_seg))
			break;
		cur_seg--;
	}

	if (cur_seg <= 0)
		panic("out of segments");

	setbit(pmap_seg_tab, cur_seg);
	max_free_seg = cur_seg;

	return(cur_seg);
}


/*
 *	Deallocate a formerly used segment (duplicate deallocates are NOT ok).
 */
static void
pmap_dealloc_seg(segid)
	short	segid;
{
	if (isclr(pmap_seg_tab,segid))
		panic("pmap_dealloc_seg: dup free");

	clrbit(pmap_seg_tab,segid);
	if (max_free_seg < segid)
		max_free_seg = segid;
}

/*
 *	Initialize the segment allocation system.
 */
static void
pmap_init_seg_alloc()
{
	int	current_seg = ROMP_NSEGS;

	while (current_seg >= 0) {
		clrbit(pmap_seg_tab,current_seg);
		current_seg--;
	}

	setbit(pmap_seg_tab, ROMP_SID_SYSTEM);	/* Allocate system segment */
}

/*
 *	Routine:	pmap_remove_all
 *	Function:
 *		Removes this physical page from
 *		all physical maps in which it resides.
 *		Reflects back modify bits to the pager.
 *
 *		Since a physical page may have at most one virtual
 * 		address on the rosetta, just map it out.
 *
 *		XXX - pmap locking.
 */
static void
pmap_remove_all(phys)
	vm_offset_t	phys;
{
	vm_offset_t	addr = phys;
	int		s = splvm();

	DDISPLAY(5);

	tlb_invalidate_all();

	pmap_cache_remove_page(addr);

	/*
	 *  If the cache is in place, the is_in bit can
	 *  be clear, but the page still can belong to a pmap.
	 */
	pmap_remove_ppage_entry(addr);
	
	if (isset(is_in,addr>>rose_page_shift)) {
		DDISPLAY(35);
		mapout(a_to_rosetta_page(addr),ppages_per_vmpage);
		DDISPLAY(5);
	}

#ifndef	MACH_XP
	for (;addr < PAGE_SIZE + phys; addr += rose_page_size) {
		if (get_mod_bit(addr>>0xb))
			vm_page_set_modified(PHYS_TO_VM_PAGE(addr));
	}
#endif	MACH_XP

	DDISPLAY(0);
	splx(s);
}

static void
pmap_add_ppage_entry(pmap,v,p)
	struct pmap	*pmap;
	vm_offset_t	v;
	vm_offset_t	p;
{
    pmap_list_entry_t entry;

    /*
     * before memory allocator initialized?
     */
 
    if (pmap_list_entries == 0) return;

    entry = &pmap_list_entries[atop(p)];
    if (entry->in_use) {
	printf("pmap_add_ppage_entry pmap %x, entry %x",
		pmap,entry);
	panic("pmap_add_ppage_entry : in_use page");
    }

    entry->vaddr = v;
    entry->paddr = p;
    entry->in_use = TRUE;
    entry->pmap = pmap;
    queue_enter(&pmap->ppage_list,entry,pmap_list_entry_t,pmap_pages);
    pmap->ppage_count++;
}

static void
pmap_remove_ppage_entry(paddr)
vm_offset_t paddr;
{
    	pmap_t			pmap;
	queue_head_t		*list;
	struct pmap_list_entry	*entry;


#ifdef	notyet
	entry = &pmap_list_entries[atop(paddr)];
	pmap = entry->pmap;
	queue_remove(&pmap->ppage_list,entry,pmap_list_entry_t, pmap_pages);
	pmap->ppage_count--;
	entry->in_use = FALSE;
#else	notyet
	/*
	 * Sanity check
	 */
	 
	pmap = pmap_list_entries[atop(paddr)].pmap;

	/*
	 * Its entirely possible that with the cache in place,
	 * we get here with no pmap associated with this page.
	 * Probably we should find the current owner and reset
	 * this field, but for now, we'll just punt.
	 */
	if (pmap == PMAP_NULL) return;

	list = &pmap->ppage_list;

	entry = (struct pmap_list_entry *) queue_first(list);
	while (!queue_end(list,(queue_entry_t) entry)) {
	    if (entry->paddr == paddr) {
		goto sanity_ok;
	    }
	    entry = (pmap_list_entry_t) queue_next(&entry->pmap_pages);
	}
	printf("pmap=%x, list=%x, paddr=%x, entry=%x\n",
		pmap,&pmap->ppage_list,paddr,&pmap_list_entries[atop(paddr)]);
	panic("pmap_remove_ppage_entry failed sanity check!");

sanity_ok:
	/*
	 * remove it from the pmap ppage list,
	 * and then free it.
	 */
	queue_remove(list,entry,pmap_list_entry_t, pmap_pages);
	pmap->ppage_count--;	
	entry->in_use = FALSE;
	entry->pmap = PMAP_NULL;
#endif	notyet

}

/*
 *	Bootstrap the system enough to run with virtual memory.
 *	Called with mapping ON.
 *
 *	Parameters:
 *	load_start:	PA where kernel was loaded
 *	avail_start	PA of first available physical page
 *	avail_end	PA of last available physical page
 *	virtual_avail	VA of first available page (after kernel bss)
 *	virtual_end	VA of last available page (end of kernel addr space)
 *
 *	&start		start of kernel text
 *	&etext		end of kernel text
 *
 */
void
pmap_bootstrap(load_start, avail_start, avail_end, virtual_avail, virtual_end)
	vm_offset_t	load_start;
	vm_offset_t	*avail_start;	/* IN/OUT */
	vm_offset_t	*avail_end;	/* IN/OUT */
	vm_offset_t	*virtual_avail;	/* IN/OUT */
	vm_offset_t	*virtual_end;	/* OUT */
{
	int			index;
	void			pmap_pr_phys();
	int 			i;
	vm_size_t		s;

	/*
	 *	Initialize the "is it mapped in" table.  lohatipt makes
	 *	sure that all pages are indeed mapped in when we start up.
	 *
	 * 	Also initialize the table of back pointers.  Initially,
	 *	the code in lohatipt.s sets up the page table such that
	 * 	system page i maps to real page i.  	
	 */
	for (index = 0;index < physmem;index++) {
	    	struct hatipt_coarse *ipte, *next;

		ipte = &ROMP_HATIPT[index];
		if ((ipte->hat_ptr != ROMP_UNDEF_PTR) &&
		    (!ROMP_ENDCHAIN(ipte->hat_ptr))) {
			next = &ROMP_HATIPT[ipte->hat_ptr];
			next->rev_ptr = IPT_HATLINK | index;
		} 
		setbit(is_in,index);
	}

	/*
	 *	The kernel's pmap is statically allocated so we don't
	 *	have to use pmap_create, which is unlikely to work
	 *	correctly at this part of the boot sequence.
	 */

	lock_init(&pmap_lock, FALSE);
	simple_lock_init(&kernel_pmap->lock);
	queue_init(&kernel_pmap->ppage_list);
	queue_init(&kernel_pmap->cache_entries);
	kernel_pmap->ppage_count = 0;
	kernel_pmap->cache_count = 0;

	pmap_init_seg_alloc();/* Initialize the rosetta segment allocator. */


	for (index = 0; index < NUSR_SEGS; index++)
		kernel_pmap->sidtab[index] = 0;
	/*
 	 *	We do not need to set the system segment in the kernel pmap
 	 *	because the system segment (SYS_SEG) is not considered part
	 * 	of a pmap, but rather is a constant (and ser(SYS_SEG) is never
	 *	changed after it is initialized by lohatipt.s).
	 */
	kernel_pmap->ref_count = 1;
	
	*virtual_avail = round_page(VM_MIN_KERNEL_ADDRESS + mem_size); /*XXX*/
	*virtual_end = trunc_page(VM_MAX_KERNEL_ADDRESS);
	
	ppages_per_vmpage = (PAGE_SIZE / rose_page_size);

	/*
	 * Grab some physical memory for the auxiallary page table
	 * information.
	 */
	s = round_page(atop(mem_size) * sizeof(struct pmap_list_entry));
	*avail_end -= s;
	
#ifndef	SIMULATOR
	pmap_list_entries = (pmap_list_entry_t) (*avail_end + load_start);
#else
	printf("pmap_list_entries would be at 0x%08x for 0x%x bytes\n",
		*avail_end + load_start, s);
	pmap_list_entries = (pmap_list_entry_t)malloc(s);
	printf("allocated %d bytes for pmap_list_entries at 0x%x\n", s,
		pmap_list_entries);
	if (!pmap_list_entries)
		panic("can't allocate pmap_list_entries");
#endif

	for (i = 0;  i < atop(mem_size); i++) {
	    	pmap_list_entry_t entry = &pmap_list_entries[i];
		vm_offset_t p_addr = (vm_offset_t) ( i << PAGE_SHIFT);
		entry->pmap = kernel_pmap;
		entry->in_use = TRUE;
		entry->vaddr = p_addr + load_start;
		entry->paddr = p_addr;
		queue_enter(&kernel_pmap->ppage_list,
				entry,
				pmap_list_entry_t,
				pmap_pages);
		kernel_pmap->ppage_count++;
	}

	/*
	 *	Initialize the shared-memory cache hash table.
	 */
	pmap_cache_init(load_start, avail_end);
	
	min_user_pa = round_page((vm_offset_t)*avail_end) >> rose_page_shift;
				 /*No user physical pages yet*/
				 
	/*
	 *	Remove unused mappings so that consistency code will work.
	 */

	mapout(a_to_rosetta_page(*avail_start),
	       a_to_rosetta_page((*avail_end) - (*avail_start)));

	return;
}

/*
 *	Routine:	virt_eq_real
 *	Function:
 *		Set segment 0 virtual_equal_real mode if turnon,
 *		clear otherwise; in any case, return the old 
 *		value.
 */

static
virt_eq_real(turnon)
	int	turnon;
{
	register int tcr = ior (ROMP_TCR);

	if (turnon)
		iow(ROMP_TCR,tcr | TCR_V);
	else
		iow(ROMP_TCR,tcr & (~TCR_V));

	return((tcr & TCR_V) ? 1 : 0);
}

/*
 * mapout - remove a given physical address from the hat/ipt.
 */
 
void
mapout(ppage,count)
	vm_offset_t	ppage;
	int		count;
{

#ifndef	USE_ASSEMBLY
	register struct hatipt_coarse *ipte, *prev, *next;
	int i;
#endif	/* USE_ASSEMBLY */

#ifdef	USE_ASSEMBLY
	lo_mapout(ppage,count);
#else
	

	for (i=0; i < count; i++, ppage += 1) {

	    if (isclr(is_in,ppage)) {
		if (pmapdebug)
		    printf("mapout: already out ppage %x count %d\n",
		    	ppage,count);
#ifdef	ROMP_CACHE
	/*
	 * It is ok to try to mapout a page which is already out.
	 *  lo_lookup() may do this.
	 *  pmap_remove() may also do it.
	 */
		return;
#else
		panic("mapout: already out");
#endif	/* ROMP_CACHE */
	    }
	    clrbit(is_in,ppage);
  	    ipte = &ROMP_HATIPT[ppage];

	    if (!ROMP_ENDCHAIN(ipte->ipt_ptr)) {
	    	next = &ROMP_HATIPT[ipte->ipt_ptr];
	    	next->rev_ptr = ipte->rev_ptr;
	    }
	    
	    if (ROMP_ENDCHAIN(ipte->rev_ptr)) {
		/*
		 * The reverse link points to a hat entry
		 */
		prev = &ROMP_HATIPT[ipte->rev_ptr&IPT_HATLINK_MASK];
		prev->hat_ptr = ipte->ipt_ptr;
	    } else {
		/*
		 * The reverse link points to just another ipt entry.
		 */
		prev = &ROMP_HATIPT[ipte->rev_ptr];
		prev->ipt_ptr = ipte->ipt_ptr;
	    }
	    ipte->ipt_ptr = ROMP_UNDEF_PTR;
	    ipte->rev_ptr = ROMP_UNDEF_PTR;	    
	}
#endif	/* USE_ASSEMBLY */
}

/*
 * mapin - add a given physical/virtual address translation to the hat/ipt.
 */
void
mapin(ppage, vpage, prot, count, sid)
	vm_offset_t	ppage;
	vm_offset_t	vpage;
	int		prot;
	int		count;
	int		sid;
{
#ifndef	USE_ASSEMBLY
	register struct hatipt_coarse *ipte, *head, *next_ipte;
	u_int key;
	int i;
	int hash_head;
#endif	/* USE_ASSEMBLY */

#ifdef	USE_ASSEMBLY
	lo_mapin(ppage, vpage, prot, count, sid);
#else
	if (isset(is_in,ppage)) {
	        printf("mapin: ppage %x, vpage %x, sid %x, count %d\n",
			ppage,vpage,sid,count);
		panic("mapin: page already in");
	}

	if ((sid != ROMP_SID_SYSTEM) && (ppage < min_user_pa))
		min_user_pa = ppage;

	for (i = 0; i < count ; i++, ppage += 1, vpage += 1) {
		/* point to the ipte for this page frame */
		ipte = &ROMP_HATIPT[ppage];

		/* determine the key, write, tid, lockbits */
		key = prot;	  /* non-special segments only */

		/* fill in the fields to reflect the new virt addr and prot */
		ipte->key_addrtag = (key << ROMP_KEY_SHIFT)
			|(sid << ROMP_VPAGE_BITS)
			|((rose_page_size == 0x1000) ? vpage << 1 : vpage);
			
		ipte->w = 1;		  /* special segments only */
		ipte->tid = 0 ;		  /* special segments only */
		ipte->lockbits = -1;	  /* special segments only */

		/* hash to find the head of the ipte chain */
		head = &ROMP_HATIPT[hash_head = ROMP_HASH(sid, vpage)];

		if (!ROMP_ENDCHAIN(head->hat_ptr)) {
		    next_ipte = &ROMP_HATIPT[head->hat_ptr];
		    next_ipte->rev_ptr = ppage;
		}
		ipte->ipt_ptr = head->hat_ptr;
		ipte->rev_ptr = hash_head | IPT_HATLINK;
		
		set_hatptr(head, ppage);

		/* Mark it in in our external table. */
		setbit(is_in,ppage);
	}
#endif	/* USE_ASSEMBLY	 */
}

/*
 * remap - reassign the physical/virtual address translation for a given page
 */
void
remap(sid,vpage,ppage,prot)
	int		sid;
	vm_offset_t	vpage;
	vm_offset_t	ppage;
	int		prot;
{
#ifndef	USE_ASSEMBLY
    struct hatipt_coarse *ipte, *prev, *next, *head;
    int i;
    int hash_head;
#endif	/* USE_ASSEMBLY */

#ifdef	USE_ASSEMBLY
	lo_remap(sid, vpage, ppage, prot);
#else
    for (i = 0; i < ppages_per_vmpage; ppage++, vpage++, i++) {
  	    ipte = &ROMP_HATIPT[ppage];

	    if (isset(is_in,ppage)) {
		if (!ROMP_ENDCHAIN(ipte->ipt_ptr)) {
		    next = &ROMP_HATIPT[ipte->ipt_ptr];
		    next->rev_ptr = ipte->rev_ptr;
		}

		if (ROMP_ENDCHAIN(ipte->rev_ptr)) {
		    /*
		     * The reverse link points to a hat entry
		     */
		    prev = &ROMP_HATIPT[ipte->rev_ptr&IPT_HATLINK_MASK];
		    prev->hat_ptr = ipte->ipt_ptr;
		} else {
		    /*
		     * The reverse link points to just another ipt entry.
		     */
		    prev = &ROMP_HATIPT[ipte->rev_ptr];
		    prev->ipt_ptr = ipte->ipt_ptr;
		}
	    } else {
		setbit(is_in,ppage);
	    }

	    /*
	     * fill in the fields to reflect the new virt addr and prot
	     */
	     
	    ipte->key_addrtag = (prot << ROMP_KEY_SHIFT)
			|(sid << ROMP_VPAGE_BITS)
			|((rose_page_size == 0x1000) ? vpage << 1 : vpage);

	    ipte->w = 1;		  /* special segments only */
	    ipte->tid = 0 ;		  /* special segments only */
	    ipte->lockbits = -1;	  /* special segments only */

	    /* hash to find the head of the ipte chain */
	    head = &ROMP_HATIPT[hash_head = ROMP_HASH(sid, vpage)];
  	    if (!ROMP_ENDCHAIN(head->hat_ptr)) {
		    next = &ROMP_HATIPT[head->hat_ptr];
		    next->rev_ptr = ppage;
	    }
	    ipte->ipt_ptr = head->hat_ptr;	    
	    ipte->rev_ptr = hash_head | IPT_HATLINK;
	    set_hatptr(head, ppage);
    }
#endif	/* USE_ASSEMBLY */
}

/*
 *	This function takes a virtual address and returns the corresponding
 *	physical address if it is mapped in, else returns -1.
 *
 *	NOTE: This was found to be wanting for speed, and has been moved 
 *	to loutil.s
 *
 */
#if	defined(SIMULATOR)

static long
trans_virt_addr(va)
	vm_offset_t	va;
{
	struct hatipt *ipte;
	int ppage, seg, page_ind;

	seg=get_rosetta_seg(atoseg(va))>>2&0xfff;  /* only seg number */
	page_ind=a_to_rosetta_page(segoffset(va)); /* page number in segment */
	printf("trans_virt_addr: seg=0x%03x  page index=0x%04x\n", seg,
		page_ind);
	ipte = (struct hatipt *)&ROMP_HATIPT[ROMP_HASH(seg, page_ind)];
	if (!ipte->hat_empty) {
	    ppage=ipte->hat_ptr;
	    for(;;) {
		ipte=(struct hatipt *)&ROMP_HATIPT[ppage];
		if ((ipte->sid == seg) && 
		    (ipte->vpage) == page_ind) {
		    printf("trans_virt_addr: found mapping to page 0x%04x\n",
			ppage);
		    return(ppage);
		}
		if (ipte->ipt_last) {
		    printf("trans_virt_addr: No mapping found, exhausted IPT\n");
		    return(-1);
		}
		ppage=ipte->ipt_ptr;
	    }
	}
	printf("trans_virt_addr: No mapping found, empty HAT\n");
	return(-1);
}

#elif	!defined(USE_ASSEMBLY)

static long
trans_virt_addr(va)
	vm_offset_t	va;
{
	int	phys;

	rosetta_write_lrar(va);
	phys = rosetta_read_trar();
	if (phys < 0)
		return(-1);
	return(phys);
}
#endif	/* USE_ASSEMBLY */
