
#include "kbd.h"
#include <machine/pio.h>

static int active_kbd;

static char
kbd_read (loop)
int	loop;
{
    char c, c1;

    while (! ((c = inb(READ_8255_PC)) & PA_INPUT_FULL)) {
	delay(1);
	if (loop && --loop == 0) {
	    return (0);
	}
    }
    c1 = inb(READ_8255_PA);
    return (c1);
}


/* issue a keyboard-specific command */
static char
kbd_cmd(cmd)
int cmd;
{
    char c;

    /*
     * wait for an empty buffer.  Write the command, expect an ack.
     * if busy-rejected, try it again.
     */
    do {
	/* wait till the output buffer is empty */
	while (! (inb(READ_8255_PC) & PA_OUTPUT_EMPTY))
	    delay(10);

	outw(WRITE_8255_PA, cmd);
	c = kbd_read (100000);
	if (c == KBW_ACCEPT)
	    break;
    } while (c == KBW_BUSY);
    return c;
}

/* tech ref, page 5-141 */
kbd_init() 
{
    char c, c1;

    active_kbd = 0;

    /* initiate a reset via CRRB */
    c = inb(CRRB);
    c1 = c & KB_CRRB_MASK;
    outb(CRRB, c1);

    /* configure 8255 */
    outb(WRITE_8255_CONF, 0xC3);

    /* release adapter reset */
    delay(150000);
    c1 = c | ~KB_CRRB_MASK;
    outb(CRRB, c1);

    /* check results of self-test */
    delay(150000);
    c = inb(READ_8255_PC);
    if ((c & AIIR_MASK) != INT_8051_TEST) {
	printf ("keyboard 8051 self-test failed, PC=0x%x\n", c);
	return 0;
    }
    c = inb(READ_8255_PA);

    if (c != 0xae) {
	printf ("Keyboard 8051 self-test failed, PA=0x%x\n", c);
	return 0;
    }

    /* reset the keyboard */
    c1 = kbd_cmd (PA_CMD(WR_KEYBOARD,KBC_RESET));
    if (c1) {
	printf ("unexpected reset return from keyboard, PA=0x%x\n", c1);
	return 0;
    }
    delay(600000);

    /* check the results of the reset */
    c = kbd_read (10000);
    if (c == 0xe0) {
	printf ("no keyboard\n");
	return 0;
    }
    if (c != 0xaa) {
	printf ("invalid keyboard test completion code\n");
	return 0;
    }
    c = kbd_read (10000);	/* should get 0xbf */
    c = kbd_read (10000);	/* should get 0xb0 */

    /* enable keyboard scanning, turn on make-mode */
    kbd_cmd (PA_CMD(WR_KEYBOARD,KBC_ENABLE));

#ifdef bad_idea
    kbd_cmd (PA_CMD(WR_KEYBOARD,KBC_ALLMAKE));
#endif

    return ++active_kbd;
}

static char keymod;	/* keyboard modifier */
static char keyled;	/* keyboard leds */
extern char kbd_keymap[];
extern char kbd_shiftmap[];

#define islower(c)	((c) >= 'a' && (c) <= 'z')
#define toupper(c)	((c) - 'a' + 'A')

/*
 * expecting to be in make/break mode, where a key-press/key-release
 * produces the sequence key-code,break-code,key-code.
 */
int
kbd_getc()
{
    char c, okeyled, lastscan;
    int sawbreak = 0, wasbreak;
    int scan_again;

    if (!active_kbd)
	return 0;

    okeyled = keyled;
    lastscan = 0;
    do {
	wasbreak = scan_again = 0;
	c = kbd_read(5000);	/* long enough delay to prevent "bounces" */

	switch (c) {
	    case KEY_BREAK:
		wasbreak ++;
		sawbreak ++;
		scan_again ++;
		break;

	    case KEY_CAPSLOCK:
		if (lastscan == c)	/* need to disable repeat */
		    break;
		if (sawbreak) 
		    break;
		keymod ^= KEY_MOD_CAPSLOCK;
		keyled ^= KBC_LED_CAPSLOCK;
		break;

	    case KEY_SHIFT_LEFT:
	    case KEY_SHIFT_RIGHT:
		if (lastscan == c)	/* need to disable repeat */
		    break;
		if (sawbreak) {
		    keymod &= ~KEY_MOD_SHIFTED;
		    break;
		}
		keymod |= KEY_MOD_SHIFTED;
		break;

	    case KEY_CONTROL:
		if (lastscan == c)	/* need to disable repeat */
		    break;
		if (sawbreak) {
		    keymod &= ~KEY_MOD_CONTROL;
		    break;
		}
		keymod |= KEY_MOD_CONTROL;
		break;

	    /* this group is not implemented yet */
	    case KEY_ALT_LEFT:
	    case KEY_ALT_RIGHT:
	    case KEY_ACTION:
	    case KEY_NUMLOCK:
		scan_again++;
		break;

	    case 0:
/* 		printf ("."); */
		break;

	    default:
		if (sawbreak)
		    scan_again ++;
		break;
	}
	/* if this key wasn't the break sequence, but the previous
	 * sequence was a break sequence, reset sawbreak for the next time.
	 */
	if (!wasbreak && sawbreak)
	    sawbreak = 0;
	lastscan = c;
    } while (scan_again);

    /* update LEDs */
#ifdef notworking
    if (keyled != okeyled) {
	char c1;
	c1 = kbd_cmd (PA_CMD(WR_KEYBOARD,KBC_SET_LED));
	printf ("setled command returned 0x%x\n", c1);
	c1 = kbd_cmd (PA_CMD(WR_KEYBOARD,keyled));
	printf ("setled value returned 0x%x\n", c1);
    }
#endif

    /* get the key according to a map */
    if (keymod & KEY_MOD_SHIFTED) {
	c = kbd_shiftmap[c];		/* shift everything */
    }
    else if (keymod & KEY_MOD_CAPSLOCK) {
	c = kbd_keymap[c];
	if (islower(c))
	    c = toupper(c);		/* shift alphabetics only */
    }
    else {
	c = kbd_keymap[c];		/* normal keymap */
    }
    if (keymod & KEY_MOD_CONTROL)
	c &= 0x1f;

    return c;
}
