#define PROBE	1
/*	if_adx.c	*/

#include "adx.h"
#if NADX > 0

/*
 * ADAX SDMA synchronous communications board driver.
 *
 *	Drew D. Perkins
 *	Carnegie Mellon University
 *
 * The flags field of the config file is interpreted like so:
 * bits		meaning
 * ----		-------
 * 0		Soft Carrier
 * 1		Internal Clock
 * 8-10		Receive DMA channel
 * 12-14	Transmit DMA channel
 *
 * Routines defined in this module are referenced via a few structures:
 * Called via iocc_driver structure by autoconf and interrupt (locore) system:
 *	adx_probe()	- probes for existence of device
 *	adx_attach()	- attaches device to system
 *	adx_int()	- handles device interrupt
 *
 * Called via lapb_driver structure by LAPB layer:
 *	adx_init()	- initialize device
 *	adx_down()	- shutdown device
 *	adx_reset()	- reset device (after bus reset)
 *	adx_output()	- output a frame
 *
 * Routines in LAPB called from this module are:
 *	lapb_attach()	- attach device to LAPB system
 *	lapb_start()	- start output of next frame
 *	lapb_input()	- input a frame
 *	lapb_modem()	- device modem controls changed state
 *
 * TODO:
 *	Enter Hunt Phase after rx overrun doesn't seem to work.
 *	BCNT transmit mode doesn't work.
 */

#if MACH
#include "../ca/rosetta.h"
#else
#include "../machine/pte.h"
#endif

#include "param.h"
#include "systm.h"
#include "mbuf.h"
#include "buf.h"
#include "protosw.h"
#include "socket.h"
#include "vmmac.h"
#include "kernel.h"
#include "ioctl.h"
#include "errno.h"
#include "uio.h"

#include "../net/if.h"

#include "../net/if_lapbvar.h"

#include "../ca/io.h"
#include "../caio/ioccvar.h"
#include "../caio/dmareg.h"
#include "../caio/dmavar.h"

#include "../caif/mpsccreg.h"
#include "../caif/if_adxvar.h"


#define DEBUG_PROC	0x1	/* Procedure entrance/exit */
#define DEBUG_SOFTERROR	0x2	/* Soft errors */
#define DEBUG_HARDERROR	0x4	/* Hard errors */
#define DEBUG_DMA	0x8	/* DMA operations */
#define DEBUG_DMATCW	0x10	/* DMA TCW info */

#if 0
#define DEBUG	DEBUG_HARDERROR|DEBUG_SOFTERROR|DEBUG_PROC|DEBUG_DMA|DEBUG_DMATCW
#define DEBUG	DEBUG_HARDERROR|DEBUG_SOFTERROR|DEBUG_PROC
#define DEBUG	DEBUG_HARDERROR|DEBUG_SOFTERROR
#define DEBUG	DEBUG_HARDERROR
#endif

#define DEBUG	DEBUG_HARDERROR

#ifdef DEBUG
u_char adx_debug = DEBUG;	/* Control's printf's for debugging */
#define DEBUGF(cond,stmt) if (cond) stmt	/* do the stmt (printf) */
#else
#define DEBUGF(cond,stmt)		/* do nothing if not debugging */
#endif

int adx_probe(), adx_attach(), adx_init(), adx_output(), adx_int(), adx_put();
int adx_dma_done();
void adx_reset(), adx_down(), adx_dtr(), adx_dcd_timeout(), adx_dtr_timeout();
void adx_dma_setup();
struct mbuf *adx_get();
struct ifnet *lapb_attach();

/*
 * Standard device locations.
 */
caddr_t adx_std[] = {
    (caddr_t) 0xf0000380,	/* Card 1 */
    (caddr_t) 0xf00003c0,	/* Card 2 */
    (caddr_t) 0xf0000100,	/* Card 3 */
    (caddr_t) 0xf0000120,	/* Card 4 */
    0				/* No more... */
};

/*
 * Describe each available device to the system.
 */
struct iocc_device *adx_info[NADX];

/*
 * Describe the driver to the system.
 */
struct iocc_driver adxdriver = {
/*  probe, slave, attach, dgo, addr, dname, dinfo,
    mname, minfo, intr, csr, chanrelse, flags */
    adx_probe, 0, adx_attach, 0, adx_std, "adx", adx_info,
    0, 0, adx_int, 0, 0, 0
};

/*
 * Describe the driver to LAPB.
 */
struct lapb_driver adx_lapbdriver = {
/*  init, down, reset, output */
    adx_init, adx_down, adx_reset, adx_output
};

/*
 * Software status associated with each device.  One per card.
 */
struct adx_softc adx_softc[NADX];
char adx_softCAR[NADX];		/* Software DCD override */


/*
 * adx_probe - probe for the device.
 *
 * Called by autoconfiguration code to determine if the device is present.
 * We don't support probe interrupts yet.
 */
int adx_probe(a)
    caddr_t a;
{
#if PROBE
    struct adx_device *addr = (struct adx_device *) a;
    int i;
#endif

    DEBUGF(adx_debug & DEBUG_PROC,
	printf("adx probe begin\n");
    );

#ifdef PROBE
    /*
     * Reset the PC-SDMA MPSCC chip.
     */
    i = addr->chA_csr;		/* Make sure register pointer = 0 */
    addr->chA_csr = MPSCC_RST_CHANNEL;
    i = addr->chB_csr;		/* Make sure register pointer = 0 */
    addr->chB_csr = MPSCC_RST_CHANNEL;

    /*
     * Waste a little time.
     */
    DELAY(10000);

    /*
     * Set DMA mode/Nonvectored Interrupts/Rx INT mask.
     */
    addr->chA_csr = MPSCC_CR2;
    addr->chA_csr = MPSCC_DMA_MODE0|MPSCC_DMA_PRI1|MPSCC_IASR_MODE2;

    /*
     * Set interrupt vector.
     */
    addr->chB_csr = MPSCC_CR2;
    addr->chB_csr = 0x00;

    /*
     * Set Condition Affects Vector mode.
     */
    addr->chB_csr = MPSCC_CR1;
    addr->chB_csr = MPSCC_CAV;

    /*
     * Set mode/x1 clk/sync (must be done first).
     */
    addr->chA_csr = MPSCC_CR4;
    addr->chA_csr = MPSCC_CLK_x1|MPSCC_SYNC_ENA;

    /*
     * Set DTR for B high to enable DMA/interrupts.
     */
    addr->chB_csr = MPSCC_CR5;
    addr->chB_csr = MPSCC_DTR;

    /*
     * Enable Transmitter and External Interrupts.
     */
    addr->chA_csr = MPSCC_CR1;
    addr->chA_csr = MPSCC_ENA_TxINT|MPSCC_ENA_EXTINT;
    addr->chA_csr = MPSCC_CR1;
    addr->chA_csr = MPSCC_ENA_TxINT|MPSCC_ENA_EXTINT|MPSCC_ENA_BCNT_MODE;
    addr->chA_csr = 0x02;
    addr->chA_csr = 0x0;

    /*
     * Set transmit character length to 8 bits, enable transmitter,
     */
    addr->chA_csr = MPSCC_CR5;
    addr->chA_csr = MPSCC_Tx_8BITS|MPSCC_Tx_ENA;
#endif

    /*
     * Waste a little time.
     */
    PROBE_DELAY(10000);

    /*
     * Reset the PC-SDMA MPSCC chip.
     */
    i = addr->chA_csr;		/* Make sure register pointer = 0 */
    addr->chA_csr = MPSCC_RST_CHANNEL;
    i = addr->chB_csr;		/* Make sure register pointer = 0 */
    addr->chB_csr = MPSCC_RST_CHANNEL;

    DEBUGF(adx_debug & DEBUG_PROC,
	printf("adx probe end\n");
    );

#if PROBE
    return (PROBE_OK);
#else
    return (PROBE_NOINT);
#endif
}

 
/*
 * adx_attatch - attach device to system.
 *
 * The interface exists, make it available by attaching to LAPB
 * layer.  LAPB will initialize the interface when it is ready to
 * accept frames.
 */
adx_attach(iod)
    register struct iocc_device *iod;
{
    register struct adx_softc *axc = &adx_softc[iod->iod_unit];

    axc->adx_dcd_tmo = ADX_DFLT_DCD_TMO;
    axc->adx_dtr_tmo = ADX_DFLT_DTR_TMO;

    /*
     * Initialize driver specific variables.
     */
    axc->adx_flags = 0;		/* Reset flags */
    if (iod->iod_flags == 0) {
	printf("adx%d: no flags specified in configuration file!\n");
	return;
    }
    adx_softCAR[iod->iod_unit] = iod->iod_flags & 0x01;
    if ((iod->iod_flags >> 1) & 1)
	axc->adx_flags |= ADX_INTERNALCLK;
    axc->adx_tx_dma_chan = (iod->iod_flags >> 12) & 7;
    axc->adx_rx_dma_chan = (iod->iod_flags >> 8) & 7;
    printf("adx%d: tx dma channel %d; rx dma channel %d; %sternal clock%s\n",
	   iod->iod_unit, axc->adx_tx_dma_chan, axc->adx_rx_dma_chan,
	   (axc->adx_flags & ADX_INTERNALCLK) ? "in" : "ex",
	   adx_softCAR[iod->iod_unit] ? "; hard-wired" : "");

    /*
     * Attach to LAPB layer.
     * Don't include trailer in MTU.
     */
    axc->adx_if = lapb_attach(&adx_lapbdriver, "adx", iod->iod_unit,
			      ADX_MTU - 2);
}


/*
 * adx_reset - reset interface after bus reset.
 */
void adx_reset(unit)
    register unsigned int unit;
{
    register struct adx_softc *axc = &adx_softc[unit];

    DEBUGF(adx_debug & DEBUG_PROC,
	printf("adx%d reset begin\n", unit);
    );

    /*
     * We're no longer initialized and running.
     */
    axc->adx_if->if_flags &= ~IFF_RUNNING;
    axc->adx_flags &= ~(ADX_STARTED|ADX_DCDWARN);

    DEBUGF(adx_debug & DEBUG_PROC,
	printf("adx%d reset end\n", unit);
    );
}


/*
 * adx_init - initialize interface.
 */
int adx_init(unit)
    register unsigned int unit;
{
    register struct adx_softc *axc = &adx_softc[unit];
    register struct adx_device *addr =
	(struct adx_device *) adx_info[unit]->iod_addr;
    register int i, s;

    /*
     * Check to see that the device hasn't already been started.
     */
    if (axc->adx_flags & ADX_STARTED)
	return (0);

    DEBUGF(adx_debug & DEBUG_PROC,
	printf("adx%d init begin\n", unit);
    );

    s = splimp();

    /*
     * Reset the PC-SDMA MPSCC chip.
     */
    i = addr->chA_csr;		/* Make sure register pointer = 0 */
    addr->chA_csr = MPSCC_RST_CHANNEL;
    i = addr->chB_csr;		/* Make sure register pointer = 0 */
    addr->chB_csr = MPSCC_RST_CHANNEL;

    /*
     * Waste a little time.
     */
    DELAY(10000);

    /*
     * Set DMA mode/Nonvectored Interrupts/Rx INT mask.
     */
    addr->chA_csr = MPSCC_CR2;
    addr->chA_csr = MPSCC_DMA_MODE3|MPSCC_DMA_PRI1|MPSCC_IASR_MODE2|MPSCC_RxINT_MASK|MPSCC_ENA_SYNCB;

    /*
     * Set interrupt vector.
     */
    addr->chB_csr = MPSCC_CR2;
    addr->chB_csr = 0x00;

    /*
     * Set Condition Affects Vector mode.
     */
    addr->chB_csr = MPSCC_CR1;
    addr->chB_csr = MPSCC_CAV;

    /*
     * Set sdlc mode/x1 clk/sync mode/no parity (must be done first).
     */
    addr->chA_csr = MPSCC_CR4;
    addr->chA_csr = MPSCC_SYNC_SDLC|MPSCC_SYNC_ENA;

    /*
     * Set DMA request on 1st received char only.
     * Enable Transmitter and External Interrupts.
     */
    addr->chA_csr = MPSCC_CR1;
#ifdef BCNT
    addr->chA_csr = MPSCC_ENA_FRSTCHR|MPSCC_ENA_TxINT|MPSCC_ENA_EXTINT|
	MPSCC_ENA_BCNT_MODE;
    addr->chA_csr = 0xff;
    addr->chA_csr = 0xff;
#else
    addr->chA_csr = MPSCC_ENA_FRSTCHR|MPSCC_ENA_TxINT|MPSCC_ENA_EXTINT;
#endif

    /*
     * Set receive character length to 8 bits, enter hunt phase,
     * enable receiver CRC checking, enable receiver.
     */
    addr->chA_csr = MPSCC_CR3;
#ifdef AUTO_ENA
    addr->chA_csr = MPSCC_Rx_8BITS|MPSCC_AUTO_ENA|MPSCC_ENTER_HUNT|MPSCC_RxCRC_ENA|MPSCC_Rx_ENA;
#else
    addr->chA_csr = MPSCC_Rx_8BITS|MPSCC_ENTER_HUNT|MPSCC_RxCRC_ENA|MPSCC_Rx_ENA;
#endif

    /*
     * Assert DTR, set transmit character length to 8 bits, 
     * enable transmitter, select CCITT CRC polynomial, assert RTS,
     * and enable tx CRC generation.
     */
    addr->chA_csr = MPSCC_CR5;
    addr->chA_csr = MPSCC_DTR|MPSCC_Tx_8BITS|MPSCC_POLY_SDLC|MPSCC_Tx_ENA|MPSCC_RTS|MPSCC_TxCRC_ENA;

    /*
     * Set SDLC link address.
     */
#define LINK_ADDRESS	0	/* XXX Doesn't matter at this point */
    addr->chA_csr = MPSCC_CR6;
    addr->chA_csr = LINK_ADDRESS;

    /*
     * Set the SDLC flag ==> 01111110.
     */
    addr->chA_csr = MPSCC_CR7;
    addr->chA_csr = 0x7e;

    /*
     * Set DTR for B high to enable DMA/interrupts.
     */
    addr->chB_csr = MPSCC_CR5;
    addr->chB_csr = MPSCC_DTR;

    /*
     * Now for some total bogosity.  Read the contents of the status
     * registers and clear some interrupt conditions.
     * First, get Modified Vector from sr2, chan B.
     */
    addr->chB_csr = MPSCC_CR2;
    i = (int) addr->chB_csr;

    /*
     * Now get External Status status from sr0, chan A.
     */
    i = (int) addr->chA_csr;

    /*
     * Now reset everything in sight.
     */
    addr->chA_csr = MPSCC_RST_EXTINT;
    addr->chA_csr = MPSCC_RST_TxINTP;
    addr->chA_csr = MPSCC_ERROR_RST;
    addr->chA_csr = MPSCC_ENA_INTRxC;
    addr->chA_csr = MPSCC_END_OF_INT;

    /*
     * Get current state of DCD.  If already asserted, inform LAPB.  Else
     * start dcd timeout timer.
     */
    i = (int) addr->chA_csr;	/* Get External Status Status */
    if (i & MPSCC_DCD) {
	axc->adx_flags |= ADX_DCD;
	lapb_modem(axc->adx_if, 1);
    } else {
	axc->adx_flags &= ~ADX_DCD;
	if (axc->adx_dcd_tmo)
	    timeout(adx_dcd_timeout, unit, mstohz(axc->adx_dcd_tmo));
    }

#ifdef PROVIDE_CLOCK
    /*
     * Set up the counter/timer to provide us with a clock.
     */
    addr->ctr_csr = 0x36;  /*  setting the clock                     */
#if BAUD == 9600 */
    addr->ctr_0 = 0xf2;    /*  setting low order byte of counter     */
    addr->ctr_0 = 0x01;    /*  setting high order byte               */
#endif
#if BAUD == 76800
    addr->ctr_0 = 0x3e;    /*  setting low order byte of counter     */
    addr->ctr_0 = 0x00;    /*  setting high order byte               */
#endif
#if BAUD == 57600
    addr->ctr_0 = 0x5d;    /*  setting low order byte of counter     */
    addr->ctr_0 = 0x00;    /*  setting high order byte               */
#endif
#if BAUD == 38400
    addr->ctr_0 = 0x7c;    /*  setting low order byte of counter     */
    addr->ctr_0 = 0x00;    /*  setting high order byte               */
#endif
#if BAUD == 19200
    addr->ctr_0 = 0xf8;    /*  setting low order byte of counter     */
    addr->ctr_0 = 0x00;    /*  setting high order byte               */
#endif
#if BAUD == 9600
    addr->ctr_0 = 0xf0;    /*  setting low order byte of counter     */
    addr->ctr_0 = 0x01;    /*  setting high order byte               */
#endif
#endif

    /*
     * Setting up the DMA paramters.
     * DMA channel assignments for unit 0, 1, and 2 are hard coded.
     */
    if (unit == 0) {
	axc->adx_tx_dma_chan = DMA_CHAN1; /* 8-bit transmit channel */
	axc->adx_rx_dma_chan = DMA_CHAN3; /* 8-bit receive channel */
    } else if (unit == 1) {
	axc->adx_tx_dma_chan = DMA_CHAN2; /* 8-bit transmit channel */
	axc->adx_rx_dma_chan = DMA_CHAN0; /* 8-bit receive channel */
    } else {
	axc->adx_tx_dma_chan = DMA_CHAN6; /* 16-bit transmit channel */
	axc->adx_rx_dma_chan = DMA_CHAN7; /* 16-bit receive channel */
    }

    /*
     * Allocate transmit and receive buffers.  Buffers must be
     * contiguous and must be atleast ADX_MTU in length for interfaces
     * using the 8-bit DMA controller.  The buffers for interfaces using
     * the 16-bit DMA controller must be twice this long.
     * Since 2*ADX_MTU > MCLBYTES, we may need more than one cluster.
     * Therefore, we can't use the MCLALLOC macro because it can only
     * allocate 1 cluster (currently).  Instead, we call m_clalloc()
     * directly to allocate mbuf page space and avoid mucking with mbufs
     * and clusters.
     */
#ifndef MCLBYTES
#define MCLBYTES CLBYTES
#endif
#ifndef MCLSHIFT
#define MCLSHIFT CLSHIFT
#endif /* !MCLSHIFT */
#define MSPCALLOC(dma_channel)\
    ((caddr_t) m_clalloc(((((ADX_MTU + ADX_SLOP) *\
			    (DMA_CTLR_1(dma_channel) ? 1 : 2)) +\
			   (MCLBYTES - 1)) >> MCLSHIFT),\
			 MPG_SPACE, M_DONTWAIT))

    if (axc->adx_rx_buffer0 == NULL) {
	axc->adx_rx_buffer0 = MSPCALLOC(axc->adx_rx_dma_chan);
	if (axc->adx_rx_buffer0 == NULL) {
	    splx(s);
	    printf("adx%d: can't allocate rx buffer\n", unit);
	    return (-1);
	}
    }
    if (axc->adx_rx_buffer1 == NULL) {
	axc->adx_rx_buffer1 = MSPCALLOC(axc->adx_rx_dma_chan);
	if (axc->adx_rx_buffer1 == NULL) {
	    splx(s);
	    printf("adx%d: can't allocate rx buffer\n", unit);
	    return (-1);
	}
    }
    if (axc->adx_tx_buffer == NULL) {
	axc->adx_tx_buffer = MSPCALLOC(axc->adx_tx_dma_chan);
	if (axc->adx_tx_buffer == NULL) {
	    splx(s);
	    printf("adx%d: can't allocate tx buffer\n", unit);
	    return (-1);
	}
    }
    axc->adx_which_rx_buffer = 0; /* Start with first one */

    /*
     * We're running!
     */
    axc->adx_if->if_flags |= IFF_RUNNING;
    axc->adx_flags |= ADX_STARTED;

    /*
     * Set up DMA to receive frames.
     */
    adx_dma_setup(axc->adx_rx_dma_chan,
		  axc->adx_which_rx_buffer ?
		   axc->adx_rx_buffer1 :
		   axc->adx_rx_buffer0,
		  ADX_MTU + ADX_SLOP, DM_READ);

    splx(s);

    DEBUGF(adx_debug & DEBUG_PROC,
	printf("adx%d init end\n", unit);
    );

    return (0);
}

 
/*
 * adx_down - shut down interface.
 */
void adx_down(unit)
    register unsigned int unit;
{
    register struct adx_softc *axc = &adx_softc[unit];
    register struct adx_device *addr =
	(struct adx_device *) adx_info[unit]->iod_addr;
    register int i, s;

    DEBUGF(adx_debug & DEBUG_PROC,
	printf("adx%d down begin\n", unit);
    );

    s = splimp();

    /*
     * Remember that we're no longer initialized and running.
     */
    axc->adx_if->if_flags &= ~IFF_RUNNING;
    axc->adx_flags &= ~(ADX_STARTED|ADX_DCDWARN);

    /*
     * Reset the PC-SDMA MPSCC chip.
     */
    i = addr->chA_csr;		/* Make sure register pointer = 0 */
    addr->chA_csr = MPSCC_RST_CHANNEL;

    /*
     * Cancel pending timeouts.
     */
    untimeout(adx_dcd_timeout, unit);
    untimeout(adx_dtr_timeout, unit);

    /*
     * Waste a little time.
     */
    DELAY(100);

    splx(s);

    DEBUGF(adx_debug & DEBUG_PROC,
	printf("adx%d down end\n", unit);
    );
}


/*
 * adx_dtr - Set the state of the DTR pin.
 */
void adx_dtr(unit, dtr)
    register unsigned int unit;
    int dtr;
{
    register struct adx_device *addr =
	(struct adx_device *) adx_info[unit]->iod_addr;

    DEBUGF(adx_debug & DEBUG_PROC,
	printf("adx%d dtr begin\n", unit);
    );

    if (dtr) {
	/*
	 * Set DTR/RTS high.
	 */
	addr->chA_csr = MPSCC_CR5;
	addr->chA_csr = MPSCC_DTR|MPSCC_Tx_8BITS|MPSCC_POLY_SDLC|
	    MPSCC_Tx_ENA|MPSCC_RTS|MPSCC_TxCRC_ENA;
    } else {
	/*
	 * Set DTR/RTS low.
	 */
	addr->chA_csr = MPSCC_CR5;
	addr->chA_csr = MPSCC_Tx_8BITS|MPSCC_POLY_SDLC|MPSCC_Tx_ENA|
	    MPSCC_TxCRC_ENA;
    }

    DEBUGF(adx_debug & DEBUG_PROC,
	printf("adx%d dtr end\n", unit);
    );
}


/*
 * adx_output - start output of an mbuf chain.
 *
 * Should be called at splimp or higher(?).
 */
int adx_output(unit, m)
    register unsigned int unit;
    struct mbuf *m;
{
    register struct adx_softc *axc = &adx_softc[unit];
    register struct adx_device *addr =
	(struct adx_device *) adx_info[unit]->iod_addr;
    register int i;
    int len;
    caddr_t bufaddr;

    DEBUGF(adx_debug & DEBUG_PROC,
	printf("adx%d output begin\n", unit);
    );

    /*
     * Transfer frame from mbuf chain into output buffer and free mbufs.
     */
    bufaddr = (caddr_t) axc->adx_tx_buffer;
    len = adx_put(unit, m, bufaddr, DMA_CTLR_1(axc->adx_tx_dma_chan) ? 1 : 2);
    m_freem(m);

    axc->adx_oactive = 1;	/* Mark as busy transmitting */
    axc->adx_tx_retries = 0;	/* Reset retry count */
    axc->adx_tx_len = i = len;	/* Remember length for retry */

#ifndef BCNT
    bufaddr += DMA_CTLR_1(axc->adx_tx_dma_chan) ? 1 : 2;
    i--;			/* Account for first character primer */
#endif

    addr->chA_csr = MPSCC_RST_TxINTP;

    /*
     * Set up DMA channel to supply frame on demand from the ADX.
     */
    adx_dma_setup(axc->adx_tx_dma_chan, bufaddr, i, DM_WRITE);

#ifdef BCNT
    /*
     * Start the ball rolling by writing the tx length registers.
     */
    addr->chA_csr = MPSCC_RST_TxUNDER;
    addr->chA_csr = MPSCC_RST_EXTINT|MPSCC_CR1;
    addr->chA_csr = MPSCC_ENA_FRSTCHR|MPSCC_ENA_TxINT|MPSCC_ENA_EXTINT|
	MPSCC_ENA_BCNT_MODE;		    
    addr->chA_csr = i & 0xff;
    addr->chA_csr = (i >> 8) & 0xff;
#else
    /*
     * Start the ball rolling by writing the first character.
     */
    addr->chA_data = *bufaddr;
    addr->chA_csr = MPSCC_RST_TxUNDER;
#endif

    DEBUGF(adx_debug & DEBUG_PROC,
	printf("adx%d output end\n", unit);
    );

    return (len);
}


/*
 * adx_int - interrupt handler.
 */
int adx_int(unit)
    int unit;
{
    register struct adx_softc *axc = &adx_softc[unit];
    register struct adx_device *addr =
	(struct adx_device *) adx_info[unit]->iod_addr;
    register int sr0, sr1, sr2, sr3, sr4, length, ret;
    caddr_t bufaddr;

    ret = 1;
top:
    /*
     * Read the modified interrupt vector, register 2B, followed,
     * by register 0.  The vector must be read first to make RR0 valid.
     * RR1 can then be read and is valid because RR0 was read.
     */
    /* Get Modified Vector from sr2, chan B. */
    addr->chB_csr = MPSCC_CR2;
    sr2 = (int) addr->chB_csr;

    /* Get External Status status from sr0, chan A. */
    sr0 = (int) addr->chA_csr;

    /* Get Special Receive Status from sr1, chan A. */
    addr->chA_csr = MPSCC_CR1;
    sr1 = (int) addr->chA_csr;

    if (!(sr0 & MPSCC_INT_PENDING)) {	/* Spurious interrupt? */
	DEBUGF(adx_debug & DEBUG_PROC,
	       printf("adx%d int end\n", unit);
	);
	return (ret);
    }
    ret = 0;

    switch (sr2 & 0x07) {	/* Mask unwanted bits just in case */
	case MPSCC_CHNA_TxIP:	/* Transmitter buffer empty */
	    DEBUGF(adx_debug & DEBUG_PROC,
		printf("adx%d int TxIP begin\n", unit);
	    );

	    addr->chA_csr = MPSCC_RST_TxINTP;
	    break;

	case MPSCC_CHNA_ESIP:	/* External status change */
	    DEBUGF(adx_debug & DEBUG_PROC,
		printf("adx%d int ESIP(sr0=%x sr1=%x) begin\n",
		       unit, sr0, sr1);
	    );

	    /*
	     * Check Break/Abort status.
	     */
	    if (sr0 & MPSCC_BREAK) {
		if (!(axc->adx_flags & ADX_ABORT)) {
		    /* Now aborting */
		    DEBUGF(adx_debug & DEBUG_SOFTERROR,
			   printf("adx%d aborting\n", unit);
		    );
		    axc->adx_flags |= ADX_ABORT;
		}
	    }
	    else {
		if (axc->adx_flags & ADX_ABORT) {
		    /* No longer aborting */
		    DEBUGF(adx_debug & DEBUG_SOFTERROR,
			   printf("adx%d not aborting\n", unit);
		    );
		    axc->adx_flags &= ~ADX_ABORT;

		    addr->chA_csr = MPSCC_ERROR_RST;
		    addr->chA_csr = MPSCC_ENA_INTRxC;

		    /*
		     * Set up DMA to receive frames.
		     */
		    adx_dma_setup(axc->adx_rx_dma_chan,
				  axc->adx_which_rx_buffer ?
				   axc->adx_rx_buffer1 :
				   axc->adx_rx_buffer0,
				ADX_MTU + ADX_SLOP, DM_READ);
		}
	    }

	    /*
	     * Check Sync status.
	     */
	    if (sr0 & MPSCC_SYNC) {
		if (axc->adx_flags & ADX_SYNC) {
		    DEBUGF(adx_debug & DEBUG_PROC,
			printf("adx%d lost sync\n", unit);
		    );
		    axc->adx_flags &= ~ADX_SYNC;
		}
	    }
	    else {
		if (!(axc->adx_flags & ADX_SYNC)) {
		    DEBUGF(adx_debug & DEBUG_PROC,
			printf("adx%d rx sync\n", unit);
		    );
		    axc->adx_flags |= ADX_SYNC;
		}
	    }

	    /*
	     * Check CTS status.
	     */
	    if (sr0 & MPSCC_CTS) {
		if (!(axc->adx_flags & ADX_CTS)) {
		    DEBUGF(adx_debug & DEBUG_PROC,
			printf("adx%d cts raised\n", unit);
		    );
		    axc->adx_flags |= ADX_CTS;
		}
	    }
	    else {
		if (axc->adx_flags & ADX_CTS) {
		    DEBUGF(adx_debug & DEBUG_PROC,
			printf("adx%d cts dropped\n", unit);
		    );
		    axc->adx_flags &= ~ADX_CTS;
		}
	    }

	    /*
	     * Check DCD status.
	     */
	    if (sr0 & MPSCC_DCD) {
		if (!(axc->adx_flags & ADX_DCD)) {
		    DEBUGF(adx_debug & DEBUG_PROC,
			printf("adx%d dcd raised\n", unit);
		    );
		    axc->adx_flags |= ADX_DCD;

		    /* Raise DTR */
		    adx_dtr(unit, 1);
		    untimeout(adx_dtr_timeout, unit);
		    untimeout(adx_dcd_timeout, unit);

		    /* Inform LAPB */
		    lapb_modem(axc->adx_if, 1);
		}
	    }
	    else {
		if (axc->adx_flags & ADX_DCD) {
		    DEBUGF(adx_debug & DEBUG_PROC,
			printf("adx%d dcd dropped\n", unit);
		    );
		    axc->adx_flags &= ~ADX_DCD;

		    /*
		     * Toggle DTR.
		     */
		    adx_dtr(unit, 0);
		    timeout(adx_dtr_timeout, unit, mstohz(axc->adx_dtr_tmo));

		    /* Inform LAPB */
		    lapb_modem(axc->adx_if, 0);
		}
	    }

	    /*
	     * Check for Tx Buffer Empty Condition.
	     */
	    if ((sr0 & MPSCC_TxBUF_EMPTY) &&
		axc->adx_oactive) {
		DEBUGF(adx_debug & DEBUG_PROC,
		    printf("adx%d tx buf empty\n", unit);
		);
		addr->chA_csr = MPSCC_RST_TxINTP;
	    }

	    /*
	     * Check for All Sent Condition.
	     */
	    if ((sr1 & MPSCC_ALL_SENT) &&
		axc->adx_oactive) {
		DEBUGF(adx_debug & DEBUG_PROC,
		    printf("adx%d tx all sent\n", unit);
		);

		/*
		 * Check sr3 and sr4.
		 */
		addr->chA_csr = MPSCC_CR3;
		sr3 = (int) addr->chA_csr;
		addr->chA_csr = MPSCC_CR4;
		sr4 = (int) addr->chA_csr;
		addr->chA_csr = MPSCC_RST_TxINTP;
		addr->chA_csr = MPSCC_RST_EXTINT;

#ifdef BCNT
		if (sr3 || sr4) {
		    DEBUGF(adx_debug & DEBUG_HARDERROR,
			printf("adx%d tx not all sent 0x%x\n",
			       unit, (sr4 << 8) | sr3);
		    );
		    axc->adx_if->if_oerrors++;
		}
#endif

		/*
		 * Check DMA status.  If DMA not complete and not too many
		 * retries, retry current frame.  Else, start next LAPB frame.
		 */
		if (length = adx_dma_done(axc->adx_tx_dma_chan)) {
		    axc->adx_if->if_oerrors++;
		    if (axc->adx_tx_retries++ >= ADX_MAX_TX_RETRIES) {
			DEBUGF(adx_debug & DEBUG_HARDERROR,
			       printf("adx%d max tx retries exceeded, dma not done (0x%x)!\n",
				      unit, length);
			);
		    }
		    else {
			DEBUGF(adx_debug & DEBUG_SOFTERROR,
			       printf("adx%d dma not done (0x%x)!  Retrying...\n",
				      unit, length);
			);
			bufaddr = (caddr_t) axc->adx_tx_buffer;
			length = axc->adx_tx_len;

#ifndef BCNT
			bufaddr += DMA_CTLR_1(axc->adx_tx_dma_chan) ? 1 : 2;
			length--; /* Account for first character primer */
#endif

			/*
			 * Set up DMA channel to supply frame on demand
			 * from the ADX.
			 */
			adx_dma_setup(axc->adx_tx_dma_chan, bufaddr, length,
				      DM_WRITE);

#ifdef BCNT
			/*
			 * Start the ball rolling by writing the tx length
			 * registers.
			 */
			addr->chA_csr = MPSCC_RST_TxUNDER;
			addr->chA_csr = MPSCC_RST_EXTINT|MPSCC_CR1;
			addr->chA_csr = MPSCC_ENA_BCNT_MODE|MPSCC_ENA_FRSTCHR|
			    MPSCC_ENA_TxINT|MPSCC_ENA_EXTINT;
			addr->chA_csr = length & 0xff;
			addr->chA_csr = (length >> 8) & 0xff;
#else
			/*
			 * Start the ball rolling by writing the first
			 * character.
			 */
			addr->chA_data = *bufaddr;
			addr->chA_csr = MPSCC_RST_TxUNDER;
#endif
			break;
		    }
		}

		/* mark as done transmitting */
		axc->adx_oactive = 0;
		lapb_start(axc->adx_if);
	    } else {
		addr->chA_csr = MPSCC_RST_EXTINT;
	    }
            break;

	case MPSCC_CHNA_RxIP:	/* Received character available */
	    DEBUGF(adx_debug & DEBUG_PROC,
		printf("adx%d int RxIP begin\n", unit);
	    );

	    addr->chA_csr = MPSCC_RST_EXTINT;
	    break;

	case MPSCC_CHNA_SRxIP:	/* Special receive condition */
	    DEBUGF(adx_debug & DEBUG_PROC,
		printf("adx%d int SRxIP begin\n", unit);
	    );

	    /*
	     * Get length of received frame.  Account for checksum.
	     */
	    length = ADX_MTU + ADX_SLOP - adx_dma_done(axc->adx_rx_dma_chan) - 2;

	    /*
	     * Start input on next frame using *other* buffer.
	     */
	    adx_dma_setup(axc->adx_rx_dma_chan,
			  axc->adx_which_rx_buffer ?
			   axc->adx_rx_buffer0 :
			   axc->adx_rx_buffer1,
			  ADX_MTU + ADX_SLOP, DM_READ);
	    addr->chA_csr = MPSCC_ERROR_RST;

	    /*
	     * On rx overrun,  check for additional rx overrun, then
	     * reenter hunt phase to abort reception of current frame.
	     */
	    if (sr1 & MPSCC_Rx_OVERRUN) {
		/* Get Special Receive Status from sr1, chan A. */
		addr->chA_csr = MPSCC_CR1;
		sr1 = (int) addr->chA_csr;
		if (sr1 & MPSCC_Rx_OVERRUN)
		    addr->chA_csr = MPSCC_ERROR_RST;

		addr->chA_csr = MPSCC_CR3;
#ifdef AUTO_ENA
		addr->chA_csr = MPSCC_Rx_8BITS|MPSCC_AUTO_ENA|MPSCC_ENTER_HUNT|MPSCC_RxCRC_ENA|MPSCC_Rx_ENA;
#else
		addr->chA_csr = MPSCC_Rx_8BITS|MPSCC_ENTER_HUNT|MPSCC_RxCRC_ENA|MPSCC_Rx_ENA;
#endif
	    }

	    addr->chA_csr = MPSCC_ENA_INTRxC;
	    addr->chA_csr = MPSCC_END_OF_INT;

	    /*
	     * Switch buffers.
	     */
	    axc->adx_which_rx_buffer =
		!axc->adx_which_rx_buffer;

	    /*
	     * If an invalid frame, just ignore it, else process it.
	     */
	    if ((sr1 & (MPSCC_EOF|MPSCC_CRC_FRAMING)) ==
		(MPSCC_EOF|MPSCC_CRC_FRAMING)) {
		DEBUGF(adx_debug & DEBUG_SOFTERROR,
		    printf("adx%d crc error\n", unit);
		);
		axc->adx_if->if_ierrors++;
	    }
	    else if (sr1 & MPSCC_Rx_OVERRUN) {
		DEBUGF(adx_debug & DEBUG_SOFTERROR,
		    printf("adx%d rx overrun\n", unit);
		);
		axc->adx_if->if_ierrors++;
	    }
	    else if (length == 0) {
		DEBUGF(adx_debug & DEBUG_SOFTERROR,
		    printf("adx%d rx length = 0\n", unit);
		);
		axc->adx_if->if_ierrors++;
	    } else if (sr1 & MPSCC_EOF) {
		struct mbuf *m;

		/*
		 * "Get" frame into mbuf's.  Frame is returned with
		 * struct ifnet *axc->adc_if at head of frame.
		 */
		m = adx_get(axc->adx_if, (axc->adx_which_rx_buffer ?
					  axc->adx_rx_buffer0 :
					  axc->adx_rx_buffer1),
			    length,
			    DMA_CTLR_1(axc->adx_rx_dma_chan) ? 1 : 2);
		if (m != 0)
		    lapb_input(axc->adx_if, m);
		else {
		    axc->adx_if->if_ierrors++;
		    DEBUGF(adx_debug & DEBUG_SOFTERROR,
			   printf("adx%d no mbuf rx'd\n",
				  unit);
			   );
		}
	    }

	    /*
	     * Check DCD status.
	     */
	    if (sr0 & MPSCC_DCD) {
		if (!(axc->adx_flags & ADX_DCD)) {
		    DEBUGF(adx_debug & DEBUG_HARDERROR,
			printf("adx%d missed dcd raise int\n", unit);
		    );
		    axc->adx_flags |= ADX_DCD;

		    /* Raise DTR */
		    adx_dtr(unit, 1);
		    untimeout(adx_dtr_timeout, unit);
		    untimeout(adx_dcd_timeout, unit);

		    /* Inform LAPB */
		    lapb_modem(axc->adx_if, 1);
		}
	    }
	    goto ret;

	default:
	    DEBUGF(adx_debug & DEBUG_PROC,
		printf("adx%d int unknown begin\n", unit);
	    );

	    addr->chA_csr = MPSCC_RST_EXTINT;
	    break;
    }				/* end switch */
    addr->chA_csr = MPSCC_END_OF_INT;

ret:
    goto top;
}

 
/*
 * adx_dcd_timeout - DCD timeout routine.
 *
 * Called sometime after DTR has been raised but DCD has not been asserted.
 * Drop's DTR and schedules a timeout for it to be raised.
 */
void adx_dcd_timeout(unit)
    int unit;
{
    register struct adx_softc *axc = &adx_softc[unit];
    int s = splimp();

    DEBUGF(adx_debug & DEBUG_PROC,
	   printf("adx%d dtr begin\n", unit);
    );

    if ((adx_softc[unit].adx_flags & ADX_DCDWARN) == 0) {
	printf("adx%d DCD not asserted, is your cable connected?\n", unit);
	adx_softc[unit].adx_flags |= ADX_DCDWARN;
    }

    /*
     * Toggle DTR.
     */
    adx_dtr(unit, 0);
    timeout(adx_dtr_timeout, unit, mstohz(axc->adx_dtr_tmo));

    splx(s);

    DEBUGF(adx_debug & DEBUG_PROC,
	   printf("adx%d dtr end\n", unit);
    );
}


/*
 * adx_dtr_timeout - DTR timeout routine.
 *
 * Called sometime after a DCD event in order to raise DTR.
 */
void adx_dtr_timeout(unit)
    int unit;
{
    register struct adx_softc *axc = &adx_softc[unit];
    int s = splimp();

    DEBUGF(adx_debug & DEBUG_PROC,
	   printf("adx%d dtr begin\n", unit);
    );

    adx_dtr(unit, 1);
    if (axc->adx_dcd_tmo)
	timeout(adx_dcd_timeout, unit, mstohz(axc->adx_dcd_tmo));
    splx(s);

    DEBUGF(adx_debug & DEBUG_PROC,
	   printf("adx%d dtr end\n", unit);
    );
}


/*
 * adx_put - copy the mbuf chain into a contiguous buffer.
 *
 * Called from output routine to copy an mbuf chain into the output
 * buffer.
 */
int adx_put(unit, m, buf, stride)
    register unsigned int unit;
    register struct mbuf *m;	/* Source MBUF chain */
    caddr_t buf;		/* Desination buffer */
    int stride;			/* Buf stride */
{
    int mbufs;
    register int length = 0;
    caddr_t from;
    int i;

    DEBUGF(adx_debug & DEBUG_PROC,
	   printf("adx%d put begin\n", unit);
    );

    for (mbufs = 0; m; m = m->m_next, mbufs++) {
	if (mbufs > 22) {
	    printf("adx%d: Too many mbufs in chain\n", unit);
	    return (-1);
	}

	if (stride == 1) {
	    bcopy(mtod(m, caddr_t), buf, m->m_len);
	    buf += (int) m->m_len;
	}
	else {
	    for (from = mtod(m, caddr_t), i = m->m_len; i; i--, buf += stride)
		*buf = *from++;
	}
	length += (int) m->m_len;
    }

    DEBUGF(adx_debug & DEBUG_PROC,
	   printf("adx put end\n", unit);
    );

    return (length);
}


/*
 * adx_get - get frame into mbuf chain.
 *
 * Called from input routine to copy a devices input buffer into an mbuf
 * chain.  Deals correctly with data at a stride other than 1 (i.e. because
 * of 16-bit DMA channels).
 */
struct mbuf *adx_get(ifp, buf, len, stride)
    struct ifnet *ifp;		/* Interface pointer */
    register caddr_t buf;	/* Data */
    register int len;		/* Length of data */
    int stride;			/* Data stride */
{
    register struct mbuf *m, **mp;
    register unsigned count;
    struct mbuf *top = NULL;
    caddr_t to;
#ifdef DEBUG
    int unit = ifp->if_unit;
#endif

    DEBUGF(adx_debug & DEBUG_PROC,
	   printf("adx%d get begin\n", ifp->if_unit);
    );

#if 0
    for (count = 0; count < 24; count++)
	printf("%x ", buf[count]);
    printf("\n");
#endif

    mp = &top;
    while (len > 0) {
	MGET(m, M_DONTWAIT, MT_DATA);
	if ((*mp = m) == NULL) {
	    m_freem(top);
	    DEBUGF(adx_debug & DEBUG_SOFTERROR,
		   printf("adx_get: couldn't allocate mbuf\n");
	    );
	    return (NULL);
	}

	to = mtod(m, caddr_t);
	if (ifp == NULL) {
	    count = MIN(len, MLEN);
	    m->m_len = 0;
	}
	else {
	    bcopy(&ifp, to, sizeof (struct ifnet *));
	    m->m_len = sizeof (struct ifnet *);
	    to += sizeof (struct ifnet *);
	    count = MIN(len, MLEN - sizeof (struct ifnet *));
	    ifp = NULL;
	}

	if (stride == 1) {
	    bcopy(buf, to, count); /* Simply bcopy */
	    buf += count;
	}
	else {
	    int i;

	    for (i = count; i; i--, buf += stride)
		*to++ = *buf;
	}

	m->m_len += count;
	len -= count;
	mp = &m->m_next;
    }

#if 0
    {
	char *buf = mtod(top, char *);
	for (count = 0; count < 24; count++)
	    printf("%x ", buf[count]);
	printf("\n");
    }
#endif

    DEBUGF(adx_debug & DEBUG_PROC,
	   printf("adx%d get end\n", unit);
    );

    return (top);
}


/*
 * Mapping from 8237 channels to address offsets.
 */
int adx_channel_map[] = {
	2, 1, 0, 3,		/* First 8237 */
	0, 1, 2, 3,		/* Second 8237 */
	0			/* Coprocessor */
};

/*
 * Pointers to DMA controller registers.
 */
struct dma_8bit_device *dma8 = (struct dma_8bit_device *) DM_CTL1_BASE;
struct dma_16bit_device *dma16 = (struct dma_16bit_device *) DM_CTL2_BASE;


/*
 * This routine will set up the corresponding TCW (mapping registers) and
 * set up the corresponding 8237 channel.
 */
void adx_dma_setup(channel, buffer, length, dir)
    int channel;		/* Channel number to set up */
    caddr_t buffer;		/* Buffer to dma */
    short length;		/* Length of buffer (bytes/words) */
    int dir;			/* Direction to dma: DM_WRITE = write, DM_READ = read */
{
    short displacement;		/* 2k displacement */
    short prefix;		/* tcw prefix */
    u_char mode;		/* mode byte */
    int tcw_num;		/* tcw number */
    int ch;			/* 8237 channel */
    int i;

    DEBUGF(adx_debug & DEBUG_PROC,
	printf("adx dma_setup begin\n");
    );

    if (((int) buffer) & 0x10000000)
	panic("adx_dma_setup: address not in kernel space");

    /*
     * Find # of necessary TCWs and set them up.
     */
    displacement = (short) (((unsigned) buffer) & TCW_PAGEMASK);
    tcw_num = ((length << (DMA_CTLR_1(channel) ? 0 : 1)) +
	       displacement + TCW_PAGESIZE - 1) / TCW_PAGESIZE;
    if (tcw_num > 32)
	panic("adx_dma_setup: more than 32 TCWs => boundary error");

    DEBUGF(adx_debug & DEBUG_DMATCW,
	printf(" buffer=%x displacement=%x tcw_num=%x\n",
	       buffer, displacement, tcw_num);
    );

    /*
     * Calculate  TCWs.  Write prefixes in TCWs.
     */
    for (i = 0; i < tcw_num; i++, buffer += TCW_PAGESIZE) {
#if MACH
	prefix = (short) ((trans_virt_addr(buffer) >> TCW_PAGESHIFT) & 0x1fff);
#else
	prefix = (short) ((vtop(buffer) >> TCW_PAGESHIFT) & 0x1fff);
#endif
	TCW(channel, i) = prefix | TCW_RSC_ACC | TCW_REAL_ACC;

	DEBUGF(adx_debug & DEBUG_DMATCW,
	    printf(" tcw addr=%x tcw[%d]=%x read tcw=%x\n",
		   &TCW(channel, i), i, prefix | TCW_RSC_ACC | TCW_REAL_ACC,
		   TCW(channel, i));
	);
    }

    /*
     * Setup IOCC registers.
     */
    DMA_SET_PAGE_MODE(channel);	/* page mode only on 3 party */

    /*
     * Set up 8237 controller.
     * Set base address, low order byte first.
     * Set length count, low order byte first.
     */
    ch = adx_channel_map[channel];
    length -= 1;		/* Screwy hardware is off by 1 */
    mode = DMA_SINGLE | dir | ch;

    DEBUGF(adx_debug & DEBUG_DMA,
	printf(" channel=%d ch=%d length=0x%x mode=%x\n",
	       channel, ch, length, mode);
    );
 
    if (DMA_CTLR_1(channel)) {
	dma8->dm_mode = mode;	/* Issue mode byte */
	dma8->dm_ff = 0;	/* Clear first last flip/flop */
	dma8->dm_chan[ch].dm_base = (u_char) displacement;
	dma8->dm_chan[ch].dm_base = (u_char) (displacement >> 8);
	dma8->dm_ff = 0;
	dma8->dm_chan[ch].dm_count = (u_char) length;
	dma8->dm_chan[ch].dm_count = (u_char) (length >> 8);
    }
    else {
/* XXX Use Rotating Priority to avoid receive overruns when higher priority
   DMA source are hogging the bus (i.e the IBM Token Ring) */
	dma16->dm_cmd = (DM_CTLR_ENABLE | DM_PRIO_ROTATE | DM_WR_EXTEND);
	dma16->dm_mode = mode;	/* Issue mode byte */
	dma16->dm_ff = 0;	/* Clear first last flip/flop */
	dma16->dm_chan[ch].dm_base = (u_char) displacement >> 1;
	dma16->dm_chan[ch].dm_base = (u_char) (displacement >> 9);
	dma16->dm_ff = 0;
	dma16->dm_chan[ch].dm_count = (u_char) length;
	dma16->dm_chan[ch].dm_count = (u_char) (length >> 8);
	length <<=  1;		/* Adjust for use below */
    }

    /*
     * Enable specific channel.
     */
    if (DMA_CTLR_1(channel))
	dma8->dm_smask = ch;
    else
	dma16->dm_smask = ch;

    DEBUGF(adx_debug & DEBUG_PROC,
	printf("adx dma_setup end\n");
    );

    return;
}


/*
 * This routine will be called from the requesting device interrupt
 * handler.  It returns 0 if the transfer was completed, else the number
 * of bytes/words left to transfer.  It also sets the mask inhibiting further
 * transfers.
 */
int adx_dma_done(channel)
    int channel;
{
    u_char ch;
    int length;
#ifdef DEBUG
    int status = 0;
#endif

    DEBUGF(adx_debug & DEBUG_PROC,
	printf("adx dma_done begin\n");
    );

    /*
     * Map virtual channel number to real number.
     */
    ch = adx_channel_map[channel];

    /*
     * Retrieve length of transfer and disable channel.
     * The length register contains the number of bytes/words left to transfer
     * minus 1, so we must adjust for that.
     * The length register contains 0xffff if the transfer finished.
     * adding 1 to this yields 0x10000 when we really want 0.
     */
    length = 1;
    if (DMA_CTLR_1(channel)) {
	status = dma8->dm_cmd;
	length += dma8->dm_chan[ch].dm_count;
	length += dma8->dm_chan[ch].dm_count << 8;
	dma8->dm_smask = 0x4 | ch;
	if (length == 0x10000)
	    length = 0;
    }
    else {
	status = dma16->dm_cmd;
	length += dma16->dm_chan[ch].dm_count;
	length += dma16->dm_chan[ch].dm_count << 8;
	dma16->dm_smask = 0x4 | ch;
	if (length == 0x10000)
	    length = 0;
    }

    DEBUGF(adx_debug & DEBUG_DMA,
	printf(" channel=%d ch=%d status=%x length=0x%x\n",
	       channel, ch, status, length);
    );

    DEBUGF(adx_debug & DEBUG_PROC,
	printf("adx dma_done end\n");
    );

    return (length);
}
#endif
