/*	if_lapb.c	*/

/*
 * X.25 LAPB interface driver (upper half).
 *
 * The LAPB protocol is handled in this module, leaving
 * hardware specifics to a lower level interface driver.
 * 
 * The two modules are coupled at "attach time", and ever after,
 * through the lapb_driver and lapb_device structures.  Higher level
 * protocols, e.g. IP, interact with this module.
 *
 *	Drew D. Perkins
 *	Carnegie Mellon University
 *
 * Routines defined in this module are referenced via a few structures:
 * Called via the ifnet structure by the interface layer:
 *	lapb_init()	- initialize
 *	lapb_ioctl()	- process and ioctl request
 *	lapb_reset()	- device was reset
 *	lapb_output()	- output a packet
 *
 * Called by lower half hardware drivers:
 *	lapb_attach()	- attach device to LAPB system
 *	lapb_start()	- start output of next frame
 *	lapb_input()	- input a frame
 *	lapb_modem()	- device readiness changed state
 */


#include "lapb.h"
#if NLAPB > 0

#include "../machine/pte.h"

#include "param.h"
#include "systm.h"
#include "mbuf.h"
#include "buf.h"
#include "protosw.h"
#include "socket.h"
#include "vmmac.h"
#include "time.h"
#include "kernel.h"
#include "ioctl.h"
#include "errno.h"

#include "../net/if.h"
#include "../net/netisr.h"
#include "../net/route.h"

#ifdef INET
#include "../netinet/in.h"
#include "../netinet/in_systm.h"
#include "../netinet/in_var.h"
#include "../netinet/ip.h"
#include "../netinet/ip_var.h"
#endif

#include "../net/if_lapbvar.h"
#include "../net/if_lapb.h"

#define DEBUG_PROC	0x1	/* Procedure entrance/exit */
#define DEBUG_MBUF	0x2	/* Mbuf handling */
#define DEBUG_CBUF	0x4	/* xxx */
#define DEBUG_HDR	0x8	/* Header processing */
#define DEBUG_ERROR	0x10	/* Errors */

/*
#define DEBUG	DEBUG_PROC|DEBUG_MBUF|DEBUG_CBUF|DEBUG_HDR|DEBUG_ERROR
#define DEBUG	DEBUG_ERROR
*/
#define DEBUG	0

#ifdef DEBUG
char lapb_debug = DEBUG;
#define DEBUGF(cond,stmt) if (cond) stmt
#else
#define DEBUGF(cond,stmt)
#endif

struct ifnet *lapb_attach();
void lapb_init(), lapb_start(), lapb_input();
void lapb_modem();
int lapb_ioctl(), lapb_reset(), lapb_output();
#ifdef DEBUG
void lapb_printh();
#endif

int lapb_units = 0;		/* Count of actual devices attached */

struct lapb_softc lapb_softc[NLAPB];


/*
 * lapb_attach - attach a hardware device to the system.
 *
 * Called from hardware device attach routine at configuration time
 * with a pointer to a lapb driver structure and other info.
 * Sets up local state and attaches the interface to the system.
 */
struct ifnet *lapb_attach(driver, name, unit, mtu)
    struct lapb_driver *driver;
    char *name;
    short unit;
    short mtu;
{
    struct lapb_softc *sc;
    struct ifnet *ifp;

    DEBUGF(lapb_debug & DEBUG_PROC,
	   printf("lap %s%d attach begin\n", name, unit);
    );

    if (lapb_units >= NLAPB) {
	printf("lapb: %s%d not attached!  Increase NLAPB\n", name, unit);
	return (NULL);
    }

    /*
     * Initialize LAPB layer variables.
     */
    sc = &lapb_softc[lapb_units];
    sc->lapb_driver = driver;
    sc->lapb_unit = unit;

    /*
     * Initialize interface layer variables.
     * Adjust the hardware MTU to account for LAPB header.
     */
    ifp = &sc->lapb_if;
    ifp->if_name = name;
    ifp->if_unit = lapb_units++;
    ifp->if_mtu = mtu - sizeof (struct lapb_header);
    ifp->if_flags = IFF_POINTOPOINT;
    ifp->if_init = (int (*)()) lapb_init;
    ifp->if_output = lapb_output;
    ifp->if_ioctl = lapb_ioctl;
    ifp->if_reset = lapb_reset;

    /*
     * Attach to interface layer.
     */
    if_attach(ifp);

    DEBUGF(lapb_debug & DEBUG_PROC,
	   printf("lap %s%d attach end\n", name, unit);
    );

    return (ifp);
}


/*
 * lapb_reset - reset LAPB interface after bus reset.
 *
 * Called by if layer after bus reset.  First notify device then try
 * to re-initialize.
 */
int lapb_reset(unit)
    register short unit;
{
    struct lapb_softc *sc = &lapb_softc[unit];

    sc->lapb_driver->dr_reset(sc->lapb_unit);
    lapb_init(&sc->lapb_if);
}


/*
 * lapb_init - initialize a LAPB interface.
 *
 * Called by LAPB ioctl routine.  Calls hardware module to setup
 * hardware resources and init state.
 */
void lapb_init(ifp)
    register struct ifnet *ifp;
{
    int s = splimp();
    struct lapb_softc *sc = &lapb_softc[ifp->if_unit];

    DEBUGF(lapb_debug & DEBUG_PROC,
	   printf("lap %s%d init begin\n", ifp->if_name, ifp->if_unit);
    );

    /*
     * Don't initialize unless an address has been set.
     */
    if (ifp->if_addrlist == 0)
	return;

    /*
     * Assume LAPB initial state.
     */
    sc->lapb_state = LAPB_INITIAL;
    sc->lapb_oactive = 0;

    /*
     * We're running!
     */
    ifp->if_flags &= ~IFF_UP;
    ifp->if_snd.ifq_maxlen = 20;

    /*
     * Initialize hardware.  We will be notified when the hardware and
     * physical layer are ready.
     */
    if ((*sc->lapb_driver->dr_init)(sc->lapb_unit) == -1) {
	ifp->if_flags &= ~(IFF_UP|IFF_RUNNING);
	splx(s);
	return;
    }

    DEBUGF(lapb_debug & DEBUG_PROC,
	   printf("lap %s%d init end\n", ifp->if_name, ifp->if_unit);
    );

    splx(s);
}


/*
 * lapb_down - shutdown a LAPB interface.
 *
 * Called by LAPB ioctl routine.  Calls hardware module to shutdown
 * hardware.  Possibly attempts closing connection with remote end.
 */
void lapb_down(ifp)
    register struct ifnet *ifp;
{
    int s = splimp();
    struct lapb_softc *sc = &lapb_softc[ifp->if_unit];

    DEBUGF(lapb_debug & DEBUG_PROC,
	   printf("lap %s%d down begin\n", ifp->if_name, ifp->if_unit);
    );

    /*
     * Shutdown hardware and drop DTR.
     */
    (*sc->lapb_driver->dr_down)(sc->lapb_unit);

    /*
     * Assume LAPB initial state. XXX
     */
    sc->lapb_state = LAPB_INITIAL;
    ifp->if_flags &= ~IFF_UP;
    sc->lapb_oactive = 0;

    DEBUGF(lapb_debug & DEBUG_PROC,
	   printf("lap %s%d down end\n", ifp->if_name, ifp->if_unit);
    );

    splx(s);
}


/*
 * lapb_ioctl - Process an ioctl request.
 *
 * Called as a result of a user process executing an ioctl() call.
 */
int lapb_ioctl(ifp, cmd, data)
    register struct ifnet *ifp;
    int cmd;
    caddr_t data;
{
    int s = splimp(), error = 0;

    DEBUGF(lapb_debug & DEBUG_PROC,
	   printf("lap %s%d ioctl begin\n", ifp->if_name, ifp->if_unit);
    );

    switch (cmd) {
	case SIOCSIFADDR:
	    if ((ifp->if_flags & IFF_UP) &&
		((ifp->if_flags & IFF_RUNNING) == 0))
		lapb_init(ifp);
	    break;

	case SIOCSIFDSTADDR:
	    if ((ifp->if_flags & IFF_UP) &&
		((ifp->if_flags & IFF_RUNNING) == 0))
		lapb_init(ifp);
	    break;

	case SIOCSIFFLAGS:
	    /*
	     * If this ioctl is shutting us down, then IFF_UP will
	     * have been reset by the time we get here.  IFF_RUNNING
	     * won't however.  If this ioctl is turning us on,
	     * the flags won't have been changed.
	     */
	    if (((ifp->if_flags & IFF_UP) == 0) &&
		(ifp->if_flags & IFF_RUNNING))
		lapb_down(ifp); 
	    else if ((ifp->if_flags & IFF_UP) &&
		     ((ifp->if_flags & IFF_RUNNING) == 0))
		lapb_init(ifp); 

	default:
	    error = EINVAL;
    }

    splx(s);

    DEBUGF(lapb_debug & DEBUG_PROC,
	   printf("lap %s%d ioctl end\n", ifp->if_name, ifp->if_unit);
    );

    return (error);
}


/*
 * lapb_output - output a packet.
 *
 * Called from higher level protocol routines to set up messages for
 * transmission.  Sets up the header and calls lapbsnd to
 * enqueue the message for this LAPB's hardware driver.
 */
int lapb_output(ifp, m0, dst)
    register struct ifnet *ifp;
    struct mbuf *m0;
    struct sockaddr *dst;
{
    int type, error, s;
    register struct mbuf *m = m0;
    register struct lapb_header *lapb;
    struct lapb_softc *sc = &lapb_softc[ifp->if_unit];
#ifdef DEBUG
    int mbufs, totlen;
#endif

    DEBUGF(lapb_debug & DEBUG_PROC,
	   printf("lap %s%d output begin\n", ifp->if_name, ifp->if_unit);
    );

    /*
     * Don't even try if the LAPB is unavailable.
     */
#ifdef notdef
    if (lapb_softc[ifp->if_unit].lapb_state != LAPBS_CONNECTED) {
    }
#endif
    if ((ifp->if_flags & IFF_UP) == 0) {
	error = ENETDOWN;
	goto bad;
    }

    switch (dst->sa_family) {
#ifdef INET
	case AF_INET:
	    type = LAPB_IPTYPE;
	    goto gottype;
#endif

#ifdef NS
	case AF_NS:
	    type = LAPB_NSTYPE;
	    goto gottype;
#endif NS

	case AF_UNSPEC:
	    type = LAPB_PPPTYPE;
	    goto gottype;

	default:
	    printf("lap %s%d: can't handle af%d\n", ifp->if_unit,
		   dst->sa_family);
	    error = EAFNOSUPPORT;
	    goto bad;
    }

gottype:
    /*
     * Add LAPB leader.  If there's not enough space in the
     * first mbuf, allocate another.  If that should fail, we
     * drop this packet.
     */
    if (m->m_off > MMAXOFF ||
	MMINOFF + sizeof(struct lapb_header) > m->m_off) {
	m = m_get(M_DONTWAIT, MT_HEADER);
	if (m == 0) {
	    ifp->if_oerrors++;
	    error = ENOBUFS;
	    goto bad;
	}
	m->m_next = m0;
	m->m_off = MMINOFF;
	m->m_len = sizeof(struct lapb_header);
    } else {
	m->m_off -= sizeof(struct lapb_header);
	m->m_len += sizeof(struct lapb_header);
    }

    lapb = mtod(m, struct lapb_header *);
    lapb->lapb_type = htons((u_short) type);

    /*
     * Queue packet on interface and start output if interface
     * not yet active.  If queue is full, increment "collision count".
     */
    s = splimp();
    if (IF_QFULL(&ifp->if_snd)) {
	IF_DROP(&ifp->if_snd);
	splx(s);
	m0 = m;
	ifp->if_collisions++;
	error = ENOBUFS;
	goto bad;
    }

    IF_ENQUEUE(&ifp->if_snd, m);
#ifdef DEBUG
    for (totlen = 0, mbufs = 0; m; m = m->m_next, mbufs++)
	totlen += m->m_len;
#endif
    DEBUGF(lapb_debug & DEBUG_MBUF,
	   printf("lap %s%d output mbufs=%d totlen=%d\n",
		  ifp->if_name, ifp->if_unit, mbufs, totlen);
    );

    /*
     * Start a new packet if none in progress.
     */
    if (!sc->lapb_oactive)
	lapb_start(ifp);

    splx(s);
    error = 0;
    goto good;

bad:
    m_freem(m0);

good:
    DEBUGF(lapb_debug & DEBUG_PROC,
	   printf("lap %s%d output end\n", ifp->if_name, ifp->if_unit);
    );

    return (error);
}


/*
 * lapb_start - start output on a LAPB interface.
 *
 * Called from either lapb_output or hardware output interrupt routine to
 * start transmission of new frame.  First checks if any control frames need
 * to be sent.  If not, gets another datagram to send from the interface
 * queue.  This routine expects to be called at spl splimp or higher.
 */
void lapb_start(ifp)
    register struct ifnet *ifp;
{
    register struct lapb_softc *sc = &lapb_softc[ifp->if_unit];
    struct mbuf *m;
    struct lapb_header *lapb;
    int len;

    IF_DEQUEUE(&ifp->if_snd, m);
    if (m == 0) {
	sc->lapb_oactive = 0;
	return;
    }

    DEBUGF(lapb_debug & DEBUG_PROC,
	   printf("lap %s%d start begin\n", ifp->if_name, ifp->if_unit);
    );

    lapb = mtod(m, struct lapb_header *);
    lapb->lapb_addr = sc->lapb_rem_addr;
    lapb->lapb_cntl = LAPB_UI;

    DEBUGF(lapb_debug & DEBUG_HDR,
	   lapb_printh(lapb);
    );

    ifp->if_opackets++;
    sc->lapb_oactive = 1;
    len = (*sc->lapb_driver->dr_output)(sc->lapb_unit, m);
    if (len < sizeof (struct lapb_header))
	printf("lap %s%d bad tx len 0x%x\n", ifp->if_name, ifp->if_unit, len);

    DEBUGF(lapb_debug & DEBUG_CBUF,
	   printf("lap %s%d start len=%d\n", ifp->if_name, ifp->if_unit, len);
    );

    DEBUGF(lapb_debug & DEBUG_PROC,
	   printf("lap %s%d start end\n", ifp->if_name, ifp->if_unit);
    );
}


/*
 * lapb_input - input a frame.
 *
 * Called from hardware input interrupt routine to handle LAPB frames.
 */
void lapb_input(ifp, m)
    struct ifnet *ifp;
    register struct mbuf *m;
{
    register struct lapb_softc *sc = &lapb_softc[ifp->if_unit];
    register struct lapb_header *lapb;
    register struct ifqueue *inq;
    char *cp;
    int type;
#ifdef DEBUG
    int totlen, mbufs;
#endif

    DEBUGF(lapb_debug & DEBUG_PROC,
	   printf("lap %s%d input begin\n", ifp->if_name, ifp->if_unit);
    );

    DEBUGF(lapb_debug & DEBUG_CBUF,
	   printf("lap %s%d input\n", ifp->if_name, ifp->if_unit);
    );

    /*
     * Make sure header mbuf is long enough.
     */
    ifp->if_ipackets++;
    if (((m->m_off > MMAXOFF) ||
	 (m->m_len < sizeof (struct lapb_header) + sizeof (struct ifnet *))) &&
	((m = m_pullup(m, sizeof (struct lapb_header) +
		          sizeof (struct ifnet *))) == 0)) {
	ifp->if_ierrors++;
	DEBUGF(lapb_debug & DEBUG_ERROR,
	       printf("lap %s%d bad rx len 0x%x\n",
		      ifp->if_name, ifp->if_unit);
	);
	goto drop;
    }

    /*
     * Figure out packet type, remove header and stick ifp in front of packet.
     */
    cp = mtod(m, char *);
    lapb = (struct lapb_header *) (cp + sizeof (struct ifnet *));
    DEBUGF(lapb_debug & DEBUG_HDR,
	   lapb_printh(lapb);
    );
    type = htons(lapb->lapb_type);

    /*
     * Remove lapb_header.  Copy struct ifnet * over lapb_header.
     */
    bcopy(&ifp, cp + sizeof(struct lapb_header), sizeof(struct ifnet *));
    m->m_off += sizeof(struct lapb_header);
    m->m_len -= sizeof(struct lapb_header);

    switch (type) {
#ifdef INET
    case LAPB_IPTYPE:
	schednetisr(NETISR_IP);
	inq = &ipintrq;
	break;
#endif

    default:
	ifp->if_ierrors++;
	DEBUGF(lapb_debug & DEBUG_ERROR,
	       printf("lap %s%d bad rx type 0x%x\n",
		      ifp->if_name, ifp->if_unit, type);
	);
	goto drop;
    }

    if (IF_QFULL(inq)) {
	IF_DROP(inq);
	m_free(m);
	ifp->if_ierrors++;
	DEBUGF(lapb_debug & DEBUG_ERROR,
	       printf("lap %s%d if queue full, dropping\n",
		      ifp->if_name, ifp->if_unit);
	);
    } else {
	IF_ENQUEUE(inq, m);
    }

#ifdef DEBUG
    for (totlen = 0, mbufs = 0; m; m = m->m_next, mbufs++)
	totlen += m->m_len;

    DEBUGF(lapb_debug & DEBUG_CBUF,
	   printf("lap %s%d input mbufs=%d totlen=%d\n",
		  ifp->if_name, ifp->if_unit, mbufs, totlen);
    );
#endif

drop:
    DEBUGF(lapb_debug & DEBUG_PROC,
	   printf("lap %s%d input end\n", ifp->if_name, ifp->if_unit);
    );
}


/*
 * lapb_modem - modem control lead transition.
 *
 * Called from hardware input interrupt routine to handle modem control lead
 * transitions.  Going down resets the connection.
 */
void lapb_modem(ifp, state)
    struct ifnet *ifp;
    int state;
{
    register struct lapb_softc *sc = &lapb_softc[ifp->if_unit];

    DEBUGF(lapb_debug & DEBUG_PROC,
	   printf("lap %s%d modem begin\n", ifp->if_name, ifp->if_unit);
    );

    printf("lap %s%d physical layer %s\n", ifp->if_name, ifp->if_unit,
	   state ? "up" : "down");

    if (!state) {
	/*
	 * Mark interface down.
	 */
	ifp->if_flags &= ~IFF_UP;

	/*
	 * Reset the connection.  I.e. Stop all timers and enter
	 * INITIAL state.
	 */
	sc->lapb_state = LAPB_INITIAL;
    } else {
	/*
	 * XXX Mark interface up.
	 */
	ifp->if_flags |= IFF_UP;

	/*
	 * Reset the connection.  I.e. Stop all timers and enter
	 * INITIAL state.
	 */
	sc->lapb_state = LAPB_INITIAL;
    }

    DEBUGF(lapb_debug & DEBUG_PROC,
	   printf("lap %s%d modem end\n", ifp->if_name, ifp->if_unit);
    );
}


#ifdef DEBUG
/*
 * LAPB header pretty print routine:
 * Pretty print a LAPB header.
 */
void lapb_printh(lapb)
	struct lapb_header *lapb;
{
	printf("lapb_header(%x) Address: %d\tCntl: %x\tType: %d\n",
		lapb, lapb->lapb_addr, lapb->lapb_cntl, lapb->lapb_type);
}
#endif
#endif
