/*	if_lapb.h	*/

/*
 * LAPB packet encapsulation.  The type allows the link to be multiplexed
 * among several protcols.
 */
struct lapb_header {
    u_char lapb_addr;		/* Destination address */
    u_char lapb_cntl;		/* Control field */
    u_short lapb_type;		/* Packet type */
};


/*
 * Packet types.
 */
#define LAPB_IPTYPE	1
#define LAPB_PPPTYPE	2
#define LAPB_XNSTYPE	3
#define LAPB_ISOTYPE	4

/*
 * There are only two valid addresses with LAPB.  The DTE uses ADDRESS A = 3.
 * The DCE uses ADDRESS B = 1.  Commands are sent using the remote address.
 * Responses are sent using the local address.
 */
#define LAPB_ADDRESS_A	0x03
#define LAPB_ADDRESS_B	0x01

/*
 * HDLC Link Access Procedure-Balanced control field encoding.
 */
#define LAPB_IS_MASK	0x01	/* Mask to decide Information or Supervisory/Unnumbered */
#define LAPB_INFO	0x00	/* Information frame */
#define LAPB_SU_MASK	0x03	/* Mask to decide Supervisory or Unnumbered */
#define LAPB_SUPER	0x01	/* Supervisory frame */
#define LAPB_UNNUM	0x03	/* Unnumbered frame */
#define LAPB_S_MASK	0x0f	/* Mask to extract supervisory command */
#define LAPB_U_MASK	0xef	/* Mask to extract unnumbered command */
#define LAPB_P_F	0x10	/* Poll/Final Bit */
#define LAPB_NS_MASK	0x0e	/* The N(S) sequence number is here */
#define LAPB_NR_MASK	0xe0	/* The N(R) sequence number is here */
#define LAPB_NS_SHIFT	1	/* Shift field this number of bits */
#define LAPB_NR_SHIFT	5	/* Shift field this number of bits */

/*
 * Supervisory frame command and response encoding.
 */
#define LAPB_RR		(0x00 | LAPB_SUPER)
#define LAPB_RNR	(0x04 | LAPB_SUPER)
#define LAPB_REJ	(0x08 | LAPB_SUPER)
#define LAPB_SREJ	(0x0c | LAPB_SUPER) /* Not LAPB */

/*
 * Unnumbered frame command and response encoding.
 */
#define LAPB_UI		(0x00 | LAPB_UNNUM) /* Not LAPB */
#define LAPB_DISC	(0x40 | LAPB_UNNUM)
#define LAPB_UA		(0x60 | LAPB_UNNUM)
#define LAPB_FRMR	(0x84 | LAPB_UNNUM)
#define LAPB_DM		(0x0c | LAPB_UNNUM)
#define LAPB_SABM	(0x2c | LAPB_UNNUM)


/*
 * LAPB states.
 */
/* connection state */
#define LAPB_INITIAL	0	/* Initial unconnected state */
#define LAPB_CON_ACTIVE	1	/* Actively attempting connection */
#define LAPB_CON_PASSIVE 2	/* Passively attempting connection */
#define LAPB_CONNECTED	3	/* Connection established */
#define LAPB_PREL_DISC	4	/* Preliminary disconnecting */
#define LAPB_FRMR_STATE	5	/* Frame reject */
#define LAPB_DISC_ACTIVE 6	/* Actively attempting disconnection */
#define LAPB_DISC_PASSIVE 7	/* Passively attempting disconnection */
#define LAPB_PREL_CONN 8	/* Preliminary connecting */

/* transmitter state */
#define LAPB_TX_NORMAL	0	/* Transmitter normal */
#define LAPB_TX_BUSY	1	/* Remote end busy */

/* receiver state */
#define LAPB_RX_NORMAL	0	/* Receiver normal */
#define LAPB_RX_BUSY	1	/* Receiver busy */
#define LAPB_RX_REJ	2	/* Receiver rejecting */

/*
 * Which side, DTE or DCE?
 */
#define LAPB_DTE	0	/* DTE, start in CON_ACTIVE */
#define LAPB_DCE	1	/* DCE, start in CON_PASSIVE */

/*
 * Convert number of milliseconds to hz.  Hz is the number of clock ticks
 * desired before a timeout is called.
 * Shifting by 10 is the same as dividing by 1024 which is "close" to
 * dividing by 1000.
 */
#define mstohz(ms)	(((ms) * hz) >> 10)

/*
 * Increment a sequence number variable.
 */
#define LAPB_SEQ_MASK	0x07	/* Sequence number mask */
#define LAPB_INC_SEQ(n)	((n) = (n + 1) & LAPB_SEQ_MASK)
#define LAPB_NUMSEQNUM	0x08	/* Number of sequence numbers available */

#define LAPB_WAIT_F_TMO		100	/* Timeout waiting for Final bit */
#define LAPB_WAIT_ACK_TMO	100	/* Timeout waiting for ACK */
#define LAPB_FRMR_TMO		100	/* Timeout waiting for FRMR */
#define LAPB_LOCAL_BUSY_TMO	100	/* Timeout waiting for input queue to become not-full */


/*
 * Data structure used by the LAPB protocol module.  Contains all
 * state of the LAPB connection.
 */
struct lapb_softc {
    struct ifnet lapb_if;	/* PPP visible interface */
    struct lapb_driver *lapb_driver; /* Pointer to driver descriptor */
    int lapb_unit;		/* Device unit number */
    char lapb_oactive;		/* Output in progress */
    int lapb_which_side;	/* DTE or DCE? */
    int lapb_state;		/* Current LAPB connection state */
    int lapb_VS;		/* Next send sequence number to be sent (V(S)) */
    int lapb_LR;		/* Last recieved N(R) (last send sequence number acknowledged) */
    int lapb_tx_state;		/* State of transmitter (Normal, Busy) */
    int lapb_tx_retries;	/* Number of times frame retransmitted */
    int lapb_window_closed;	/* Window open/closed */
    int lapb_VR;		/* Next receive sequence number expected (V(R))*/
    int lapb_rx_state;		/* State of receiver (Normal, Busy, Rej) */
    int lapb_tx_ack;		/* Must send an ack */
    int lapb_tx_f;		/* Must respond with F bit set */
    int lapb_frames_discarded;	/* Discarded frames while busy */
    u_char lapb_s_cmd;		/* Supervisor command requested (RR, RNR, REJ) */
    u_char lapb_s_rsp;		/* Supervisor response requested (RR, RNR, REJ) */
    u_char lapb_u_cmd;		/* Unnumbered command requested (SABM, DISC) */
    u_char lapb_u_rsp;		/* Unnumbered response requested (UA, DM, FRMR) */
    u_char lapb_our_addr;	/* Our HDLC address */
    u_char lapb_rem_addr;	/* Remote HDLC address */
    struct mbuf *lapb_ack_q[LAPB_NUMSEQNUM];
				/* Queue of tx packets sent and waiting
				    for acknowledgement. */
    int lapb_ack_q_len;		/* Number of entries in the ack queue */
}; 
