/*-
 * Copyright (c) 1990 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <stdio.h>
#include <errno.h>
#include <a.out.h>
#include <unistd.h>
#include <stdlib.h>
#include <ar.h>
#include <string.h>
#include "shlib.h"

#define MAX_FILENAME 1024
#define PAGE_ALIGN(x) ((((x) + PAGESIZE - 1) / PAGESIZE) * PAGESIZE)

static FILE *in;
static struct exec in_header;

static const char *program_name;

/* command line arguments */

static const char *input_module,
                  *shared_library,
                  *linker_library,
                  *module_directory;

static int        version;

static caddr_t    text_address,
                  data_address;

/*****************************************************************************/
/**************************** Error Messages *********************************/
/*****************************************************************************/

static void out_of_memory(void) {
   fprintf(stderr, "%s: ", program_name);
   perror("Out of memory");
   exit (errno);
}

static void open_error(const char *name) {
   char message[MAX_FILENAME];

   fprintf(stderr, "%s: ", program_name);
   sprintf(message, "Can't open file \"%s\"", name);
   perror(message);
   exit(errno);
}

static void create_error(const char *name) {
   char message[MAX_FILENAME];

   sprintf(message, "Can't create file \"%s\"", name);
   fprintf(stderr, "%s: ", program_name);
   perror(message);
   exit(errno);
}

static void read_error(void) {
   fprintf(stderr, "%s: ", program_name);
   perror("Read error");
   exit (errno);
}

static void write_error(void) {
   fprintf(stderr, "%s: ", program_name);
   perror("Write error");
   exit (errno);
}

static void usage(void) {
   fprintf(stderr, "usage: %s input-module shared-library link-library "
	           "module-directory\\\n\t\tversion text-address "
	           "data-address\n"
	   "input-module, shared-library, link-library: file names\n"
	   "module-directory: directory names\n"
	   "version, text-address, data-address: numbers (use 0x for hex)\n",
	   program_name);
   exit (1);
}


/*****************************************************************************/
/**************************** Initialization *********************************/
/*****************************************************************************/

static void initialize(void) {

   if (getpagesize() != PAGESIZE) {
      fprintf(stderr, "%s: Internal error: please adjust constant PAGESIZE. "
	      "Aborted.\n", program_name);
      exit(1);
   }

   printf("Loading absolutely linked `%s' ...\n", input_module);

   if (!(in = fopen(input_module, "r"))) open_error(input_module);

   if (fread(&in_header, sizeof in_header, 1, in) != 1) read_error();

   if (N_BADMAG(in_header)) {
      fprintf(stderr, "%s: Not an executable file. Aborted.\n", program_name);
      exit(1);
   }

   if (in_header.a_trsize || in_header.a_drsize) {
      fprintf(stderr, "%s: File contains relocation information. Aborted.\n",
	      program_name);
      exit(1);
   }

   if ((caddr_t) in_header.a_entry != text_address) {
      fprintf(stderr, "%s: Text address mismatch. 0x%8.8x vs 0x%8.8x.\n",
	      program_name, in_header.a_text, text_address);
      exit(1);
   }
}

/*****************************************************************************/
/**************************** Shared Library *********************************/
/*****************************************************************************/

void create_shared_library(void) {
   ShareLibHeader shlib;
   FILE *out;
   int  i;

   /* create shared library */

   printf("Creating shared library `%s' ...\n", shared_library);

   if (!(out = fopen(shared_library, "w"))) create_error(shared_library);
   
   /* write header */

   bzero(&shlib, sizeof shlib);
   shlib.header.magic = SHLIB_MAGIC;
   shlib.header.version = version;
   shlib.header.text_size = PAGE_ALIGN(in_header.a_text);
   shlib.header.data_size = PAGE_ALIGN(in_header.a_data + in_header.a_bss);
   shlib.header.text_addr = text_address;
   shlib.header.data_addr = data_address;
   if (fwrite(&shlib, sizeof shlib, 1, out) != 1) write_error();

   printf("%s: text %8.8x-%8.8x, data %8.8x-%8.8x\n", shared_library,
	  text_address, text_address + shlib.header.text_size - 1,
	  data_address, data_address + shlib.header.data_size - 1);

   /* write text */

   fseek(in, N_TXTOFF(in_header), SEEK_SET);
   for (i = in_header.a_text; i--;) fputc(fgetc(in), out);
   for (i = shlib.header.text_size - in_header.a_text; i--;) putc(0, out);

   /* write data */

   fseek(in, N_DATOFF(in_header), SEEK_SET);
   for (i = in_header.a_data; i--;) fputc(fgetc(in), out);
   for (i = shlib.header.data_size - in_header.a_data; i--;) putc(0, out);

   fclose(out);
}

/*****************************************************************************/
/************************* Copy Member to Library ****************************/
/*****************************************************************************/

static int write_object_file(const char *full_name,
			     int absolute_symbols,
			     struct nlist *absolute_symtab,
			     FILE *library) {
   int offset = ftell(library);
   FILE *object;
   struct exec header;
   int symbols, string_size, new_string_size, i;
   struct nlist *symtab;
   char *strings, *new_strings;

   if (!(object = fopen(full_name, "r"))) open_error(full_name);

   if (fread(&header, sizeof header, 1, object) != 1) read_error();

   if (N_BADMAG(header)) {
      fprintf(stderr, "%s: Not an executable file. Aborted.\n", program_name);
      exit(1);
   }

   /* load symbol table */

   if (!(symtab = malloc(header.a_syms))) out_of_memory();
   symbols = header.a_syms / sizeof (struct nlist);
   if (fseek(object, N_SYMOFF(header), SEEK_SET)) read_error();
   if (fread(symtab, sizeof (struct nlist), symbols, object) != symbols)
      read_error();

   /* load string table */

   if (fseek(object, 0, SEEK_END)) read_error();
   i = ftell(object);
   if (fseek(object, N_STROFF(header), SEEK_SET)) read_error();
   string_size = i - ftell(object);
   strings = malloc(string_size);
   new_strings = malloc(string_size);
   if (!strings || !new_strings) out_of_memory();
   if (fread(strings, 1, string_size, object) != string_size) read_error();
   fclose(object);

   /* make references absolute, add library object */

   header.a_text = header.a_data = header.a_bss = header.a_syms = 0;
   header.a_entry = header.a_trsize = header.a_drsize = 0;
   if (fwrite(&header, sizeof header, 1, library) != 1) write_error();

   new_string_size = sizeof (long);
   if (fseek(library, offset + N_SYMOFF(header), SEEK_SET)) read_error();
   for (i = 0; i < symbols; ++i) {
      struct nlist *p = symtab + i;
      if (p->n_type & N_EXT) {
	 int type = p->n_type & N_TYPE;

	 if (type == N_TEXT || type == N_DATA || type == N_BSS ||
	     type == N_COMM || type == N_ABS) {
	    char *name = strings + p->n_un.n_strx;
	    char *s = new_strings + new_string_size;
	    int  i;

	    /* lookup symbol in absolute table */
	    for (i = 0; i < absolute_symbols; ++i) {
	       struct nlist *q = absolute_symtab + i;

	       if (!strcmp(q->n_un.n_name, name)) {
		  p->n_value = q->n_value;
		  p->n_type = N_ABS | N_EXT;
		  break;
	       }
	    }

	    /* copy string in new string table */
	    p->n_un.n_strx = new_string_size;
	    new_string_size += strlen(name) + 1;
	    while (*s++ = *name++);

	    /* write symbol */
	    if (fwrite(p, sizeof (struct nlist), 1, library) != 1)
	       write_error();

	    header.a_syms += sizeof (struct nlist);
	 }
      }
   }

   /* write new string table */
   *((long *) new_strings) = new_string_size;
   if (fseek(library, offset + N_STROFF(header), SEEK_SET)) read_error();
   if (fwrite(new_strings, 1, new_string_size, library) != new_string_size)
      write_error();

   /* update header */
   if (fseek(library, offset, SEEK_SET)) read_error();
   if (fwrite(&header, sizeof header, 1, library) != 1) write_error();

   free(symtab);
   free(strings);
   free(new_strings);

   return N_STROFF(header) + new_string_size;
}

static void copy_string(char *target, const char *source, int size) {
   while (size--) *target++ = *source ? *source++ : ' ';
}

static void copy_number(char *target, int src, int size) {
   char source[15];

   sprintf(source, "%d", src);
   copy_string(target, source, size);
}

/*****************************************************************************/
/**************** Library with References to Shared Library ******************/
/*****************************************************************************/

static void write_linker_library(void) {
   int          symbols, string_size, i;
   struct nlist *symtab;
   char         *strings;
   FILE         *library;

   /* load symbol table */

   if (!(symtab = malloc(in_header.a_syms))) out_of_memory();
   symbols = in_header.a_syms / sizeof (struct nlist);
   if (fseek(in, N_SYMOFF(in_header), SEEK_SET)) read_error();
   if (fread(symtab, sizeof (struct nlist), symbols, in) != symbols)
      read_error();

   /* load string table */

   if (fseek(in, 0, SEEK_END)) read_error();
   i = ftell(in);
   if (fseek(in, N_STROFF(in_header), SEEK_SET)) read_error();
   string_size = i - ftell(in);
   strings = malloc(string_size);
   if (!strings) out_of_memory();
   if (fread(strings, 1, string_size, in) != string_size) read_error();

   /* change strings to memory references */

   for (i = 0; i < symbols; ++i) {
      struct nlist *p = symtab + i;

      p->n_un.n_name = strings + p->n_un.n_strx;
   }

   /* create archive file */

   printf("Creating references to shared library `%s' ...\n", linker_library);

   library = fopen(linker_library, "w");
   if (!library) create_error(linker_library);

   if (fwrite(ARMAG, 1, SARMAG, library) != SARMAG) write_error();

   /* process all object files */

   for (i = 0; i < symbols; ++i) {
      struct nlist *p = symtab + i;

      if ((p->n_type & N_TYPE) == N_FN) {
	 char full_name[1024];

	 sprintf(full_name, "%s/%s", module_directory, p->n_un.n_name);
	 if (!access(full_name, R_OK)) {
	    struct ar_hdr hdr;
	    int size, start = ftell(library);

	    copy_string(hdr.ar_name, p->n_un.n_name, 16);
	    copy_number(hdr.ar_date, time(NULL), 12);
	    copy_number(hdr.ar_uid, getuid(), 6);
	    copy_number(hdr.ar_gid, getgid(), 6);
	    copy_string(hdr.ar_mode, "444", 8);
	    copy_string(hdr.ar_size, " ", 10);
	    copy_string(hdr.ar_fmag, ARFMAG, 2);
	    if (fwrite(&hdr, sizeof hdr, 1, library) != 1) write_error();

	    size = write_object_file(full_name, symbols, symtab, library);

	    copy_number(hdr.ar_size, size, 10);
	    if (fseek(library, start, SEEK_SET)) read_error();
	    if (fwrite(&hdr, sizeof hdr, 1, library) != 1) write_error();

	    if (fseek(library, start + sizeof hdr + size, SEEK_SET))
	       read_error();

	    if (size & 1) fputc('\n', library);
	 }
      }
   }
   fclose(library);
   free(strings);
   free(symtab);
}

/*****************************************************************************/
/****************************** Main Program *********************************/
/*****************************************************************************/

int main(int argc, char *argv[]) {
   char *s;

   for (program_name = s = *argv; *s;)
      if (*s++ == '/') program_name = s;

   if (argc != 8) usage();

   input_module = argv[1];
   shared_library = argv[2];
   linker_library = argv[3];
   module_directory = argv[4];

   errno = 0;
   version = strtol(argv[5], &s, 0);
   if (errno || *s) usage();

   text_address = (caddr_t) strtoul(argv[6], &s, 0);
   if (errno || *s) usage();

   data_address = (caddr_t) strtoul(argv[7], &s, 0);
   if (errno || *s) usage();

   /* open linker's output module (e.g. libc_s.o) */
   initialize();

   /* create shared library (e.g. libs_s) */
   create_shared_library();

   /* create library with references to shared library (e.g. libc_s.a) */
   write_linker_library();

   fclose(in);

   return 0;
}
