]/*-
 * Copyright (c) 1992, 1993, University of Vermont and State
 *  Agricultural College.
 * Copyright (c) 1992, 1993, Garrett A. Wollman.
 *
 * Portions:
 * Copyright (c) 1990, 1991, William F. Jolitz
 * Copyright (c) 1990, The Regents of the University of California
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	Vermont and State Agricultural College and Garrett A. Wollman,
 *	by William F. Jolitz, by the University of California,
 *	Berkeley, by Larwence Berkeley Laboratory, and its contributors.
 * 4. Neither the names of the Universities nor the names of the authors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE UNIVERSITY OR AUTHORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

/* $Id: if_ie.c,v 1.9 93/01/23 14:22:21 wollman Exp Locker: wollman $
 * $Log:	if_ie.c,v $
 * Revision 1.9  93/01/23  14:22:21  wollman
 * Added some StarLAN Fiber modifications from L. Jonas Olsson.
 * 
 * Revision 1.8  93/01/23  14:02:34  wollman
 * Added support for IP multicasting.
 * 
 * Revision 1.7  92/11/21  17:51:59  wollman
 * Added StarLAN Fiber version point to my driver.
 * 
 * Revision 1.6  92/11/20  18:21:45  wollman
 * Fixed documentation, made out-of-mbuf condition an error.
 * 
 * Revision 1.5  92/11/20  18:14:02  wollman
 * Move change of ethertype to host byte order so that it's after
 * bpf_mtap.  Hopefully, this will allow us to listen to incoming
 * IP packets.
 * 
 * Revision 1.4  92/11/20  18:11:00  wollman
 * Added code to drop packets when we run out of mbufs rather than
 * panicking.  This should alleviate the problems observed when using tcpdump.
 * 
 * Revision 1.3  92/11/14  14:12:40  wollman
 * Added BPF support code.
 * 
 * Revision 1.2  92/11/14  13:30:26  wollman
 * Added RCS keywords
 * 
 */

/*
 * Intel 82586 Ethernet chip
 * Register, bit, and structure definitions.
 *
 * Written by GAW with reference to the Clarkson Packet Driver code for this
 * chip written by Russ Nelson and others.
 *
 * BPF support code stolen directly from hpdev/if_le.c, supplied with
 * tcpdump.
 */

/*
 * The i82586 is a very versatile chip, found in many implementations.
 * Programming this chip is mostly the same, but certain details differ
 * from card to card.  This driver is written so that different cards 
 * can be automatically detected at run-time.  Currently, only the
 * AT&T EN100/StarLAN 10 series are supported.
 */

/*
Mode of operation:

We run the 82586 in a standard Ethernet mode.  We keep NFRAMES received
frame descriptors around for the receiver to use, and NBUFFS associated
receive buffer descriptors, both in a circular list.  Whenever a frame is
received, we rotate both lists as necessary.  (The 586 treats both lists
as a simple queue.)  We also keep a transmit command around so that packets
can be sent off quickly.

We configure the adapter in AL-LOC = 1 mode, which means that the
Ethernet/802.3 MAC header is placed at the beginning of the receive buffer
rather than being split off into various fields in the RFD.  This also
means that we must include this header in the transmit buffer as well.

By convention, all transmit commands, and only transmit commands, shall
have the I (IE_CMD_INTR) bit set in the command.  This way, when an
interrupt arrives at ieintr(), it is immediately possible to tell
what precisely caused it.  ANY OTHER command-sending routines should
run at splimp(), and should post an acknowledgement to every interrupt
they generate.

The 82586 has a 24-bit address space internally, and the adaptor's
memory is located at the top of this region.  However, the value we are
given in configuration is normally the *bottom* of the adaptor RAM.  So,
we must go through a few gyrations to come up with a kernel virtual address
which represents the actual beginning of the 586 address space.  First,
we autosize the RAM by running through several possible sizes and trying
to initialize the adapter under the assumption that the selected size
is correct.  Then, knowing the correct RAM size, we set up our pointers
in ie_softc[unit].  `iomem' represents the computed base of the 586
address space.  `iomembot' represents the actual configured base
of adapter RAM.  Finally, `iosize' represents the calculated size
of 586 RAM.  Then, when laying out commands, we use the interval
[iomembot, iomembot + iosize); to make 24-pointers, we subtract
iomem, and to make 16-pointers, we subtract iomem and and with 0xffff.

*/

#include "ie.h"
#if NIE > 0

#include "param.h"
#include "systm.h"
#include "mbuf.h"
#include "buf.h"
#include "protosw.h"
#include "socket.h"
#include "ioctl.h"
#include "errno.h"
#include "syslog.h"

#include "net/if.h"
#include "net/if_types.h"
#include "net/if_dl.h"
#include "net/netisr.h"
#include "net/route.h"

#include "bpfilter.h"

#ifdef INET
#include "netinet/in.h"
#include "netinet/in_systm.h"
#include "netinet/in_var.h"
#include "netinet/ip.h"
#include "netinet/if_ether.h"
#endif

#ifdef NS
#include "netns/ns.h"
#include "netns/ns_if.h"
#endif

#include "i386/isa/isa_device.h"
#include "i386/isa/ic/i82586.h"
#include "i386/isa/if_ieatt.h"
#include "i386/isa/icu.h"

#include "vm/vm.h"

#if NBPFILTER > 0
#include "net/bpf.h"
#include "net/bpfdesc.h"
#endif

#ifdef DEBUG
#define IED_RINT 1
#define IED_TINT 2
#define IED_RNR 4
#define IED_CNA 8
#define IED_READFRAME 16
int ie_debug = 0;
#endif

#ifndef ETHERMINLEN
#define ETHERMINLEN 60
#endif

#define IE_BUF_LEN 1532		/* length of some buffer? */


/* Forward declaration */
struct ie_softc;

int ieprobe(struct isa_device *dvp);
int ieattach(struct isa_device *dvp);
int ieinit(int unit);
int ieioctl(struct ifnet *ifp, int command, void *data);
int iestart(struct ifnet *ifp);
static void sl_reset_586(int unit);
static void sl_chan_attn(int unit);
int iereset(int unit, int dummy);
static void ie_readframe(int unit, struct ie_softc *ie, int bufno);
static void ie_drop_packet_buffer(int unit, struct ie_softc *ie);
static void sl_read_ether(int unit, unsigned char addr[6]);
static void find_ie_mem_size(int unit);
static int command_and_wait(int unit, int command, void volatile *pcmd, int);
static int ierint(int unit, struct ie_softc *ie);
static int ietint(int unit, struct ie_softc *ie);
static int iernr(int unit, struct ie_softc *ie);
static void start_receiver(int unit);
int ieget(int, struct ie_softc *, struct mbuf **,
		   struct ether_header *);
static caddr_t setup_rfa(caddr_t ptr, struct ie_softc *ie);
static int mc_setup(int, caddr_t, volatile struct ie_sys_ctl_block *);
#ifdef MULTICAST
static void ie_mc_reset(int unit);
#endif

#ifdef DEBUG
void print_rbd(volatile struct ie_recv_buf_desc *rbd);

int in_ierint = 0;
int in_ietint = 0;
#endif

/*
 * This tells the autoconf code how to set us up.
 */
struct isa_driver iedriver = {
  ieprobe, ieattach, "ie",
};

enum ie_hardware {
  IE_STARLAN10,
  IE_EN100,
  IE_SLFIBER,
  IE_UNKNOWN
};

const char *ie_hardware_names[] = {
  "StarLAN 10",
  "EN100",
  "StarLAN Fiber",
  "Unknown"
};

#define NFRAMES 16		/* number of frames to allow for receive */
#define NBUFFS 32		/* number of buffers to allocate */
#define IE_RBUF_SIZE 256	/* size of each buffer, MUST BE POWER OF TWO */

/*
 * Ethernet status, per interface.
 */
struct ie_softc {
  struct arpcom ns_ac;
#define ns_if ns_ac.ac_if
#define ns_addr ns_ac.ac_enaddr
  void (*ie_reset_586)(int);
  void (*ie_chan_attn)(int);
  enum ie_hardware hard_type;
  int hard_vers;

  u_short port;
  caddr_t iomem;
  caddr_t iomembot;
  unsigned iosize;
  
  int want_mcsetup;
  int promisc;
  volatile struct ie_int_sys_conf_ptr *iscp;
  volatile struct ie_sys_ctl_block *scb;
  volatile struct ie_recv_frame_desc *rframes[NFRAMES];
  volatile struct ie_recv_buf_desc *rbuffs[NBUFFS];
  volatile char *cbuffs[NBUFFS];
  int rfhead, rftail, rbhead, rbtail;

  volatile struct ie_xmit_cmd *xmit_cmd;
  volatile struct ie_xmit_buf *xmit_buf;
  u_char *xmit_cbuf;

  /*
   * The broadcast address is always put here in addition to the
   * multicast addresses.
   */
  struct ie_en_addr mcast_addrs[MAXMCAST + 1];
  int mcast_count;

#if NBPFILTER > 0
  caddr_t ie_bpf;
#endif

} ie_softc[NIE];

#define MK_24(base, ptr) ((caddr_t)((caddr_t)ptr - base))
#define MK_16(base, ptr) ((u_short)MK_24(base, ptr))

#define PORT ie_softc[unit].port
#define MEM ie_softc[unit].iomem


int ieprobe(dvp)
     struct isa_device *dvp;
{
  int unit = dvp->id_unit;
  u_char c, d;

  ie_softc[unit].port = dvp->id_iobase;
  ie_softc[unit].iomembot = dvp->id_maddr;
  ie_softc[unit].iomem = 0;

  c = inb(PORT + IEATT_REVISION);
  switch(SL_BOARD(c)) {
  case SL10_BOARD:
    ie_softc[unit].hard_type = IE_STARLAN10;
    ie_softc[unit].ie_reset_586 = sl_reset_586;
    ie_softc[unit].ie_chan_attn = sl_chan_attn;
    break;
  case EN100_BOARD:
    ie_softc[unit].hard_type = IE_EN100;
    ie_softc[unit].ie_reset_586 = sl_reset_586;
    ie_softc[unit].ie_chan_attn = sl_chan_attn;
    break;

  case SLFIBER_BOARD:
    ie_softc[unit].hard_type = IE_SLFIBER;
    ie_softc[unit].ie_reset_586 = sl_reset_586;
    ie_softc[unit].ie_chan_attn = sl_chan_attn;
    break;

    /*
     * Anything else is not recognized or cannot be used.
     */
  default:
    return 0;
  }

  ie_softc[unit].hard_vers = SL_REV(c);

  /*
   * Divine memory size on-board the card.  Ususally 16k.
   */
  find_ie_mem_size(unit);

  if(!ie_softc[unit].iosize) {
    return 0;
  }

  dvp->id_msize = ie_softc[unit].iosize;

  switch(ie_softc[unit].hard_type) {
  case IE_EN100:
  case IE_STARLAN10:
  case IE_SLFIBER:
    sl_read_ether(unit, ie_softc[unit].ns_addr);
    break;

  default:
    printf("ie%d: unknown AT&T board type code %d\n", unit, 
	   ie_softc[unit].hard_type);
    return 0;
  }

  return 1;
}

/*
 * Taken almost exactly from Bill's if_is.c.
 */
int ieattach(dvp)
     struct isa_device *dvp;
{
  int unit = dvp->id_unit;
  struct ie_softc *ie = &ie_softc[unit];
  struct ifnet *ifp = &ie->ns_if;

  ifp->if_unit = unit;
  ifp->if_name = iedriver.name;
  ifp->if_mtu = ETHERMTU;
  printf("<%s R%d> ethernet address %s", 
	 ie_hardware_names[ie_softc[unit].hard_type],
	 ie_softc[unit].hard_vers + 1,
	 ether_sprintf(ie->ns_addr));

  ifp->if_flags = IFF_BROADCAST | IFF_SIMPLEX | IFF_NOTRAILERS;
#ifdef MULTICAST
  ifp->if_flags  |= IFF_MULTICAST;
#endif /* MULTICAST */

  ifp->if_init = ieinit;
  ifp->if_output = ether_output;
  ifp->if_start = iestart;
  ifp->if_ioctl = ieioctl;
  ifp->if_reset = iereset;
  ifp->if_watchdog = iereset; /* Yes, X3J11 members, this is indeed illegal. */
  ifp->if_type = IFT_ETHER;
  ifp->if_addrlen = 6;
  ifp->if_hdrlen = 14;
  
#if NBPFILTER > 0
  printf("\n");
  bpfattach(&ie_softc[unit].ie_bpf, ifp, DLT_EN10MB,
	    sizeof(struct ether_header));
#endif

  if_attach(ifp);
  {
    struct ifaddr *ifa = ifp->if_addrlist;
    struct sockaddr_dl *sdl;
    while(ifa && ifa->ifa_addr && ifa->ifa_addr->sa_family != AF_LINK)
      ifa = ifa->ifa_next;

    if(!ifa || !ifa->ifa_addr) return;

    /* Provide our ether address to the higher layers */
    sdl = (struct sockaddr_dl *)ifa->ifa_addr;
    sdl->sdl_type = IFT_ETHER;
    sdl->sdl_alen = 6;
    sdl->sdl_slen = 0;
    bcopy(ie->ns_addr, &sdl->sdl_data[sdl->sdl_nlen], 6);
  }
}

/*
 * What to do upon receipt of an interrupt.
 */
int ieintr(unit)
     int unit;
{
  register struct ie_softc *ie = &ie_softc[unit];
  register u_short status;

  status = ie->scb->ie_status;

loop:
  if(status & IE_ST_RECV) {
#ifdef DEBUG
    in_ierint++;
    if(ie_debug & IED_RINT)
      printf("ie%d: rint\n", unit);
#endif
    ierint(unit, ie);
#ifdef DEBUG
    in_ierint--;
#endif
  }

  if(status & IE_ST_DONE) {
#ifdef DEBUG
    in_ietint++;
    if(ie_debug & IED_TINT)
      printf("ie%d: tint\n", unit);
#endif
    ietint(unit, ie);
#ifdef DEBUG
    in_ietint--;
#endif
  }

  if(status & IE_ST_RNR) {
#ifdef DEBUG
    if(ie_debug & IED_RNR)
      printf("ie%d: rnr\n", unit);
#endif
    iernr(unit, ie);
  }

#ifdef DEBUG
  if((status & IE_ST_ALLDONE)
     && (ie_debug & IED_CNA))
    printf("ie%d: cna\n", unit);
#endif

  /* Don't ack interrupts which we didn't receive */
  ie_ack(ie->scb, IE_ST_WHENCE & status, unit, ie->ie_chan_attn);

  if((status = ie->scb->ie_status) & IE_ST_WHENCE)
    goto loop;

  return unit;
}

/*
 * Process a received-frame interrupt.
 */
static int ierint(unit, ie)
     int unit;
     struct ie_softc *ie;
{
  int i, status;
  static int timesthru = 1024;

  i = ie->rfhead;
  while(1) {
    status = ie->rframes[i]->ie_fd_status;

    if((status & IE_FD_COMPLETE) && (status & IE_FD_OK)) {
      ie->ns_if.if_ipackets++;
      if(!--timesthru) {
	ie->ns_if.if_ierrors += ie->scb->ie_err_crc + ie->scb->ie_err_align +
	  ie->scb->ie_err_resource + ie->scb->ie_err_overrun;
	ie->scb->ie_err_crc = 0;
	ie->scb->ie_err_align = 0;
	ie->scb->ie_err_resource = 0;
	ie->scb->ie_err_overrun = 0;
	timesthru = 1024;
      }
      ie_readframe(unit, ie, i);
    } else {
      break;
    }
    i = (i + 1) % NFRAMES;
  }
  return 0;
}

/*
 * Process a command-complete interrupt.  These are only generated by
 * the transmission of frames.  This routine is deceptively simple, since
 * most of the real work is done by iestart().
 */
static int ietint(unit, ie)
     int unit;
     struct ie_softc *ie;
{
  int status = ie->xmit_cmd->ie_xmit_status;

  ie->ns_if.if_timer = 0;
  ie->ns_if.if_flags &= ~IFF_OACTIVE;

  if(status & IE_XS_LATECOLL) {
    printf("ie%d: late collision\n", unit);
    ie->ns_if.if_collisions++;
    ie->ns_if.if_oerrors++;
  } else if(status & IE_XS_NOCARRIER) {
    printf("ie%d: no carrier\n", unit);
    ie->ns_if.if_oerrors++;
  } else if(status & IE_XS_LOSTCTS) {
    printf("ie%d: lost CTS\n", unit);
    ie->ns_if.if_oerrors++;
  } else if(status & IE_XS_UNDERRUN) {
    printf("ie%d: DMA underrun\n", unit);
    ie->ns_if.if_oerrors++;
  } else if(status & IE_XS_EXCMAX) {
    printf("ie%d: too many collisions\n", unit);
    ie->ns_if.if_collisions += 16;
    ie->ns_if.if_oerrors++;
  } else {
    ie->ns_if.if_opackets++;
    ie->ns_if.if_collisions += status & IE_XS_MAXCOLL;
  }

  /*
   * If multicast addresses were added or deleted while we were transmitting,
   * ie_mc_reset() set the want_mcsetup flag indicating that we should do it.
   */
  if(ie->want_mcsetup) {
    ie->want_mcsetup = 0;
    mc_setup(unit, (caddr_t)ie->xmit_cbuf, ie->scb);
  }

  /* Wish I knew why this seems to be necessary... */
  ie->xmit_cmd->ie_xmit_status |= IE_STAT_COMPL;

  iestart(&ie->ns_if);
}

/*
 * Process a receiver-not-ready interrupt.  I believe that we get these
 * when there aren't enough buffers to go around.  For now (FIXME), we
 * just restart the receiver, and hope everything's ok.
 */
static int iernr(unit, ie)
     int unit;
     struct ie_softc *ie;
{
  command_and_wait(unit, IE_RU_DISABLE, 0, 0); /* just in case */
  setup_rfa((caddr_t)ie->rframes[0], ie);

  ie_softc[unit].scb->ie_recv_list = MK_16(MEM, ie_softc[unit].rframes[0]);
  command_and_wait(unit, IE_RU_ENABLE, 0, 0);

  ie_ack(ie->scb, IE_ST_WHENCE, unit, ie->ie_chan_attn);

  ie->ns_if.if_ierrors++;
  return 0;
}

/*
 * Read frame NUM from unit UNIT (pre-cached as IE).
 *
 * This routine reads the RFD at NUM, and copies in the buffers from
 * the list of RBD, then rotates the RBD and RFD lists so that the receiver
 * doesn't start complaining.  Trailers are DROPPED---there's no point
 * in wasting time on confusing code to deal with them.  Hopefully,
 * this machine will never ARP for trailers anyway.
 */
static void ie_readframe(unit, ie, num)
     int unit;
     struct ie_softc *ie;
     int num;			/* frame number to read */
{
  struct ie_recv_frame_desc rfd = *(ie->rframes[num]);
  struct mbuf *m = 0;
  struct ether_header eh;

  /* Immediately advance the RFD list, since we we have copied ours now. */
  ie->rframes[num]->ie_fd_status = 0;
  ie->rframes[num]->ie_fd_last |= IE_FD_LAST;
  ie->rframes[ie->rftail]->ie_fd_last &= ~IE_FD_LAST;
  ie->rftail = (ie->rftail + 1) % NFRAMES;
  ie->rfhead = (ie->rfhead + 1) % NFRAMES;

  if(rfd.ie_fd_status & IE_FD_OK) {
    if(ieget(unit, ie, &m, &eh)) {
      ie->ns_if.if_ierrors++;	/* this counts as an error */
      return;
    }
  }

#ifdef DEBUG
  if(ie_debug & IED_READFRAME) {
    printf("ie%d: frame from ether %s type %x\n", unit,
	   ether_sprintf(eh.ether_shost), (unsigned)eh.ether_type);
  }
  if(ntohs(eh.ether_type) > ETHERTYPE_TRAIL
     && ntohs(eh.ether_type) < (ETHERTYPE_TRAIL + ETHERTYPE_NTRAILER))
    printf("received trailer!\n");
#endif

  if(!m) return;

#if NBPFILTER > 0 || defined(MULTICAST)
#if NBPFILTER > 0
  /*
   * Check for a BPF filter; if so, hand it up.
   * Note that we have to stick an extra mbuf up front, because
   * bpf_mtap expects to have the ether header at the front.
   * It doesn't matter that this results in an ill-formatted mbuf chain,
   * since BPF just looks at the data.
   */
  if(ie->ie_bpf) {
    struct mbuf *m2, *m3;

    /* Get an mbuf */
    MGET(m2, M_DONTWAIT, MT_DATA);
    if(m2) {
      m2->m_len = sizeof eh;

      bcopy(&eh, mtod(m2, char *), sizeof eh);
      m2->m_next = m;

      /* Pass it up */
      bpf_mtap(ie->ie_bpf, m);

      /* Get rid of it */
      MFREE(m2, m3);		/* m2 is a dummy */

    }
#endif /* NBPFILTER > 0 */
    /* From tcpdump: hpdev/if_le.c */
    /*
     * If we are promiscuous, we have to check if this packet is really ours.
     */
#ifdef MULTICAST
    if(ie->promisc && bcmp(eh.ether_dhost, ie->ns_addr, sizeof ie->ns_addr) ) {
      if(!(eh.ether_dhost[0] & 1))
	return;			/* not multicast, can't possibly match */

      /* If not catching all multicasts, filter. */
      if(!(ie->ns_if.if_flags & IFF_ALLMULTI)) {
	int i;
	for(i = 0; i < ie->mcast_count; i++) {
	  if(!bcmp(eh.ether_dhost, &ie->mcast_addrs[i], sizeof ie->ns_addr))
	    break;
	}
	if(i >= ie->mcast_count)
	  return;
	/* otherwise matched a MC address */
      }
    }
#else /* not MULTICAST */
    if(ie->promisc
       && bcmp(eh.ether_dhost, ie->ns_addr, sizeof ie->ns_addr) != 0
       && bcmp(eh.ether_dhost, &ie->mcast_addrs[0], sizeof ie->ns_addr) != 0)
      return;
#endif /* MULTICAST */
#endif /* NBPFILTER > 0 || defined(MULTICAST) */

  eh.ether_type = ntohs(eh.ether_type);

  /*
   * Finally pass this packet up to higher layers.
   */
  ether_input(&ie->ns_if, &eh, m);
}

/*
 * Read data off the interface, and turn it into an mbuf chain.
 */
int ieget(unit, ie, mp, ehp)
     int unit;
     struct ie_softc *ie;
     struct mbuf **mp;
     struct ether_header *ehp;
{
  struct mbuf *m, *top, **mymp;
  int i, j;
  int offset;
  int thislen, totlen = 0, len;
  int ohead;

  MGETHDR(*mp, M_DONTWAIT, MT_DATA);
  if(!*mp) {
    ie_drop_packet_buffer(unit, ie);
    return -1;
  }

  m = *mp;
  m->m_pkthdr.rcvif = (struct ifnet *)ie;
  m->m_len = MHLEN;

  top = 0;
  mymp = &top;
  
  i = ohead = ie->rbhead;

  do {
    /*
     * This means we are somehow out of sync.  So, we reset the
     * adapter.
     */
    if(!(ie->rbuffs[i]->ie_rbd_actual & IE_RBD_USED)) {
#ifdef DEBUG
      print_rbd(ie->rbuffs[i]);
#endif
      log(LOG_ERR, "ie%d: receive descriptors out of sync at %d\n",
	  unit, i);
      m_freem(*mp);
      iereset(unit, 0);
      return -1;
    }

    offset = 0;
    /*
     * We want to isolate the bits that have meaning...  This assumes that
     * IE_RBUF_SIZE is an even power of two.  If somehow the act_len exceeds
     * the size of the buffer, then we are screwed anyway.
     */
    thislen = ie->rbuffs[i]->ie_rbd_actual
      & (IE_RBUF_SIZE | (IE_RBUF_SIZE - 1));

    /*
     * If this is the first receive buffer, immediately copy
     * the Ethernet header into ehp.
     */
    if(i == ohead) {
      bcopy((caddr_t)ie->cbuffs[i], (caddr_t)ehp, sizeof *ehp);
      thislen -= sizeof *ehp;
      offset = sizeof *ehp;
    }

    len = thislen;

    do {
      if(top) {
	MGET(m, M_DONTWAIT, MT_DATA);
	if(!m) {
	  m_freem(top);
	  ie_drop_packet_buffer(unit, ie); /* the depends on rbhead being
					      correct */
	  return -1;
	}
	m->m_len = MLEN;
      }

      /*
       * If this is big enough to merit an mbuf cluster, allocate one.
       * Else, just use a regular mbuf.
       */
      if(len >= MINCLSIZE) {
	MCLGET(m, M_DONTWAIT);
	if(m->m_flags & M_EXT)
	  m->m_len = len = min(len, MCLBYTES);
	else
	  len = m->m_len;
      } else {
	if(len < m->m_len) {
	  if(!top && len + max_linkhdr <= m->m_len)
	    m->m_data += max_linkhdr;
	  m->m_len = len;
	} else {
	  len = m->m_len;
	}
      }
      bcopy((caddr_t)(ie->cbuffs[i] + offset),
	    mtod(m, caddr_t), (unsigned)len);
      offset += len;
      *mymp = m;		/* first time through, this sets top */
      mymp = &m->m_next;
      thislen -= len;
      totlen += len;
    } while(thislen > 0);

    j = ie->rbuffs[i]->ie_rbd_actual;

    ie->rbuffs[i]->ie_rbd_actual = 0;

    ie->rbuffs[ie->rbhead]->ie_rbd_length |= IE_RBD_LAST;
    ie->rbhead = (ie->rbhead + 1) % NBUFFS;
    ie->rbuffs[ie->rbtail]->ie_rbd_length &= ~IE_RBD_LAST;
    ie->rbtail = (ie->rbtail + 1) % NBUFFS;

    i = (i + 1) % NBUFFS;
  } while(!(j & IE_RBD_LAST));

  (*mp)->m_pkthdr.len = totlen;

  return 0;
}

static void ie_drop_packet_buffer(int unit, struct ie_softc *ie) {
  int i;

  do {
    /*
     * This means we are somehow out of sync.  So, we reset the
     * adapter.
     */
    if(!(ie->rbuffs[ie->rbhead]->ie_rbd_actual & IE_RBD_USED)) {
#ifdef DEBUG
      print_rbd(ie->rbuffs[ie->rbhead]);
#endif
      log(LOG_ERR, "ie%d: receive descriptors out of sync at %d\n",
	  unit, i);
      iereset(unit, 0);
      return;
    }

    i = ie->rbuffs[i]->ie_rbd_actual & IE_RBD_LAST;

    ie->rbuffs[ie->rbhead]->ie_rbd_actual = 0;
    ie->rbuffs[ie->rbhead]->ie_rbd_length |= IE_RBD_LAST;
    ie->rbhead = (ie->rbhead + 1) % NBUFFS;
    ie->rbuffs[ie->rbtail]->ie_rbd_length &= ~IE_RBD_LAST;
    ie->rbtail = (ie->rbtail + 1) % NBUFFS;
  } while(!i);
}


/*
 * Start transmission on an interface.
 */
int iestart(ifp)
     struct ifnet *ifp;
{
  struct ie_softc *ie = &ie_softc[ifp->if_unit];
  struct mbuf *m0, *m;
  unsigned char *buffer;
  u_short len;
  int i;

  if(!(ifp->if_flags & IFF_RUNNING))
    return 0;
  if(ifp->if_flags & IFF_OACTIVE)
    return 0;

  do {
    /*
     * If last transmission didn't finish, forget it.
     */
    if(!(ie->xmit_cmd->ie_xmit_status & IE_STAT_COMPL)) {
      printf("ie%d: output not active but complete not set\n", ifp->if_unit);
      iereset(ifp->if_unit, 0);
      return 0;
    }

    IF_DEQUEUE(&ie->ns_if.if_snd, m);
    if(!m)
      return;

    buffer = ie->xmit_cbuf;
    len = 0;

    for(m0 = m; m && len < IE_BUF_LEN; m = m->m_next) {
      bcopy(mtod(m, caddr_t), buffer, m->m_len);
      buffer += m->m_len;
      len += m->m_len;
    }

    m_freem(m0);
    len = MAX(len, ETHERMINLEN);

#if NBPFILTER > 0
    /*
     * See if bpf is listening on this interface, let it see the packet
     * before we commit it to the wire.
     */
    if(ie->ie_bpf)
      bpf_tap(ie->ie_bpf, ie->xmit_cbuf, len);
#endif

    ie->xmit_buf->ie_xmit_flags = IE_XMIT_LAST | len;
    ie->xmit_buf->ie_xmit_next = 0xffff;
    ie->xmit_buf->ie_xmit_buf = MK_24(ie->iomem, ie->xmit_cbuf);

    ie->xmit_cmd->com.ie_cmd_cmd = IE_CMD_LAST | IE_CMD_XMIT | IE_CMD_INTR;
    ie->xmit_cmd->ie_xmit_status = 0;
    ie->xmit_cmd->ie_xmit_desc = MK_16(ie->iomem, ie->xmit_buf);

    ie->scb->ie_command_list = MK_16(ie->iomem, ie->xmit_cmd);

    /*
     * By passing the command pointer as a null, we tell
     * command_and_wait() to pretend that this isn't an action
     * command.  I wish I understood what was happening here.
     */
    command_and_wait(ifp->if_unit, IE_CU_START, 0, 0);

    ifp->if_flags |= IFF_OACTIVE;
    ifp->if_timer = 10;		/* set watchdog to expire in ten seconds */
  } while(0);			/* someday, we may have more than one... */

  return 0;
}

/*
 * Check to see if there's an 82586 out there.
 */
int check_ie_present(unit, where, size)
     int unit;
     caddr_t where;
     unsigned size;
{
  volatile struct ie_sys_conf_ptr *scp;
  volatile struct ie_int_sys_conf_ptr *iscp;
  volatile struct ie_sys_ctl_block *scb;
  u_long realbase;
  int s;

  s = splimp();

  realbase = (u_long)where + size - (1 << 24);

  scp = (void *)(realbase + IE_SCP_ADDR);
  bzero((char *)scp, sizeof *scp);
  
  /*
   * First we put the ISCP at the bottom of memory; this tests to make
   * sure that our idea of the size of memory is the same as the controller's.
   * This is NOT where the ISCP will be in normal operation.
   */
  iscp = (void *)where;
  bzero((char *)iscp, sizeof *iscp);

  scb = (void *)where;
  bzero((char *)scb, sizeof *scb);

  scp->ie_bus_use = 0;		/* 16-bit */
  scp->ie_iscp_ptr = (caddr_t)((caddr_t)iscp - (caddr_t)realbase);

  iscp->ie_busy = 1;
  iscp->ie_scb_offset = MK_16(realbase, scb) + 256;

  (*ie_softc[unit].ie_reset_586)(unit);
  (*ie_softc[unit].ie_chan_attn)(unit);

  DELAY(100);			/* wait a while... */

  if(iscp->ie_busy) {
    splx(s);
    return 0;
  }

  /*
   * Now relocate the ISCP to its real home, and reset the controller
   * again.
   */
  iscp = (void *)Align((caddr_t)(realbase + IE_SCP_ADDR - 
				 sizeof(struct ie_int_sys_conf_ptr)));
  bzero((char *)iscp, sizeof *iscp);

  scp->ie_iscp_ptr = (caddr_t)((caddr_t)iscp - (caddr_t)realbase);

  iscp->ie_busy = 1;
  iscp->ie_scb_offset = MK_16(realbase, scb);

  (*ie_softc[unit].ie_reset_586)(unit);
  (*ie_softc[unit].ie_chan_attn)(unit);

  DELAY(100);

  if(iscp->ie_busy) {
    splx(s);
    return 0;
  }

  ie_softc[unit].iosize = size;
  ie_softc[unit].iomem = (caddr_t)realbase;

  ie_softc[unit].iscp = iscp;
  ie_softc[unit].scb = scb;

  /*
   * Acknowledge any interrupts we may have caused...
   */
  ie_ack(scb, IE_ST_WHENCE, unit, ie_softc[unit].ie_chan_attn);
  splx(s);

  return 1;
}

/*
 * Divine the memory size of ie board UNIT.
 * Better hope there's nothing important hiding just below the ie card...
 */
static void find_ie_mem_size(unit)
     int unit;
{
  unsigned size;

  ie_softc[unit].iosize = 0;

  for(size = 65536; size >= 16384; size -= 16384) {
    if(check_ie_present(unit, ie_softc[unit].iomembot, size)) {
      return;
    }
  }

  return;
}

void sl_reset_586(unit)
     int unit;
{
  outb(PORT + IEATT_RESET, 0);
}

void sl_chan_attn(unit)
     int unit;
{
  outb(PORT + IEATT_ATTN, 0);
}

void sl_read_ether(unit, addr)
     int unit;
     unsigned char addr[6];
{
  int i;

  for(i = 0; i < 6; i++)
    addr[i] = inb(PORT + i);
}


int iereset(unit, dummy)
     int unit, dummy;
{
  int s = splimp();

  if(unit >= NIE) {
    splx(s);
    return -1;
  }

  printf("ie%d: reset\n", unit);

  /*
   * Stop i82586 dead in its tracks.
   */
  if(command_and_wait(unit, IE_RU_ABORT | IE_CU_ABORT, 0, 0))
    printf("ie%d: abort commands timed out\n", unit);

  if(command_and_wait(unit, IE_RU_DISABLE | IE_CU_STOP, 0, 0))
    printf("ie%d: disable commands timed out\n", unit);

  if(!check_ie_present(unit, ie_softc[unit].iomembot, ie_softc[unit].iosize))
    panic("ie disappeared!\n");

  ieinit(unit);
  splx(s);
  return 0;
}

/*
 * This is called if we time out.
 */
static int chan_attn_timeout(rock)
     caddr_t rock;
{
  *(int *)rock = 1;
}

/*
 * Send a command to the controller and wait for it to either
 * complete or be accepted, depending on the command.  If the
 * command pointer is null, then pretend that the command is
 * not an action command.  If the command pointer is not null,
 * and the command is an action command, wait for
 * ((volatile struct ie_cmd_common *)pcmd)->ie_cmd_status & MASK
 * to become true.
 */
static int command_and_wait(unit, cmd, pcmd, mask)
     int unit;
     int cmd;
     volatile void *pcmd;
     int mask;
{
  volatile struct ie_cmd_common *cc = pcmd;
  volatile int timedout = 0;
  extern int hz;

  ie_softc[unit].scb->ie_command = (u_short)cmd;
  
  if(IE_ACTION_COMMAND(cmd) && pcmd) {
    (*ie_softc[unit].ie_chan_attn)(unit);

    /*
     * According to the packet driver, the minimum timeout should be
     * .369 seconds, which we round up to .37.
     */
    timeout(chan_attn_timeout, (caddr_t)&timedout, 37 * hz / 100);

    /*
     * Now spin-lock waiting for status.  This is not a very nice
     * thing to do, but I haven't figured out how, or indeed if, we
     * can put the process waiting for action to sleep.  (We may
     * be getting called through some other timeout running in the
     * kernel.)
     */
    while(1) {
      if((cc->ie_cmd_status & mask) || timedout)
	break;
    }

    untimeout(chan_attn_timeout, (caddr_t)&timedout);

    return timedout;
  } else {
    
    /*
     * Otherwise, just wait for the command to be accepted.
     */
    (*ie_softc[unit].ie_chan_attn)(unit);

    while(ie_softc[unit].scb->ie_command)
      ;				/* spin lock */

    return 0;
  }
}

/*
 * Run the time-domain reflectometer...
 */
static void run_tdr(unit, cmd)
     int unit;
     struct ie_tdr_cmd *cmd;
{
  int result;

  cmd->com.ie_cmd_status = 0;
  cmd->com.ie_cmd_cmd = IE_CMD_TDR | IE_CMD_LAST;
  cmd->com.ie_cmd_link = 0xffff;
  cmd->ie_tdr_time = 0;

  ie_softc[unit].scb->ie_command_list = MK_16(MEM, cmd);
  cmd->ie_tdr_time = 0;

  if(command_and_wait(unit, IE_CU_START, cmd, IE_STAT_COMPL))
    result = 0x2000;
  else
    result = cmd->ie_tdr_time;

  ie_ack(ie_softc[unit].scb, IE_ST_WHENCE, unit, 
	 ie_softc[unit].ie_chan_attn);

  if(result & IE_TDR_SUCCESS)
    return;

  if(result & IE_TDR_XCVR) {
    printf("ie%d: transceiver problem\n", unit);
  } else if(result & IE_TDR_OPEN) {
    printf("ie%d: TDR detected an open %d clocks away\n", unit, 
	   result & IE_TDR_TIME);
  } else if(result & IE_TDR_SHORT) {
    printf("ie%d: TDR detected a short %d clocks away\n", unit,
	   result & IE_TDR_TIME);
  } else {
    printf("ie%d: TDR returned unknown status %x\n", result);
  }
}

static void start_receiver(unit)
     int unit;
{
  int s = splimp();

  ie_softc[unit].scb->ie_recv_list = MK_16(MEM, ie_softc[unit].rframes[0]);
  command_and_wait(unit, IE_RU_START, 0, 0);

  ie_ack(ie_softc[unit].scb, IE_ST_WHENCE, unit, ie_softc[unit].ie_chan_attn);

  splx(s);
}

/*
 * Here is a helper routine for iernr() and ieinit().  This sets up
 * the RFA.
 */
static caddr_t setup_rfa(caddr_t ptr, struct ie_softc *ie) {
  volatile struct ie_recv_frame_desc *rfd = (void *)ptr;
  volatile struct ie_recv_buf_desc *rbd;
  int i;
  int unit = ie - &ie_softc[0];

  /* First lay them out */
  for(i = 0; i < NFRAMES; i++) {
    ie->rframes[i] = rfd;
    bzero((char *)rfd, sizeof *rfd);
    rfd++;
  }

  ptr = (caddr_t)Align((caddr_t)rfd);
  
  /* Now link them together */
  for(i = 0; i < NFRAMES; i++) {
    ie->rframes[i]->ie_fd_next =
      MK_16(MEM, ie->rframes[(i + 1) % NFRAMES]);
  }
  
  /* Finally, set the EOL bit on the last one. */
  ie->rframes[NFRAMES - 1]->ie_fd_last |= IE_FD_LAST;

  /*
   * Now lay out some buffers for the incoming frames.  Note that
   * we set aside a bit of slop in each buffer, to make sure that
   * we have enough space to hold a single frame in every buffer.
   */
  rbd = (void *)ptr;

  for(i = 0; i < NBUFFS; i++) {
    ie->rbuffs[i] = rbd;
    bzero((char *)rbd, sizeof *rbd);
    ptr = (caddr_t)Align(ptr + sizeof *rbd);
    rbd->ie_rbd_length = IE_RBUF_SIZE;
    rbd->ie_rbd_buffer = MK_24(MEM, ptr);
    ie->cbuffs[i] =  (void *)ptr;
    ptr += IE_RBUF_SIZE;
    rbd = (void *)ptr;
  }
  
  /* Now link them together */
  for(i = 0; i < NBUFFS; i++) {
    ie->rbuffs[i]->ie_rbd_next = MK_16(MEM, ie->rbuffs[(i + 1) % NBUFFS]);
  }
  
  /* Tag EOF on the last one */
  ie->rbuffs[NBUFFS - 1]->ie_rbd_length |= IE_RBD_LAST;

  /* We use the head and tail pointers on receive to keep track of
   * the order in which RFDs and RBDs are used. */
  ie->rfhead = 0;
  ie->rftail = NFRAMES - 1;
  ie->rbhead = 0;
  ie->rbtail = NBUFFS - 1;

  ie->scb->ie_recv_list = MK_16(MEM, ie->rframes[0]);
  ie->rframes[0]->ie_fd_buf_desc = MK_16(MEM, ie->rbuffs[0]);

  ptr = Align(ptr);
  return ptr;
}

/*
 * Run the multicast setup command.
 * Call at splimp().
 */
static int mc_setup(int unit, caddr_t ptr, 
		    volatile struct ie_sys_ctl_block *scb) {
  struct ie_softc *ie = &ie_softc[unit];
  volatile struct ie_mcast_cmd *cmd = (void *)ptr;
  
  cmd->com.ie_cmd_status = 0;
  cmd->com.ie_cmd_cmd = IE_CMD_MCAST | IE_CMD_LAST;
  cmd->com.ie_cmd_link = 0xffff;
  
  bcopy((caddr_t)ie->mcast_addrs, (caddr_t)cmd->ie_mcast_addrs,
	ie->mcast_count * sizeof *ie->mcast_addrs);

  cmd->ie_mcast_number = ie->mcast_count;
  
  scb->ie_command_list = MK_16(MEM, cmd);
  if(command_and_wait(unit, IE_CU_START, cmd, IE_STAT_COMPL)
     || !(cmd->com.ie_cmd_status & IE_STAT_OK)) {
    printf("ie%d: multicast address setup command failed\n", unit);
    return 0;
  }
  return 1;
}

/*
 * This routine takes the environment generated by check_ie_present()
 * and adds to it all the other structures we need to operate the adapter.
 * This includes executing the CONFIGURE, IA-SETUP, and MC-SETUP commands,
 * starting the receiver unit, and clearing interrupts.
 *
 * THIS ROUTINE MUST BE CALLED AT splimp() OR HIGHER.
 */
int ieinit(unit)
     int unit;
{
  struct ie_softc *ie = &ie_softc[unit];
  volatile struct ie_sys_ctl_block *scb = ie->scb;
  caddr_t ptr;
  int s;

  s = splimp();
  if(~s & splhigh()) {
    splx(s);
    panic("ieinit not called at splimp() or higher");
  }
  splx(s);

  ptr = (caddr_t)Align((caddr_t)scb + sizeof *scb);

  /*
   * Send the configure command first.
   */
  {
    volatile struct ie_config_cmd *cmd = (void *)ptr;

    ie_setup_config(cmd, ie->promisc);
    cmd->com.ie_cmd_status = 0;
    cmd->com.ie_cmd_cmd = IE_CMD_CONFIG | IE_CMD_LAST;
    cmd->com.ie_cmd_link = 0xffff;

    scb->ie_command_list = MK_16(MEM, cmd);

    if(command_and_wait(unit, IE_CU_START, cmd, IE_STAT_COMPL)
       || !(cmd->com.ie_cmd_status & IE_STAT_OK)) {
      printf("ie%d: configure command failed\n", unit);
      return 0;
    }
  }
  /*
   * Now send the Individual Address Setup command.
   */
  {
    volatile struct ie_iasetup_cmd *cmd = (void *)ptr;

    cmd->com.ie_cmd_status = 0;
    cmd->com.ie_cmd_cmd = IE_CMD_IASETUP | IE_CMD_LAST;
    cmd->com.ie_cmd_link = 0xffff;

    bcopy((char *)ie_softc[unit].ns_addr, (char *)&cmd->ie_address,
	  sizeof cmd->ie_address);

    scb->ie_command_list = MK_16(MEM, cmd);
    if(command_and_wait(unit, IE_CU_START, cmd, IE_STAT_COMPL)
       || !(cmd->com.ie_cmd_status & IE_STAT_OK)) {
      printf("ie%d: individual address setup command failed\n", unit);
      return 0;
    }
  }

  /*
   * Set up standard Ether broadcst address.
   */
  {
    u_char *p = ie->mcast_addrs[0].data;
    int i;
    for(i = 0; i < 6; i++)
      p[i] = 0xff;
  }
  ie->mcast_count = 1;
  
  if(!mc_setup(unit, ptr, scb)) return 0;

  /*
   * Now run the time-domain reflectometer.
   */
  run_tdr(unit, (void *)ptr);

  /*
   * Acknowledge any interrupts we have generated thus far.
   */
  ie_ack(ie->scb, IE_ST_WHENCE, unit, ie->ie_chan_attn);

  /*
   * Set up the RFA.
   */
  ptr = setup_rfa(ptr, ie);

  /*
   * Finally, the transmit command and buffer are the last little bit of work.
   */
  ie->xmit_cmd = (void *)ptr;
  ptr += sizeof *ie->xmit_cmd;
  ptr = Align(ptr);
  ie->xmit_buf = (void *)ptr;
  ptr += sizeof *ie->xmit_buf;
  ptr = Align(ptr);
  ie->xmit_cbuf = (void *)ptr;

  bzero((caddr_t)ie->xmit_cmd, sizeof *ie->xmit_cmd);
  bzero((caddr_t)ie->xmit_buf, sizeof *ie->xmit_buf);

  /*
   * This must be coordinated with iestart() and ietint().
   */
  ie->xmit_cmd->ie_xmit_status = IE_STAT_COMPL;

  start_receiver(unit);
  ie->ns_if.if_flags |= IFF_RUNNING; /* tell higher levels that we are here */
}

static void ie_stop(unit)
    int unit;
{
  command_and_wait(unit, IE_RU_DISABLE, 0, 0);
}

int ieioctl(ifp, command, data)
     struct ifnet *ifp;
     int command;
     void *data;
{
  struct ifaddr *ifa = (struct ifaddr *)data;
  struct ie_softc *ie = &ie_softc[ifp->if_unit];
  struct ifreq *ifr = (struct ifreq *)data;
  int s, error = 0;

  s = splimp();

  switch(command) {
  case SIOCSIFADDR:
    ifp->if_flags |= IFF_UP;

    switch(ifa->ifa_addr->sa_family) {
#ifdef INET
    case AF_INET:
      ieinit(ifp->if_unit);
      ((struct arpcom *)ifp)->ac_ipaddr =
	IA_SIN(ifa)->sin_addr;
      arpwhohas((struct arpcom *)ifp, &IA_SIN(ifa)->sin_addr);
      break;
#endif /* INET */

#ifdef NS
      /* This magic copied from if_is.c; I don't use XNS, so I have no
       * way of telling if this actually works or not.
       */
    case AF_NS:
      {
	struct ns_addr *ina = &(IS_SNS(ifa)->sns_addr);

	if(ns_nullhost(*ina)) {
	  ina->x_host = *(union ns_host *)(ns->ns_addr);
	} else {
	  ifp->if_flags &= ~IFF_RUNNING;
	  bcopy((caddr_t)ina->x_host.c_host,
		(caddr_t)ns->ns_addr,
		sizeof ns->ns_addr);
	}

	ieinit(ifp->if_unit);
      }
      break;
#endif /* NS */

    default:
      ieinit(ifp->if_unit);
      break;
    }
    break;

  case SIOCSIFFLAGS:
    /*
     * Note that this device doesn't have an "all multicast" mode, so we
     * must turn on promiscuous mode and do the filtering manually.
     */
    if((ifp->if_flags & IFF_UP) == 0 &&
       (ifp->if_flags & IFF_RUNNING)) {
      ifp->if_flags &= ~IFF_RUNNING;
      ie_stop(ifp->if_unit);
    } else if((ifp->if_flags & IFF_UP) &&
	      (ifp->if_flags & IFF_RUNNING) == 0) {
      ie_softc[ifp->if_unit].promisc = 
	!!(ifp->if_flags & (IFF_PROMISC|IFF_ALLMULTI));
      ieinit(ifp->if_unit);
    } else if(!ie_softc[ifp->if_unit].promisc &&
	      (ifp->if_flags & (IFF_PROMISC | IFF_ALLMULTI))) {
      ie_softc[ifp->if_unit].promisc = 1;
      ieinit(ifp->if_unit);
    } else if(ie_softc[ifp->if_unit].promisc &&
	      !(ifp->if_flags & (IFF_PROMISC | IFF_ALLMULTI))) {
      ie_softc[ifp->if_unit].promisc = 0;
      ieinit(ifp->if_unit);
    }
    break;

#ifdef useSIOCGHWADDR
  case SIOCGHWADDR:
    bcopy((caddr_t)ns->ns_addr, (caddr_t)&ifr->ifr_data, sizeof ns->ns_addr);
    break;
#endif /* useSIOCGHWADDR */

#ifdef MULTICAST
  case SIOCADDMULTI:
  case SIOCDELMULTI:
    /*
     * Update multicast listeners
     */
    error = ((command == SIOCADDMULTI)
	     ? ether_addmulti((struct ifreq *)data, &ie->ns_ac)
	     : ether_delmulti((struct ifreq *)data, &ie->ns_ac));

    if(error == ENETRESET) {
      /* reset multicast filtering */
      ie_mc_reset(ifp->if_unit);
      error = 0;
    }
    break;
#endif /* MULTICAST */

  default:
    error = EINVAL;
  }

  splx(s);
  return error;
}

#ifdef MULTICAST
static void ie_mc_reset(int unit) {
  struct ie_softc *ie = &ie_softc[unit];
  struct ether_multi *enm;
  struct ether_multistep step;

  /*
   * Step through the list of addresses.
   */
  ie->mcast_count = 1;
  ETHER_FIRST_MULTI(step, &ie->ns_ac, enm);
  while(enm) {
    if(ie->mcast_count > MAXMCAST
       || bcmp(enm->enm_addrlo, enm->enm_addrhi, 6) != 0) {
      ie->ns_if.if_flags |= IFF_ALLMULTI;
      ieioctl(&ie->ns_if, SIOCSIFFLAGS, (void *)0);
      goto setflag;
    }

    bcopy(enm->enm_addrlo, &(ie->mcast_addrs[ie->mcast_count]), 6);
    ie->mcast_count++;
    ETHER_NEXT_MULTI(step, enm);
  }

setflag:
  if(ie->ns_if.if_flags & IFF_OACTIVE) {
    ie->want_mcsetup = 1;
  } else {
    mc_setup(unit, (caddr_t)ie->xmit_cbuf, ie->scb);
  }
}

#endif

#ifdef DEBUG
void print_rbd(volatile struct ie_recv_buf_desc *rbd) {
  printf("RBD at %08lx:\n"
	 "actual %04x, next %04x, buffer %08x\n"
	 "length %04x, mbz %04x\n",
	 (unsigned long)rbd,
	 rbd->ie_rbd_actual, rbd->ie_rbd_next, rbd->ie_rbd_buffer,
	 rbd->ie_rbd_length, rbd->mbz);
}
#endif /* DEBUG */
#endif /* NIE > 0 */

